from datetime import datetime

log_data = [
    "2024-03-15 08:00:00 INFO: User login successful [user=johndoe, source_ip=192.168.1.50]",
    "2024-03-15 08:15:30 WARNING: Disk space low [disk=/dev/sda1, usage=95%]",
    "2024-03-15 08:30:15 ERROR: Database connection failed [source_ip=10.0.0.100]",
    "2024-03-15 09:00:00 INFO: Service started [service=webserver]",
    "2024-03-15 09:45:20 CRITICAL: System crash [source_ip=10.0.0.100, error_code=0xAB12]",
    "2024-03-15 10:15:00 INFO: User logout successful [user=johndoe, source_ip=192.168.1.50]",
    "2024-03-15 10:30:45 ERROR: Invalid request [source_ip=192.168.1.60]",
]

def parse_log_entry(entry):
    parts = entry.split(" ", 2)  # Split into timestamp, level, and the rest
    timestamp_str = parts[0] + " " + parts[1]
    timestamp = datetime.strptime(timestamp_str, "%Y-%m-%d %H:%M:%S")
    level = parts[2].replace(":", "") # Remove the colon

    message_parts = parts[2].split("[", 1)
    message = message_parts[0].replace(f"{level}: ", "").strip()
    attributes_str = message_parts[1].rstrip("]") if len(message_parts) > 1 else ""

    attributes = {}
    if attributes_str:
        kv_pairs = attributes_str.split(",")
        for pair in kv_pairs:
            key, value = pair.strip().split("=")
            attributes[key.strip()] = value.strip()

    return {
        "timestamp": timestamp,
        "level": level,
        "message": message,
        "attributes": attributes,
    }

def filter_logs(logs, level=None, start_time=None, end_time=None, keyword=None):
    filtered_logs = []
    for log in logs:
        if level and log["level"] != level:
            continue
        if start_time and log["timestamp"] < start_time:
            continue
        if end_time and log["timestamp"] > end_time:
            continue
        if keyword and keyword.lower() not in log["message"].lower():
            continue
        filtered_logs.append(log)
    return filtered_logs

def generate_report(logs):
    total_entries = len(logs)
    level_counts = {}
    source_ips = set()

    for log in logs:
        level = log["level"]
        level_counts[level] = level_counts.get(level, 0) + 1
        if "source_ip" in log["attributes"]:
            source_ips.add(log["attributes"]["source_ip"])

    report = f"Log Report:\n"
    report += f"  Total Entries: {total_entries}\n"
    report += f"  Level Counts:\n"
    for level, count in level_counts.items():
        report += f"    - {level}: {count}\n"
    report += f"  Unique Source IPs: {', '.join(source_ips) if source_ips else 'None'}\n"
    return report


# --- Main program execution ---
parsed_logs = []
for entry in log_data:
  parsed_logs.append(parse_log_entry(entry))

# 1. Show all parsed logs
print("--- All Logs ---")
for log in parsed_logs:
    print(log)

# 2. Demonstrate filtering.
print("\n--- Filtered Logs (ERROR Level) ---")
error_logs = filter_logs(parsed_logs, level="ERROR")
for log in error_logs:
    print(log)

start = datetime(2024, 3, 15, 9, 0, 0)
end = datetime(2024, 3, 15, 10, 0, 0)
print(f"\n--- Filtered Logs (Between {start} and {end}) ---")

time_filtered_logs = filter_logs(parsed_logs, start_time=start, end_time=end)
for log in time_filtered_logs:
  print(log)

print(f"\n--- Filtered Logs (Keyword = 'user') ---")
keyword_filtered_logs = filter_logs(parsed_logs, keyword="user")
for log in keyword_filtered_logs:
  print(log)

# 3. Generate and print the report.
print("\n--- Log Report ---")
report = generate_report(parsed_logs)
print(report)