from collections import namedtuple
import copy
import os

# -- ex_1 solution -- #
def ex_1(collection, item):
    """
    Adds an item to a collection, handling both mutable and immutable collections.

    Args:
        collection: A collection (list, tuple, etc.)
        item: Item to add to the collection

    Returns:
        For mutable collections: True if successful
        For immutable collections: A new collection with the item added
    """
    # Check if the collection is mutable (list, set, dict)
    if isinstance(collection, list):
        collection.append(item)
        return True
    elif isinstance(collection, set):
        collection.add(item)
        return True
    elif isinstance(collection, dict):
        # If item is a tuple of (key, value)
        if isinstance(item, tuple) and len(item) == 2:
            key, value = item
            collection[key] = value
            return True
        else:
            raise ValueError("For dictionaries, item must be a (key, value) tuple")
    # Handle immutable collections (tuple, frozenset)
    elif isinstance(collection, tuple):
        # Create and return a new tuple with the added item
        return collection + (item,)
    elif isinstance(collection, frozenset):
        # Create and return a new frozenset with the added item
        return frozenset(collection.union({item}))
    else:
        raise TypeError(f"Unsupported collection type: {type(collection)}")



# -- ex_3 solution -- #
def ex_3():
    """
    Demonstrates the difference between shallow and deep copies.

    Returns:
        Boolean indicating whether the deep copy remained unchanged
    """
    import copy

    # Create a nested list
    original = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]

    # Create a shallow copy using slice
    shallow_copy = original[:]

    # Create a deep copy
    deep_copy = copy.deepcopy(original)

    # Verify that all three are different objects at the top level
    print("Top-level IDs (should all be different):")
    print(f"Original ID: {id(original)}")
    print(f"Shallow copy ID: {id(shallow_copy)}")
    print(f"Deep copy ID: {id(deep_copy)}")

    # But the nested lists share IDs in the shallow copy
    print("\nNested list IDs (shallow copy should match original):")
    print(f"Original first inner list ID: {id(original[0])}")
    print(f"Shallow copy first inner list ID: {id(shallow_copy[0])}")
    print(f"Deep copy first inner list ID: {id(deep_copy[0])}")

    # Modify the first element of the first nested list in the original
    print("\nBefore modification:")
    print(f"Original: {original}")
    print(f"Shallow copy: {shallow_copy}")
    print(f"Deep copy: {deep_copy}")

    original[0][0] = 'X'

    print("\nAfter modification (original[0][0] = 'X'):")
    print(f"Original: {original}")
    print(f"Shallow copy: {shallow_copy}")
    print(f"Deep copy: {deep_copy}")

    # Check if deep copy remained unchanged
    deep_copy_unchanged = deep_copy[0][0] != 'X'

    print("\nExplanation:")
    print("1. With a shallow copy, only the top-level list is copied.")
    print("   The inner lists are still references to the same objects.")
    print("2. With a deep copy, all nested objects are recursively copied.")
    print(f"3. Deep copy remained unchanged: {deep_copy_unchanged}")

    return deep_copy_unchanged


# -- ex_4 solution -- #
def ex_4():
    """
    Demonstrates the utility of tuples as dictionary keys and tuple unpacking.
    """
    # Create a dictionary with coordinate tuples as keys
    grid = {
        (0, 0): "treasure",
        (0, 1): "monster",
        (1, 0): "key",
        (1, 1): "potion",
        (2, 2): "exit"
    }

    # Function to get values from adjacent coordinates
    def get_neighbors(coord):
        x, y = coord  # Tuple unpacking
        adjacent_coords = [
            (x+1, y),  # right
            (x-1, y),  # left
            (x, y+1),  # up
            (x, y-1)   # down
        ]

        # Return values from valid adjacent coordinates
        return [grid[c] for c in adjacent_coords if c in grid]

    # Function that returns multiple values as a tuple
    def get_position_info(coord):
        if coord not in grid:
            return None, None, False

        value = grid[coord]
        neighbors = get_neighbors(coord)
        is_edge = len(neighbors) < 4

        return value, neighbors, is_edge

    # Test the functions
    print("Grid contents:")
    for coord, item in grid.items():
        print(f"Position {coord}: {item}")

    # Get neighbors of (0, 0)
    print("\nNeighbors of position (0, 0):")
    neighbors = get_neighbors((0, 0))
    print(neighbors)

    # Get neighbors of (1, 1)
    print("\nNeighbors of position (1, 1):")
    neighbors = get_neighbors((1, 1))
    print(neighbors)

    # Demonstrate tuple unpacking with function return
    print("\nInformation about position (1, 0):")
    value, neighbors, is_edge = get_position_info((1, 0))
    print(f"Value: {value}")
    print(f"Neighbors: {neighbors}")
    print(f"Is on edge: {is_edge}")

    # Try an invalid position
    print("\nInformation about invalid position (5, 5):")
    value, neighbors, is_edge = get_position_info((5, 5))
    print(f"Value: {value}")
    print(f"Neighbors: {neighbors}")
    print(f"Is on edge: {is_edge}")

    return grid


# -- ex_5 solution -- #
def ex_5():
    """
    Processes student attendance data using sets.
    """
    # Define sets for students who attended each day
    monday = {"Alice", "Bob", "Charlie", "David", "Eve"}
    tuesday = {"Alice", "Bob", "David", "Frank", "Grace"}
    wednesday = {"Alice", "Charlie", "Eve", "Frank", "Grace"}
    thursday = {"Bob", "Charlie", "David", "Eve", "Frank"}
    friday = {"Alice", "Bob", "David", "Grace", "Heidi"}

    # All attendance sets in a list for easy processing
    all_days = [monday, tuesday, wednesday, thursday, friday]
    day_names = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday"]

    # Students who attended all days (intersection of all sets)
    perfect_attendance = monday.intersection(tuesday, wednesday, thursday, friday)
    print("Students with perfect attendance:")
    print(perfect_attendance)

    # Students who attended at least one day (union of all sets)
    any_attendance = monday.union(tuesday, wednesday, thursday, friday)
    print("\nAll students who attended at least one day:")
    print(any_attendance)

    # Students who attended Monday but didn't show up on Friday
    monday_not_friday = monday - friday
    print("\nStudents who attended Monday but not Friday:")
    print(monday_not_friday)

    # Students who only attended on Wednesday
    only_wednesday = wednesday - monday - tuesday - thursday - friday
    print("\nStudents who only attended Wednesday:")
    print(only_wednesday)

    # Count days attended for each student
    attendance_count = {}
    for student in any_attendance:
        # Count days where this student appears
        count = sum(1 for day in all_days if student in day)
        attendance_count[student] = count

    print("\nAttendance count for each student:")
    for student, count in sorted(attendance_count.items()):
        print(f"{student}: {count} days")

    # Additional: Students who missed most classes
    min_attendance = min(attendance_count.values())
    poor_attendance = [s for s, c in attendance_count.items() if c == min_attendance]

    print(f"\nStudents with poorest attendance ({min_attendance} days):")
    print(poor_attendance)

    return attendance_count


# -- ex_6 solution -- #
def ex_6():
    """
    Uses namedtuples to represent and process employee data.
    """
    from collections import namedtuple

    # Define an Employee namedtuple
    Employee = namedtuple('Employee', ['id', 'name', 'dept', 'salary'])

    # Create a list of employees
    employees = [
        Employee(101, "Alice Johnson", "Engineering", 85000),
        Employee(102, "Bob Smith", "Marketing", 72000),
        Employee(103, "Charlie Davis", "Engineering", 92000),
        Employee(104, "Diana Miller", "HR", 65000),
        Employee(105, "Edward Wilson", "Marketing", 68000),
        Employee(106, "Fiona Taylor", "Engineering", 79000),
        Employee(107, "George Brown", "HR", 61000)
    ]

    # Print all employees
    print("All employees:")
    for emp in employees:
        print(f"ID: {emp.id}, Name: {emp.name}, Department: {emp.dept}, Salary: ${emp.salary}")

    # Calculate average salary by department
    # First, group employees by department
    depts = {}
    for emp in employees:
        if emp.dept not in depts:
            depts[emp.dept] = []
        depts[emp.dept].append(emp)

    # Calculate average for each department
    print("\nAverage salary by department:")
    for dept, dept_employees in depts.items():
        avg_salary = sum(emp.salary for emp in dept_employees) / len(dept_employees)
        print(f"{dept}: ${avg_salary:.2f}")

    # Find highest paid employee in each department
    print("\nHighest paid employee in each department:")
    for dept, dept_employees in depts.items():
        highest_paid = max(dept_employees, key=lambda emp: emp.salary)
        print(f"{dept}: {highest_paid.name} (${highest_paid.salary})")

    # Create a new list with all employees receiving a 10% raise
    employees_with_raise = [
        emp._replace(salary=int(emp.salary * 1.1))
        for emp in employees
    ]

    print("\nEmployees after 10% raise:")
    for emp in employees_with_raise:
        print(f"ID: {emp.id}, Name: {emp.name}, Department: {emp.dept}, Salary: ${emp.salary}")

    # Additional: Calculate total salary increase
    total_before = sum(emp.salary for emp in employees)
    total_after = sum(emp.salary for emp in employees_with_raise)
    print(f"\nTotal salary before raise: ${total_before}")
    print(f"Total salary after raise: ${total_after}")
    print(f"Total increase: ${total_after - total_before}")

    return employees, employees_with_raise


# -- ex_7 solution -- #
def ex_7():
    """
    Simulates a simple inventory management system using multiple collection types.
    """
    # Initialize an empty inventory
    inventory = {}

    def add_item(inventory, item_id, item_name, quantity, price, categories):
        """Add a new item to the inventory."""
        # Create the item key (tuple)
        item_key = (item_id, item_name)

        # Check if item already exists
        if item_key in inventory:
            raise ValueError(f"Item {item_id}: {item_name} already exists in inventory")

        # Convert categories to a set if it's not already
        if not isinstance(categories, set):
            categories = set(categories)

        # Add the item
        inventory[item_key] = {
            'quantity': quantity,
            'price': price,
            'categories': categories
        }

        return True

    def update_quantity(inventory, item_id, item_name, change):
        """Update the quantity of an item."""
        item_key = (item_id, item_name)

        # Check if item exists
        if item_key not in inventory:
            raise ValueError(f"Item {item_id}: {item_name} not found in inventory")

        # Update quantity
        new_quantity = inventory[item_key]['quantity'] + change

        # Ensure quantity doesn't go below zero
        if new_quantity < 0:
            raise ValueError(f"Cannot reduce quantity below zero. Current: {inventory[item_key]['quantity']}, Change: {change}")

        inventory[item_key]['quantity'] = new_quantity
        return new_quantity

    def get_items_by_category(inventory, category):
        """Get all items that belong to a specific category."""
        # Find items that have the category in their 'categories' set
        matching_items = []

        for item_key, item_data in inventory.items():
            if category in item_data['categories']:
                # Create a copy of the item with its key info
                item_with_key = dict(item_data)
                item_with_key['id'] = item_key[0]
                item_with_key['name'] = item_key[1]
                matching_items.append(item_with_key)

        return matching_items

    def get_total_value(inventory):
        """Calculate the total value of all items in inventory."""
        total = 0
        for item_data in inventory.values():
            total += item_data['quantity'] * item_data['price']
        return total

    # Add some items to the inventory
    try:
        add_item(inventory, 101, "Laptop", 5, 1200.00, {"Electronics", "Computers"})
        add_item(inventory, 102, "Smartphone", 10, 800.00, {"Electronics", "Mobile"})
        add_item(inventory, 103, "Desk Chair", 8, 250.00, {"Furniture", "Office"})
        add_item(inventory, 104, "Desk Lamp", 12, 45.00, {"Lighting", "Office"})
        add_item(inventory, 105, "Headphones", 15, 150.00, {"Electronics", "Audio"})

        # Print the initial inventory
        print("Initial Inventory:")
        for (item_id, item_name), item_data in inventory.items():
            print(f"ID: {item_id}, Name: {item_name}, Quantity: {item_data['quantity']}, "
                  f"Price: ${item_data['price']:.2f}, Categories: {item_data['categories']}")

        # Update quantities
        print("\nUpdating quantities...")
        update_quantity(inventory, 101, "Laptop", -2)  # Sell 2 laptops
        update_quantity(inventory, 102, "Smartphone", 5)  # Add 5 more smartphones

        # Try to update a non-existent item
        try:
            update_quantity(inventory, 999, "Tablet", 3)
        except ValueError as e:
            print(f"Error: {e}")

        # Print items by category
        print("\nElectronics items:")
        electronics = get_items_by_category(inventory, "Electronics")
        for item in electronics:
            print(f"ID: {item['id']}, Name: {item['name']}, Quantity: {item['quantity']}, "
                  f"Price: ${item['price']:.2f}")

        print("\nOffice items:")
        office_items = get_items_by_category(inventory, "Office")
        for item in office_items:
            print(f"ID: {item['id']}, Name: {item['name']}, Quantity: {item['quantity']}, "
                  f"Price: ${item['price']:.2f}")

        # Calculate total inventory value
        total_value = get_total_value(inventory)
        print(f"\nTotal inventory value: ${total_value:.2f}")

    except ValueError as e:
        print(f"Error: {e}")

    return inventory


# -- ex_8 solution -- #
def ex_8():
    """
    Processes text data from a file and uses advanced data structures for analysis.
    """
    from collections import namedtuple, Counter
    import os

    # Define a namedtuple for word information
    Word = namedtuple('Word', ['text', 'count', 'first_line', 'last_line'])

    # Create a sample text file for this example
    sample_text = """Python is a programming language that lets you work quickly
and integrate systems more effectively. Python is powerful and fast;
plays well with others; runs everywhere; is friendly and easy to learn;
is open source; and has a supportive community. The Python community
is vibrant and diverse, with workshops, conferences, and meetups.
Python has a huge standard library and many third-party packages."""

    # Write sample text to a file
    with open('sample_text.txt', 'w') as f:
        f.write(sample_text)

    # Function to read the file and process it
    def analyze_text(filename):
        # Read the file
        with open(filename, 'r') as f:
            lines = f.readlines()

        # Process line by line
        word_counts = Counter()
        word_first_line = {}
        word_last_line = {}

        for i, line in enumerate(lines, 1):
            # Normalize text: lowercase, remove punctuation
            words = ''.join(c if c.isalnum() or c.isspace() else ' ' for c in line.lower()).split()

            for word in words:
                # Count word
                word_counts[word] += 1

                # Track first appearance
                if word not in word_first_line:
                    word_first_line[word] = i

                # Update last appearance
                word_last_line[word] = i

        # Create Word namedtuples for each unique word
        word_info = {}
        for word, count in word_counts.items():
            word_info[word] = Word(
                text=word,
                count=count,
                first_line=word_first_line[word],
                last_line=word_last_line[word]
            )

        return word_info

    # Analyze the text
    word_info = analyze_text('sample_text.txt')

    # Number of unique words
    print(f"Number of unique words: {len(word_info)}")

    # Most common words
    top_words = sorted(word_info.values(), key=lambda w: w.count, reverse=True)[:5]
    print("\nTop 5 most common words:")
    for word in top_words:
        print(f"'{word.text}': {word.count} occurrences, first on line {word.first_line}, last on line {word.last_line}")

    # Create a shallow copy and a deep copy
    shallow_copy = word_info.copy()
    deep_copy = copy.deepcopy(word_info)

    # Modify the original
    for word in list(word_info.keys())[:3]:  # Modify first 3 words
        word_info[word] = word_info[word]._replace(count=999)

    # Show that shallow copy shares references to the namedtuples
    print("\nDemonstrating shallow vs deep copy:")
    modified_word = next(iter(word_info.keys()))
    print(f"Original '{modified_word}' count: {word_info[modified_word].count}")
    print(f"Shallow copy '{modified_word}' count: {shallow_copy[modified_word].count}")
    print(f"Deep copy '{modified_word}' count: {deep_copy[modified_word].count}")

    # Save results to a file
    with open('word_analysis.txt', 'w') as f:
        f.write(f"WORD ANALYSIS RESULTS\n")
        f.write(f"Total unique words: {len(word_info)}\n\n")

        f.write("TOP 10 MOST COMMON WORDS\n")
        for word in sorted(deep_copy.values(), key=lambda w: w.count, reverse=True)[:10]:
            f.write(f"{word.text}: {word.count} occurrences\n")

        f.write("\nWORDS THAT APPEAR ONLY ONCE\n")
        single_words = [w for w in deep_copy.values() if w.count == 1]
        for word in single_words:
            f.write(f"{word.text} (line {word.first_line})\n")

    print(f"\nAnalysis saved to 'word_analysis.txt'")

    # Clean up the sample file
    os.remove('sample_text.txt')

    return deep_copy


if __name__ == "__main__":
    # Uncomment the function you want to run
    ex_1([1, 2, 3], 4)
    # ex_2()
    # ex_3()
    # ex_4()
    # ex_5()
    # ex_6()
    # ex_7()
    # ex_8()
