from collections import namedtuple
import json
import copy

# Define namedtuples
Customer = namedtuple('Customer', ['id', 'name', 'address'])
Product = namedtuple('Product', ['id', 'name', 'price', 'category'])

def add_order(orders, customer, products_dict, products_catalog):
    """
    Add a new order to the system

    Args:
        orders (dict): The order tracking system
        customer (Customer): Customer namedtuple
        products_dict (dict): Dictionary mapping product_id to quantity
        products_catalog (dict): Dictionary of all products (id → Product)

    Returns:
        int: The new order_id
    """
    # Validate products exist
    for product_id in products_dict:
        if product_id not in products_catalog:
            raise ValueError(f"Product {product_id} does not exist")

    # Calculate total
    total = sum(products_catalog[product_id].price * quantity
                for product_id, quantity in products_dict.items())

    # Get new order ID
    order_id = len(orders) + 1

    # Create order
    orders[order_id] = {
        'customer': customer,
        'products': copy.deepcopy(products_dict),  # Deep copy to prevent modification
        'status': 'new',
        'total': total
    }

    return order_id

def update_order(orders, order_id, new_products=None, new_status=None, products_catalog=None):
    """
    Update an existing order

    Args:
        orders (dict): The order tracking system
        order_id (int): The order to update
        new_products (dict, optional): New products to add/update
        new_status (str, optional): New status for the order
        products_catalog (dict): Dictionary of all products (needed if new_products provided)

    Returns:
        dict: The updated order
    """
    if order_id not in orders:
        raise ValueError(f"Order {order_id} does not exist")

    order = orders[order_id]

    # Update products if provided
    if new_products:
        if not products_catalog:
            raise ValueError("Products catalog is required to update products")

        # Validate products exist
        for product_id in new_products:
            if product_id not in products_catalog:
                raise ValueError(f"Product {product_id} does not exist")

        # Update or add new products
        products_copy = copy.deepcopy(order['products'])
        for product_id, quantity in new_products.items():
            if product_id in products_copy:
                products_copy[product_id] += quantity
            else:
                products_copy[product_id] = quantity

        # Recalculate total
        total = sum(products_catalog[product_id].price * quantity
                    for product_id, quantity in products_copy.items())

        order['products'] = products_copy
        order['total'] = total

    # Update status if provided
    if new_status:
        valid_statuses = {'new', 'processing', 'shipped', 'delivered', 'cancelled'}
        if new_status not in valid_statuses:
            raise ValueError(f"Invalid status: {new_status}")
        order['status'] = new_status

    return order

def cancel_order(orders, order_id):
    """
    Cancel an order

    Args:
        orders (dict): The order tracking system
        order_id (int): The order to cancel

    Returns:
        dict: The cancelled order
    """
    if order_id not in orders:
        raise ValueError(f"Order {order_id} does not exist")

    order = orders[order_id]
    if order['status'] == 'delivered':
        raise ValueError("Cannot cancel delivered order")

    order['status'] = 'cancelled'
    return order

def get_customer_order_history(orders, customer_id):
    """
    Get order history for a customer

    Args:
        orders (dict): The order tracking system
        customer_id (int): Customer ID to search for

    Returns:
        list: List of orders for the customer
    """
    customer_orders = []
    for order_id, order in orders.items():
        if order['customer'].id == customer_id:
            # Create a copy of the order and add the order_id
            order_copy = copy.deepcopy(order)
            order_copy['order_id'] = order_id
            customer_orders.append(order_copy)

    return customer_orders

def get_product_popularity(orders):
    """
    Get total quantity ordered for each product

    Args:
        orders (dict): The order tracking system

    Returns:
        dict: Dictionary mapping product_id to total quantity ordered
    """
    popularity = {}

    for order in orders.values():
        # Only count non-cancelled orders
        if order['status'] != 'cancelled':
            for product_id, quantity in order['products'].items():
                if product_id in popularity:
                    popularity[product_id] += quantity
                else:
                    popularity[product_id] = quantity

    return popularity

def serialize_orders(orders, filename):
    """
    Serialize orders to JSON file

    Args:
        orders (dict): The order tracking system
        filename (str): File to write to
    """
    # Need to convert namedtuples to dictionaries
    serializable_orders = {}

    for order_id, order in orders.items():
        serializable_order = {
            'customer': order['customer']._asdict(),
            'products': order['products'],
            'status': order['status'],
            'total': order['total']
        }
        serializable_orders[order_id] = serializable_order

    with open(filename, 'w') as f:
        json.dump(serializable_orders, f, indent=2)

def deserialize_orders(filename):
    """
    Deserialize orders from JSON file

    Args:
        filename (str): File to read from

    Returns:
        dict: The order tracking system
    """
    with open(filename, 'r') as f:
        serializable_orders = json.load(f)

    orders = {}
    for order_id, order in serializable_orders.items():
        # Convert string keys back to integers
        order_id = int(order_id)

        # Recreate Customer namedtuple
        customer_dict = order['customer']
        customer = Customer(
            id=customer_dict['id'],
            name=customer_dict['name'],
            address=customer_dict['address']
        )

        # Recreate order
        orders[order_id] = {
            'customer': customer,
            'products': order['products'],
            'status': order['status'],
            'total': order['total']
        }

    return orders

def super_challenge_1():
    # Create product catalog
    products = {
        1: Product(1, "Laptop", 1200.00, "Electronics"),
        2: Product(2, "Smartphone", 800.00, "Electronics"),
        3: Product(3, "Headphones", 150.00, "Electronics"),
        4: Product(4, "Desk Chair", 250.00, "Furniture"),
        5: Product(5, "Coffee Table", 300.00, "Furniture"),
        6: Product(6, "Bookshelf", 180.00, "Furniture"),
        7: Product(7, "Running Shoes", 120.00, "Apparel"),
        8: Product(8, "Winter Jacket", 200.00, "Apparel"),
        9: Product(9, "Blender", 80.00, "Kitchen"),
        10: Product(10, "Toaster", 40.00, "Kitchen")
    }

    # Create customers
    customers = {
        1: Customer(1, "Alice Johnson", "123 Main St"),
        2: Customer(2, "Bob Smith", "456 Elm St"),
        3: Customer(3, "Charlie Brown", "789 Oak St"),
        4: Customer(4, "Diana Prince", "101 Pine St"),
        5: Customer(5, "Edward Stark", "202 Maple St")
    }

    # Initialize order system
    orders = {}

    # Add orders
    try:
        # Order 1: Alice buys electronics
        add_order(orders, customers[1], {1: 1, 3: 1}, products)

        # Order 2: Bob buys furniture
        add_order(orders, customers[2], {4: 1, 5: 1}, products)

        # Order 3: Charlie buys a mix
        add_order(orders, customers[3], {2: 1, 7: 1, 9: 2}, products)

        # Order 4: Diana buys kitchen items
        add_order(orders, customers[4], {9: 1, 10: 2}, products)

        # Order 5: Edward buys electronics
        add_order(orders, customers[5], {1: 1, 2: 1}, products)

        # More orders with repeat customers
        add_order(orders, customers[1], {2: 1, 8: 1}, products)
        add_order(orders, customers[2], {6: 2}, products)
        add_order(orders, customers[3], {3: 1, 10: 1}, products)
        add_order(orders, customers[4], {5: 1, 6: 1}, products)
        add_order(orders, customers[5], {7: 2, 8: 1}, products)

        # Even more orders
        add_order(orders, customers[1], {9: 1, 10: 1}, products)
        add_order(orders, customers[2], {1: 1}, products)
        add_order(orders, customers[3], {4: 1, 5: 1}, products)
        add_order(orders, customers[4], {2: 1, 3: 1}, products)
        add_order(orders, customers[5], {9: 3}, products)

        # Update some orders
        update_order(orders, 3, {3: 1}, "processing", products)
        update_order(orders, 7, new_status="shipped")
        update_order(orders, 12, new_status="delivered")

        # Cancel an order
        cancel_order(orders, 5)

        # Try to cancel a delivered order (should fail)
        try:
            cancel_order(orders, 12)
            print("ERROR: Should not be able to cancel delivered order")
        except ValueError as e:
            print(f"Correctly caught exception: {e}")

        # Get customer history
        alice_orders = get_customer_order_history(orders, 1)
        print(f"Alice has {len(alice_orders)} orders")

        # Get product popularity
        popularity = get_product_popularity(orders)
        most_popular = max(popularity.items(), key=lambda x: x[1])
        print(f"Most popular product: {products[most_popular[0]].name} with {most_popular[1]} units ordered")

        # Serialize orders
        serialize_orders(orders, "orders.json")

        # Deserialize orders
        loaded_orders = deserialize_orders("orders.json")
        print(f"Successfully loaded {len(loaded_orders)} orders")

    except Exception as e:
        print(f"Error: {e}")


if __name__ == "__main__":
    super_challenge_1()
