# -- super_challenge_2_solution -- #
import os
from collections import namedtuple, Counter
import re
import json
import copy
from typing import Dict, List, Set, Tuple

# Define namedtuples
DocumentMetadata = namedtuple('DocumentMetadata',
                             ['filename', 'word_count', 'sentence_count', 'unique_words'])

WordInfo = namedtuple('WordInfo',
                     ['word', 'frequency', 'documents', 'contexts'])

# Define a Context namedtuple to store context information
Context = namedtuple('Context',
                    ['document', 'sentence', 'position', 'context_text'])

def read_files(directory):
    """
    Read all text files from a directory

    Args:
        directory (str): Directory to read from

    Returns:
        dict: Dictionary mapping filename to file content
    """
    documents = {}

    for filename in os.listdir(directory):
        if filename.endswith('.txt'):
            file_path = os.path.join(directory, filename)
            with open(file_path, 'r', encoding='utf-8') as file:
                documents[filename] = file.read()

    return documents

def extract_sentences(text):
    """
    Extract sentences from text

    Args:
        text (str): Text to process

    Returns:
        list: List of sentences
    """
    # Simple sentence splitting (this could be improved with NLP libraries)
    sentences = re.split(r'(?<=[.!?])\s+', text)
    return [s.strip() for s in sentences if s.strip()]

def extract_words(text):
    """
    Extract words from text

    Args:
        text (str): Text to process

    Returns:
        list: List of words (lowercase, no punctuation)
    """
    # Convert to lowercase and remove punctuation
    words = re.findall(r'\b\w+\b', text.lower())
    return words

def process_document(filename, content):
    """
    Process a single document

    Args:
        filename (str): Name of the file
        content (str): Content of the file

    Returns:
        tuple: (DocumentMetadata, word_frequencies, sentences, word_contexts)
    """
    # Extract sentences
    sentences = extract_sentences(content)

    # Extract words from each sentence to track context
    word_contexts = {}
    all_words = []

    for i, sentence in enumerate(sentences):
        words = extract_words(sentence)
        all_words.extend(words)

        # Track word positions and contexts
        for position, word in enumerate(words):
            # Get context (5 words before and after)
            start = max(0, position - 5)
            end = min(len(words), position + 6)
            context_text = ' '.join(words[start:end])

            if word not in word_contexts:
                word_contexts[word] = []

            context = Context(
                document=filename,
                sentence=i,
                position=position,
                context_text=context_text
            )

            word_contexts[word].append(context)

    # Count word frequencies
    word_frequencies = Counter(all_words)

    # Create document metadata
    metadata = DocumentMetadata(
        filename=filename,
        word_count=len(all_words),
        sentence_count=len(sentences),
        unique_words=len(word_frequencies)
    )

    return metadata, word_frequencies, sentences, word_contexts

def build_master_data_structure(documents):
    """
    Build master data structure from processed documents

    Args:
        documents (dict): Dictionary mapping filename to content

    Returns:
        tuple: (document_metadata, word_info, all_words_set, unique_words_per_doc)
    """
    document_metadata = {}
    word_frequencies = {}
    word_contexts = {}
    unique_words_per_doc = {}
    all_words_set = set()

    # Process each document
    for filename, content in documents.items():
        metadata, frequencies, sentences, contexts = process_document(filename, content)

        # Store metadata
        document_metadata[filename] = metadata

        # Store unique words for this document
        words_in_doc = set(frequencies.keys())
        unique_words_per_doc[filename] = words_in_doc

        # Update global word set
        all_words_set.update(words_in_doc)

        # Update word frequencies and contexts
        for word, count in frequencies.items():
            if word not in word_frequencies:
                word_frequencies[word] = Counter()
            word_frequencies[word][filename] = count

        # Update word contexts
        for word, word_context_list in contexts.items():
            if word not in word_contexts:
                word_contexts[word] = []
            word_contexts[word].extend(word_context_list)

    # Build WordInfo objects
    word_info = {}
    for word, freq_by_doc in word_frequencies.items():
        total_freq = sum(freq_by_doc.values())
        documents_with_word = list(freq_by_doc.keys())

        word_info[word] = WordInfo(
            word=word,
            frequency=total_freq,
            documents=documents_with_word,
            contexts=word_contexts.get(word, [])
        )

    return document_metadata, word_info, all_words_set, unique_words_per_doc

def get_top_words(word_info, n=10):
    """
    Get top N words across all documents

    Args:
        word_info (dict): WordInfo dictionary
        n (int): Number of top words to retrieve

    Returns:
        list: Top N words with their frequencies
    """
    # Sort words by frequency
    sorted_words = sorted(
        word_info.values(),
        key=lambda x: x.frequency,
        reverse=True
    )

    return sorted_words[:n]

def get_top_words_per_document(document_metadata, word_info, n=5):
    """
    Get top N words for each document

    Args:
        document_metadata (dict): Document metadata
        word_info (dict): WordInfo dictionary
        n (int): Number of top words to retrieve per document

    Returns:
        dict: Dictionary mapping filename to top N words
    """
    top_words = {}

    for filename in document_metadata:
        # Get words that appear in this document
        words_in_doc = [
            w for w in word_info.values()
            if filename in w.documents
        ]

        # Sort by frequency in this specific document
        sorted_words = sorted(
            words_in_doc,
            key=lambda x: x.frequency,
            reverse=True
        )

        top_words[filename] = sorted_words[:n]

    return top_words

def compute_document_similarity(doc1, doc2, unique_words_per_doc):
    """
    Compute similarity between two documents using Jaccard similarity

    Args:
        doc1 (str): First document filename
        doc2 (str): Second document filename
        unique_words_per_doc (dict): Dictionary mapping filename to set of unique words

    Returns:
        float: Similarity score (0-1)
    """
    words1 = unique_words_per_doc[doc1]
    words2 = unique_words_per_doc[doc2]

    # Calculate Jaccard similarity
    intersection = len(words1.intersection(words2))
    union = len(words1.union(words2))

    return intersection / union if union > 0 else 0

def get_concordance(word, word_info, max_contexts=10):
    """
    Get concordance (word contexts) for a given word

    Args:
        word (str): Word to look for
        word_info (dict): WordInfo dictionary
        max_contexts (int): Maximum number of contexts to return

    Returns:
        list: List of context information
    """
    if word not in word_info:
        return []

    contexts = word_info[word].contexts
    return contexts[:max_contexts]

def generate_report(document_metadata, word_info, all_words_set, unique_words_per_doc):
    """
    Generate analysis report

    Args:
        document_metadata (dict): Document metadata
        word_info (dict): WordInfo dictionary
        all_words_set (set): Set of all words across documents
        unique_words_per_doc (dict): Dictionary mapping filename to set of unique words

    Returns:
        dict: Report data
    """
    report = {
        "document_count": len(document_metadata),
        "total_words": sum(m.word_count for m in document_metadata.values()),
        "unique_words_count": len(all_words_set),
        "document_summaries": {},
        "top_words_overall": [
            {"word": w.word, "frequency": w.frequency, "document_count": len(w.documents)}
            for w in get_top_words(word_info, 20)
        ],
        "document_similarities": [],
        "words_in_all_documents": [],
        "words_in_only_one_document": []
    }

    # Add document summaries
    for filename, metadata in document_metadata.items():
        report["document_summaries"][filename] = {
            "word_count": metadata.word_count,
            "sentence_count": metadata.sentence_count,
            "unique_words": metadata.unique_words,
            "top_words": [
                {"word": w.word, "frequency": w.frequency}
                for w in get_top_words_per_document(
                    {filename: metadata}, word_info, 10
                )[filename]
            ]
        }

    # Add document similarities
    filenames = list(document_metadata.keys())
    for i, doc1 in enumerate(filenames):
        for doc2 in filenames[i+1:]:
            similarity = compute_document_similarity(doc1, doc2, unique_words_per_doc)
            report["document_similarities"].append({
                "document1": doc1,
                "document2": doc2,
                "similarity": similarity
            })

    # Find words that appear in all documents
    all_docs = set(document_metadata.keys())
    words_in_all_docs = [
        w.word for w in word_info.values()
        if set(w.documents) == all_docs
    ]
    report["words_in_all_documents"] = words_in_all_docs[:20]  # Limit to 20

    # Find words that appear in only one document
    words_in_one_doc = [
        {"word": w.word, "document": w.documents[0], "frequency": w.frequency}
        for w in word_info.values()
        if len(w.documents) == 1
    ]
    report["words_in_only_one_document"] = sorted(
        words_in_one_doc,
        key=lambda x: x["frequency"],
        reverse=True
    )[:20]  # Limit to 20

    return report

def serialize_data(data, filename):
    """Serialize data to JSON file"""
    serializable_data = {
        "document_metadata": {
            k: v._asdict() for k, v in data["document_metadata"].items()
        },
        "report": data["report"]
    }

    with open(filename, 'w', encoding='utf-8') as f:
        json.dump(serializable_data, f, indent=2)

def super_challenge_2():
    # Define directory with text files
    directory = "text_files"

    # For this example, we'll create sample text files if they don't exist
    os.makedirs(directory, exist_ok=True)

    # Create sample files if they don't exist
    sample_texts = {
        "article1.txt": "Python is a programming language. It is known for its simplicity and readability. Many developers use Python for data analysis, web development, and artificial intelligence.",
        "article2.txt": "Programming languages are tools for developers. Python is one such language. It has a large standard library and an active community. Developers love Python for its ease of use.",
        "article3.txt": "Data analysis is an important field. Python offers many libraries for data analysis. Libraries like Pandas and NumPy are widely used. Python makes data analysis accessible to many professionals."
    }

    for filename, content in sample_texts.items():
        file_path = os.path.join(directory, filename)
        if not os.path.exists(file_path):
            with open(file_path, 'w', encoding='utf-8') as f:
                f.write(content)

    # Process the text files
    documents = read_files(directory)

    # Build master data structure
    document_metadata, word_info, all_words_set, unique_words_per_doc = build_master_data_structure(documents)

    # Generate report
    report = generate_report(document_metadata, word_info, all_words_set, unique_words_per_doc)

    # Serialize data
    serialize_data({
        "document_metadata": document_metadata,
        "report": report
    }, "text_analysis_report.json")

    # Print some key findings
    print(f"Processed {len(documents)} documents")
    print(f"Found {len(all_words_set)} unique words")

    # Show top words
    top_words = get_top_words(word_info, 5)
    print("\nTop 5 words across all documents:")
    for word_info in top_words:
        print(f"- {word_info.word}: {word_info.frequency} occurrences in {len(word_info.documents)} documents")

    # Show sample concordance
    if "python" in word_info:
        print("\nConcordance for 'python':")
        for context in get_concordance("python", word_info, 3):
            print(f"- {context.context_text}")

    print("\nFull report saved to 'text_analysis_report.json'")