import logging
import time
import random

# Configure logging for Challenge 6
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)

#  Challenge 1: Safe Division
def safe_divide(a, b):
    try:
        # Check if inputs are numbers
        if not (isinstance(a, (int, float)) and isinstance(b, (int, float))):
            return "Error: Invalid input type"

        # Try to perform division
        return a / b
    except ZeroDivisionError:
        return "Error: Division by zero"


#  Challenge 2: File Reader
def read_file(filename):
    try:
        with open(filename, 'r') as file:
            return file.read()
    except FileNotFoundError:
        print(f"File not found: {filename}")
        return None
    except PermissionError:
        print(f"Permission denied for: {filename}")
        return None
    except Exception as e:
        print(f"Error reading file: {e}")
        return None


#  Challenge 3: Input Parser
def get_integer(prompt, max_attempts=3):
    attempts = 0
    while attempts < max_attempts:
        try:
            user_input = input(prompt)
            return int(user_input)
        except ValueError:
            attempts += 1
            remaining = max_attempts - attempts
            if remaining > 0:
                print(f"Invalid integer. Please try again. {remaining} attempts remaining.")
            else:
                print("Maximum attempts reached.")
    return None


#  Challenge 4: Custom Exception
class InvalidAgeError(ValueError):
    def __init__(self, age):
        self.age = age
        super().__init__(f"Invalid age: {age}. Age must be between 0 and 120.")

def validate_age(age):
    if not isinstance(age, int):
        raise TypeError("Age must be an integer")

    if age < 0 or age > 120:
        raise InvalidAgeError(age)

    return age


#  Challenge 5: Resource Manager
def write_lines(filename, lines):
    file = None
    try:
        file = open(filename, 'w')
        for line in lines:
            file.write(line + '\n')
        return True
    except Exception as e:
        print(f"Error writing to file: {e}")
        return False
    finally:
        # Ensure the file is closed even if an exception occurred
        if file is not None and not file.closed:
            file.close()


#  Challenge 6: Error Logging
def logged_divide(a, b):
    try:
        return a / b
    except TypeError as e:
        logging.warning(f"Type error in division: {e}")
        return None
    except ZeroDivisionError as e:
        logging.error(f"Division by zero attempted: {e}")
        return None


#  Challenge 7: Retry Mechanism
def retry(operation, max_attempts=3, delay=1):
    attempts = 0
    last_exception = None

    while attempts < max_attempts:
        try:
            return operation()
        except Exception as e:
            attempts += 1
            last_exception = e
            print(f"Attempt {attempts} failed: {e}")

            if attempts < max_attempts:
                print(f"Retrying in {delay} seconds...")
                time.sleep(delay)

    # If we get here, all attempts failed
    print(f"All {max_attempts} attempts failed.")
    return None


#  Challenge 8: Context Manager
class Timer:
    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = time.time()
        elapsed = self.end_time - self.start_time
        print(f"Execution time: {elapsed:.4f} seconds")

        if exc_type is not None:
            print(f"An error occurred: {exc_val}")
            # Return False to re-raise the exception
            return False
        return True


#  Challenge 9: Multiple Exception Handling
def process_value(data, key):
    try:
        # Try to get the value for the key
        value = data[key]

        # Try to convert to integer
        int_value = int(value)

        # Try to divide 100 by it
        result = 100 / int_value

        return (True, result)

    except KeyError:
        return (False, f"Key not found: {key}")

    except ValueError:
        return (False, f"Cannot convert '{data[key]}' to integer")

    except ZeroDivisionError:
        return (False, "Cannot divide by zero")


#  Challenge 10: Database Transaction Simulator
class Transaction:
    def __init__(self):
        self.db = {}
        self.backup = {}
        self.in_transaction = False

    def start(self):
        self.in_transaction = True
        self.backup = self.db.copy()
        print("Transaction started")

    def execute_query(self, query):
        if not self.in_transaction:
            raise ValueError("Transaction not started")

        # Simulate random failure (30% chance)
        if random.random() < 0.3:
            raise RuntimeError(f"Query failed: {query}")

        # Simulate query execution (just store the query as a key)
        self.db[query] = f"Result of {query}"
        print(f"Query executed: {query}")
        return True

    def commit(self):
        if not self.in_transaction:
            raise ValueError("Transaction not started")

        self.in_transaction = False
        self.backup = {}
        print("Transaction committed")
        return True

    def rollback(self):
        if not self.in_transaction:
            raise ValueError("Transaction not started")

        self.db = self.backup.copy()
        self.in_transaction = False
        self.backup = {}
        print("Transaction rolled back")

def run_transaction(queries):
    transaction = Transaction()

    try:
        transaction.start()

        for query in queries:
            transaction.execute_query(query)

        # If we got here, all queries were successful
        transaction.commit()
        print("Transaction successful")
        return transaction
    except Exception as e:
        print(f"Transaction failed: {e}")
        if transaction.in_transaction:
            transaction.rollback()
        return None


# Testing code
if __name__ == "__main__":
    # Challenge 1 Test
    print("\nChallenge 1 Tests:")
    print(f"safe_divide(10, 2) = {safe_divide(10, 2)}")
    print(f"safe_divide(10, 0) = {safe_divide(10, 0)}")
    print(f"safe_divide('10', 2) = {safe_divide('10', 2)}")

    # Challenge 2 Test
    print("\nChallenge 2 Tests:")
    print(f"read_file('nonexistent.txt') = {read_file('nonexistent.txt')}")

    # Challenge 3 Test
    print("\nChallenge 3 Tests:")
    print("Uncomment to test interactively:")
    # value = get_integer("Enter a number: ")
    # print(f"Valid input: {value}")

    # Challenge 4 Test
    print("\nChallenge 4 Tests:")
    try:
        print(f"validate_age(25) = {validate_age(25)}")
        try:
            print(f"validate_age(-5) = {validate_age(-5)}")
        except InvalidAgeError as e:
            print(f"Caught InvalidAgeError: {e}")
    except Exception as e:
        print(f"Caught unexpected exception: {e}")

    try:
        print(f"validate_age('25') = {validate_age('25')}")
    except TypeError as e:
        print(f"Caught TypeError: {e}")

    # Challenge 5 Test
    print("\nChallenge 5 Tests:")
    print(f"write_lines('test.txt', ['Line 1', 'Line 2']) = {write_lines('test.txt', ['Line 1', 'Line 2'])}")

    # Challenge 6 Test
    print("\nChallenge 6 Tests:")
    print(f"logged_divide(10, 2) = {logged_divide(10, 2)}")
    print(f"logged_divide(10, 0) = {logged_divide(10, 0)}")
    print(f"logged_divide('10', 2) = {logged_divide('10', 2)}")

    # Challenge 7 Test
    print("\nChallenge 7 Tests:")
    def test_function():
        static_var = getattr(test_function, 'called', 0)
        test_function.called = static_var + 1
        if test_function.called < 3:
            raise ValueError("Not ready yet")
        return "Success!"

    print(f"retry(test_function) = {retry(test_function)}")

    # Reset the static variable for another test
    test_function.called = 0

    # Challenge 8 Test
    print("\nChallenge 8 Tests:")
    try:
        with Timer() as timer:
            print("Starting timed operation...")
            time.sleep(1)
            print("Operation complete")
    except Exception:
        print("Exception propagated as expected")

    try:
        with Timer() as timer:
            print("Starting timed operation that will fail...")
            time.sleep(0.5)
            raise ValueError("Deliberate failure")
    except Exception as e:
        print(f"Exception propagated as expected: {e}")

    # Challenge 9 Test
    print("\nChallenge 9 Tests:")
    print(f"process_value({{'a': '10'}}, 'a') = {process_value({'a': '10'}, 'a')}")
    print(f"process_value({{'a': '10'}}, 'b') = {process_value({'a': '10'}, 'b')}")
    print(f"process_value({{'a': 'abc'}}, 'a') = {process_value({'a': 'abc'}, 'a')}")
    print(f"process_value({{'a': '0'}}, 'a') = {process_value({'a': '0'}, 'a')}")

    # Challenge 10 Test
    print("\nChallenge 10 Tests:")
    transaction1 = run_transaction(["SELECT * FROM users", "UPDATE users SET name='John'"])
    print("\nAnother transaction example:")
    transaction2 = run_transaction(["INSERT INTO products", "UPDATE inventory", "DELETE FROM old_records"])