"""
File System Operations Solutions

This file contains solutions to the challenges in challenges.py.
Each function is implemented according to its specifications.
"""

import os
import shutil
import hashlib
import datetime
import zipfile
from pathlib import Path


# Challenge 1: Basic Path Operations with os.path
def get_file_info(file_path):
    """
    Use os.path to get information about a file.
    """
    info = {
        'exists': os.path.exists(file_path),
        'is_file': os.path.isfile(file_path),
        'is_dir': os.path.isdir(file_path),
        'basename': os.path.basename(file_path),
        'dirname': os.path.dirname(file_path),
        'extension': os.path.splitext(file_path)[1],
    }

    # Get size (0 if file doesn't exist)
    info['size'] = os.path.getsize(file_path) if info['exists'] else 0

    return info


# Challenge 2: Path Operations with pathlib
def create_unique_file(directory, base_name, extension):
    """
    Create a unique filename in the specified directory.
    """
    # Make sure extension doesn't have a leading dot
    if extension.startswith('.'):
        extension = extension[1:]

    # Create the directory if it doesn't exist
    dir_path = Path(directory)
    dir_path.mkdir(parents=True, exist_ok=True)

    # Try with the base name first
    file_path = dir_path / f"{base_name}.{extension}"
    counter = 1

    # If file exists, append numbers until we find a unique name
    while file_path.exists():
        file_path = dir_path / f"{base_name}_{counter}.{extension}"
        counter += 1

    # Create an empty file to reserve the name
    file_path.touch()

    return file_path


# Challenge 3: Context Managers
def safe_write_with_backup(file_path, content):
    """
    Write content to a file, but create a backup of the file first if it exists.
    """
    backup_path = f"{file_path}.bak"

    try:
        # If the file exists, create a backup
        if os.path.exists(file_path):
            with open(file_path, 'r') as source_file:
                with open(backup_path, 'w') as backup_file:
                    backup_file.write(source_file.read())

        # Write the new content
        with open(file_path, 'w') as dest_file:
            dest_file.write(content)

        return True, backup_path if os.path.exists(backup_path) else "No backup needed (file didn't exist)"

    except Exception as e:
        return False, f"Error: {str(e)}"


# Challenge 4: File Reading and Writing
def csv_to_dict(csv_path, delimiter=','):
    """
    Read a simple CSV file and convert it to a list of dictionaries.
    """
    result = []

    with open(csv_path, 'r') as f:
        # Read all lines and strip whitespace
        lines = [line.strip() for line in f.readlines()]

        if not lines:
            return result

        # First line contains headers
        headers = lines[0].split(delimiter)

        # Process each data row
        for line in lines[1:]:
            if not line:  # Skip empty lines
                continue

            values = line.split(delimiter)
            row_dict = {}

            # Map values to headers
            for i, header in enumerate(headers):
                if i < len(values):
                    row_dict[header] = values[i]
                else:
                    row_dict[header] = ""  # Add empty value for missing fields

            result.append(row_dict)

    return result


# Challenge 5: Binary Files
def compare_binary_files(file_path1, file_path2):
    """
    Compare two binary files and return information about differences.
    """
    result = {
        'files_match': False,
        'size_diff': 0,
        'first_diff_pos': -1,
        'first_diff_bytes': None
    }

    try:
        with open(file_path1, 'rb') as f1, open(file_path2, 'rb') as f2:
            # Read both files completely
            data1 = f1.read()
            data2 = f2.read()

            # Calculate size difference
            result['size_diff'] = len(data1) - len(data2)

            # Check if files are identical
            if data1 == data2:
                result['files_match'] = True
                return result

            # Find position of first difference
            min_len = min(len(data1), len(data2))
            for i in range(min_len):
                if data1[i] != data2[i]:
                    result['first_diff_pos'] = i
                    result['first_diff_bytes'] = (data1[i], data2[i])
                    break

            # If we haven't found a difference yet but sizes differ,
            # the first difference is at the end of the shorter file
            if result['first_diff_pos'] == -1 and result['size_diff'] != 0:
                result['first_diff_pos'] = min_len
                if len(data1) > len(data2):
                    result['first_diff_bytes'] = (data1[min_len], None)
                else:
                    result['first_diff_bytes'] = (None, data2[min_len])

    except Exception:
        # If an error occurred, files don't match
        result['files_match'] = False

    return result


# Challenge 6: Directory Operations
def organize_files_by_extension(source_dir):
    """
    Organize files in the source directory by moving them into subdirectories
    based on their extensions.
    """
    extension_counts = {}

    # Ensure source directory exists
    if not os.path.exists(source_dir) or not os.path.isdir(source_dir):
        return extension_counts

    # Process each file in the source directory
    for filename in os.listdir(source_dir):
        file_path = os.path.join(source_dir, filename)

        # Skip directories
        if os.path.isdir(file_path):
            continue

        # Get file extension
        _, ext = os.path.splitext(filename)
        if not ext:
            ext = '.no_extension'  # Handle files without extension

        # Clean extension for directory name (remove the dot)
        ext_dir_name = ext[1:] if ext.startswith('.') else ext

        # Create subdirectory for this extension
        ext_dir = os.path.join(source_dir, ext_dir_name)
        os.makedirs(ext_dir, exist_ok=True)

        # Move the file
        destination = os.path.join(ext_dir, filename)
        shutil.move(file_path, destination)

        # Update our counts
        if ext in extension_counts:
            extension_counts[ext] += 1
        else:
            extension_counts[ext] = 1

    return extension_counts


# Challenge 7: File System Navigation and Multiple Directory Creation
def create_project_structure(base_dir, project_name):
    """
    Create a project directory structure.
    """
    # Create the project directory
    project_dir = Path(base_dir) / project_name

    # Define directory structure with file contents
    structure = {
        'src': {
            '__init__.py': '',  # Empty file
            'main.py': f"print('Hello from {project_name}')"
        },
        'tests': {
            '__init__.py': '',  # Empty file
            'test_main.py': f"# Tests for main.py"
        },
        'docs': {
            'README.md': f"# {project_name}\n\nProject documentation"
        }
    }

    # Create directories and files
    for dir_name, files in structure.items():
        # Create directory
        dir_path = project_dir / dir_name
        dir_path.mkdir(parents=True, exist_ok=True)

        # Create files in this directory
        for file_name, content in files.items():
            file_path = dir_path / file_name
            with open(file_path, 'w') as f:
                f.write(content)

    return str(project_dir)


# Challenge 9: Recursive Directory Walking
def find_duplicate_files(directory):
    """
    Find duplicate files within a directory tree based on file content.
    """
    # Dictionary to store files by content hash
    files_by_hash = {}

    # Helper function to calculate file hash
    def get_file_hash(filepath):
        hasher = hashlib.md5()
        with open(filepath, 'rb') as f:
            # Read file in chunks to handle large files
            chunk = f.read(65536)  # 64k chunks
            while chunk:
                hasher.update(chunk)
                chunk = f.read(65536)
        return hasher.hexdigest()

    # Walk through all files in the directory tree
    for root, _, files in os.walk(directory):
        for filename in files:
            file_path = os.path.join(root, filename)

            try:
                # Calculate hash for the file
                file_hash = get_file_hash(file_path)

                # Add to our dictionary
                if file_hash in files_by_hash:
                    files_by_hash[file_hash].append(file_path)
                else:
                    files_by_hash[file_hash] = [file_path]
            except (IOError, OSError):
                # Skip files we can't read
                continue

    # Filter to only include duplicates (2+ files with same hash)
    return {hash_val: paths for hash_val, paths in files_by_hash.items() if len(paths) >= 2}


# Challenge 10: Comprehensive Challenge
def backup_and_compress(source_dir, backup_dir, exclude_extensions=None):
    """
    Create a backup of a directory structure, excluding specified extensions,
    and compress the backup into a zip file with a timestamp.
    """
    if exclude_extensions is None:
        exclude_extensions = []

    # Make sure extensions start with a dot
    exclude_extensions = [ext if ext.startswith('.') else f'.{ext}' for ext in exclude_extensions]

    # Create backup directory if it doesn't exist
    os.makedirs(backup_dir, exist_ok=True)

    # Generate timestamp for the zip filename
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    zip_filename = f"backup_{timestamp}.zip"
    zip_path = os.path.join(backup_dir, zip_filename)

    # Create a temporary directory for collecting files
    temp_dir = os.path.join(backup_dir, f"temp_backup_{timestamp}")
    os.makedirs(temp_dir, exist_ok=True)

    try:
        # Copy files to temp directory, preserving structure
        for root, dirs, files in os.walk(source_dir):
            # Get relative path from source_dir
            rel_path = os.path.relpath(root, source_dir)

            # Create corresponding directory in temp_dir
            if rel_path != '.':
                dest_dir = os.path.join(temp_dir, rel_path)
                os.makedirs(dest_dir, exist_ok=True)
            else:
                dest_dir = temp_dir

            # Copy each file (except excluded extensions)
            for file in files:
                _, ext = os.path.splitext(file)
                if ext.lower() not in exclude_extensions:
                    src_file = os.path.join(root, file)
                    dst_file = os.path.join(dest_dir, file)
                    shutil.copy2(src_file, dst_file)

        # Create zip archive
        with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for root, _, files in os.walk(temp_dir):
                for file in files:
                    file_path = os.path.join(root, file)
                    # Add file to zip with path relative to temp_dir
                    zipf.write(file_path, os.path.relpath(file_path, temp_dir))

    finally:
        # Clean up temporary directory
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)

    return zip_path


if __name__ == "__main__":
    # Test code for the solutions

    # Challenge 1
    print("\n--- Challenge 1: Basic Path Operations ---")
    file_info = get_file_info(__file__)
    print(f"File info: {file_info}")

    # Challenge 2
    print("\n--- Challenge 2: Path Operations with pathlib ---")
    unique_file = create_unique_file(".", "test", "txt")
    print(f"Created unique file: {unique_file}")

    # Challenge 3
    print("\n--- Challenge 3: Context Managers ---")
    success, result = safe_write_with_backup("test_backup.txt", "This is a test file with backup.")
    print(f"Write result: {success}, Backup: {result}")

    # Add more tests as needed