# Programming Fundamentals

## Objectives

- What is Programming?
- Sequential Instruction
- Programming Languages
- Compiled vs Interpreted Languages
- Code and Data
- Basic Syntax Concepts

## Introductions

- Name
- Brief Background
- Prior Developer Experience

## What is Programming?

Programming is the process of giving computers a set of instructions to follow.
Think of it like writing a detailed recipe - each step must be precise and in the correct order for the desired outcome.
These instructions, called code, tell the computer exactly what to do with data.

Python is a high level, user friendly, object oriented, interpreted language, and is one of the most popular languages in the world.

Here's a simple example in Python that demonstrates basic programming concepts:

```python
# This program greets a user and calculates their age in dog years
name = input("What is your name? ")
age = int(input("How old are you? "))
dog_years = age * 7

print(f"Hello {name}! You are {age} years old.")
print(f"In dog years, you would be {dog_years} years old!")
```

!!! Note "Even though we haven't discussed a single Python concept yet, notice how "readable" this code is. This is one of the reasons Python is so popular."

## Sequential Instruction

Programs execute instructions sequentially, one after another, from top to bottom. This ordered execution is fundamental to how computers process our code.

```python
# Sequential execution example
print("Step 1: Starting the program")
x = 5
print("Step 2: Set x to", x)
y = x + 3
print("Step 3: Set y to x + 3 =", y)
result = y * 2
print("Step 4: Final result is", result)
```

!!! Note "Note on Execution Order vs. Declaration Order"
    Realistic programs often **declare** functions and `import` modules at the top of the file.  This might give the impression that the code isn't executing sequentially (top to bottom).  It's crucial to distinguish between *declaration* and *execution*.

    *   **Declaration:**  Declarations (like function definitions using `def` or module imports using `import`) tell Python *about* the existence of functions, variables, or modules. They essentially prepare these elements for later use, but they don't *execute* the code *inside* the functions or import *all* the code from the modules at that point. Think of declarations as setting up the tools and materials you'll need.

    *   **Execution:**  Execution is the actual running of the code, following the sequential order.  Even though functions are *declared* at the top, the code *inside* those functions isn't executed until the function is *called* later in the program's flow.

    **Analogy:**  Imagine building a house.  You might gather all your materials (bricks, wood, tools) first – that's like *declaration*. You're making them available.  But the *actual building* of the house (laying the foundation, erecting walls, etc.) – that's *execution* – still happens in a specific, sequential order.  You can't put the roof on before you've built the walls.  Similarly, Python executes code line by line, but it *remembers* the declared functions and modules so it can use them when they are called.

    Import statements can be thought of in a few different ways:
    - When a module is first imported, all the code in the module will be executed, including any declarations and any top level executable statements.
    - Any top level executable statements will be executed, but any declarations will be loaded into memory for later use.
    - The `import` statement itself creates a reference to the module in the current namespace.

## Programming Languages

Programming languages are tools we use to communicate with computers. Each language has its own syntax and rules, just like human languages. Python is known for its readability and straightforward syntax.

Here are two functionally identical programs written in different languages:

```python
# Python - Clean and readable
def calculate_area(length, width):
    return length * width

# Calculate rectangle area
area = calculate_area(5, 3)
print(f"The area is {area} square units")
```

The same program in C++ requires more complex syntax:

```cpp
#include <iostream>
using namespace std;

// C++ - More exacting syntax
int calculateArea(int length, int width) {
    return length * width;
}

int main() {
    int area = calculateArea(5, 3);
    cout << "The area is " << area << " square units" << endl;
    return 0;
}
```

C++ is a great and powerful language, but clearly it carries a lot of syntactic overhead that can trip up the novice programmer.

## Compiled vs Interpreted Languages

Python is an interpreted language, meaning it executes code directly, line by line. Compiled languages, like C++, must be converted to machine code first.

### Interpreted (`Python`/`bash`/`javascript`)
- Code is executed directly
- Easier to debug
- Platform independent
- Generally slower than compiled languages

!!! Note "Is Python slow"
    The concept of "speed" in programming is nuanced. If we're rendering game graphics on a GPU, then no question is a memory efficient compiled language going to be faster. If we're talking about a quick script to interact with a web API, then the ease of use and prebuilt libraries in Python are going to be the way to go. Lastly, some libraries have been optimized for performance, and virtually no noticeable difference from writing it in a lower level language.

    You are encouraged to look into data science and machine learning libraries like `numpy` and `pandas` if you are interested in this.

### Compiled (`C`/`C++`/`Java`/`Rust`)
- Must be compiled before running
- Faster execution
- Platform specific
- More complex build process


??? "Compiled Python?"

    While 99% of the time you'll hear python being referred to as an interpreted language, all languages higher then assembly are "compiled" to some intermediate representation before execution.  The python code you write is compiled into an intermediate representation called bytecode, which is then executed by the python virtual machine.

    ```python
    import dis

    def hello():
        print("Hello, World!")

    dis.dis(hello)
    ```

## Code and Data

Programs work by manipulating data using code. Data can be numbers, text, or more complex information.

```python
# Different types of data in Python
text_data = "Hello, World!"  # String (text)
numeric_data = 42           # Integer (whole number)
decimal_data = 3.14        # Float (decimal number)
is_active = True          # Boolean (True/False)

# List (collection of data)
mixed_data = [text_data, numeric_data, decimal_data, is_active]

# Dictionary (key-value pairs)
user_data = {
    "name": "Alice",
    "age": 25,
    "scores": [95, 87, 92]
}

print(f"Text: {text_data}")
print(f"Number: {numeric_data}")
print(f"List: {mixed_data}")
print(f"User data: {user_data}")
```

!!! Note "Data is KING"
    Data is the lynchpin of programming. Depending on the data you're working with, will drastically change how to approach the problem. This may be the most important concept to understand in programming as making the wrong language or data structure choices can drastically effect the performance of your program.

    Look up "Data oriented design" for more information.

## Basic Syntax Concepts

Python has several basic syntax rules that make it unique:

### Indentation
Python uses indentation to define code blocks:

```python
def check_number(x): # Start of function block
    if x > 0: # Start of if block
        print("Positive")
    elif x < 0: # Start of elif block
        print("Negative")
    else: # Start of else block
        print("Zero")

# Since the next line is fully left-justified,
# it is not part of the function block
check_number(5) # positive
```

### Comments
Comments help explain code:

```python
# This is a single-line comment

"""
This is a multi-line comment
It can span several lines
Used for longer explanations
"""

def multiply(a, b):
    # Return the product of a and b
    return a * b
```

!!! Note "Comments and Style Choices"
    In future lessons, we will talk about standards and guides that not only cover comments, but many aspects of writing clean and readable code.

### Basic Operators
Python supports various operators for calculations:

```python
# Arithmetic operators
print(5 + 3)   # Addition
print(5 - 3)   # Subtraction
print(5 * 3)   # Multiplication
print(5 / 3)   # Division
print(5 // 3)  # Integer division
print(5 % 3)   # Modulus (remainder)
print(5 ** 3)  # Exponentiation

# Comparison operators
print(5 > 3)   # Greater than
print(5 < 3)   # Less than
print(5 == 3)  # Equal to
print(5 != 3)  # Not equal to
```

!!! Tip "All of the above concepts were just an introduction and will be covered in more detail in future lessons."








