# Python Ecosystem

## Objectives

- Python Ecosystem
    - Python Installation and Versions
    - Python Execution Methods
        - Command Line
        - Interactive Shell (REPL)
    - Python Documentation
        - `help()` and `dir()`
    - The Standard Library
    - PyPI and Third-Party Packages
    - Virtual Environments (venv)
    - Package Management (pip)
    - Terminology

!!! Note "Prerequisites"
    - Python 3.10(Should be installed on VTA ops vm)
    - VS Code reccomended (and installed on local workstation)

!!! Note "We reccomend using the local VS Code to ssh to the floating IP of your VTA workstation."
    The intent of the workstations is to provide a "playground" with full root permissions and internet that the local workstations can not provide.

## Python Installation and Versions

!!! Warning "Python Versions"
    Python comes in different versions, with Python 3.x being the current standard. Python 2 is no longer supported as of 2020.

To check your Python3 version:
`python3 --version`

??? Tip "Installing Python"
    Follow [this guide](https://realpython.com/installing-python/) to install Python on the operating system of your choice.

## Python Execution Methods

### Command Line Execution

You can run Python scripts directly from the command line:

```bash
# Run a script
python script.py

# Run with arguments
python script.py arg1 arg2

# Run with the -c flag for inline code
python -c "print('Hello, World!')"
```

### Interactive Shell (REPL)

The Python REPL (Read-Eval-Print Loop) is perfect for testing quick ideas:

```python
# Start the interactive shell by typing 'python' in your terminal
>>> x = 5
>>> y = 10
>>> x + y
15
>>> print(f"The sum is {x + y}")
The sum is 15
```
!!! Info "REPL"
    There's no right or wrong way to utilize the REPL and can defintely be great for some quick calculations or testing ideas.  Eventually you'll want to primarily get in the habit of either setting up a full project, or using a `temp.py` file instead of the REPL.

## Python Documentation

Python provides built-in tools for accessing documentation.

### Using help()

```python
# Get help on a specific function
>>> help(print)

# Get help on a module
>>> import random
>>> help(random)

# Get help on a specific method
>>> help(list.append)
```

### Using dir()

The dir() function shows available attributes and methods:

```python
# List all attributes of a string
>>> print(dir("hello"))

# List all attributes of a list
>>> print(dir([1, 2, 3]))

# List all built-in functions
>>> print(dir(__builtins__))
```

## The Standard Library

Python's standard library is extensive and comes with "batteries included":

```python
# Common standard library examples
import random
import datetime
import json
import os

# Generate random numbers
print(random.randint(1, 100)) # 1-100

# Work with dates
current_time = datetime.datetime.now()
print(current_time) # 2025-02-07 15:30:00

# Work with JSON
data = {"name": "Alice", "age": 30, "single": False}
json_string = json.dumps(data)
print(json_string) # {"name": "Alice", "age": 30, "single": false}

# Work with files and directories
print(os.getcwd())  # Current working directory
# /home/student/170d/cybbh/delta-students
```

## PyPI and Third-Party Packages

!!! Warning "The python covered in this course does not require anything more then the `standard library`."

    However, It would be doing a disservice to not at least introduce the wider python ecosystem as
    it's used in a professional industry setting.

PyPI (Python Package Index) hosts thousands of third-party packages. Here are some popular ones:

```python
# requests - HTTP library
import requests
response = requests.get('https://api.github.com')
print(response.status_code)

# pandas - Data analysis
import pandas as pd
df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]})
print(df)

# numpy - Numerical computing
import numpy as np
arr = np.array([1, 2, 3])
print(arr * 2)
```
### Types of Python Packages

Python packages generally fall into two main categories: **library packages** and **CLI tools**.

#### Library Packages
These are meant to be imported and used within Python code:

```python
# Library package example using requests
import requests

response = requests.get('https://api.github.com')
data = response.json()
```

#### CLI Tools
These install commands you can run from your terminal:

```bash
python3 -m pip install black pytest pylint

# black - A code formatter
black my_file.py

# pytest - A testing framework
pytest tests/

# pylint - A code linter
pylint my_file.py
```

Some packages provide both capabilities:

```python
# jupyter can be used as both:

# As a CLI tool:
jupyter notebook  # Launches notebook server

# As a library:
from jupyter_client import KernelManager
kernel = KernelManager()
```

!!! Tip "How to tell the difference"
    - Library packages: You'll use `import` statements to use them
    - CLI tools: After installation, new commands become available in your terminal
    - Check the package's PyPI page or documentation to see how it's meant to be used

!!! Note "To download a package from PyPI, we'll use `pip`, covered later in this lesson."

## Virtual Environments (venv)

Virtual environments isolate project dependencies:

```bash

cd ~/170d/cybbh
git clone https://git.cybbh.space/kevin.m.crotty/delta-students.git
cd delta-students

# Create a virtual environment
sudo apt install python3.10-venv
python3 -m venv --help # Note that it's a cli tool as well as a python module
python -m venv myenv

# Activate the environment
# On Windows:
myenv\Scripts\activate
# On Unix or MacOS:
source myenv/bin/activate

# Deactivate when done
deactivate
```

### Why Use Virtual Environments?

```python
# Project1 needs requests version 2.25.1
# requirements1.txt
requests==2.25.1

# Project2 needs requests version 2.28.1
# requirements2.txt
requests==2.28.1

# Without virtual environments, you can't have both!
```

## Package Management (pip)

[Antigravity!!!](https://imgs.xkcd.com/comics/python.png)

Pip is Python's package installer:

```bash
# Install a package
python3 -m pip install requests

# Install specific version
python3 -m pip install requests==2.25.1

# Install from requirements file
python3 -m pip install -r requirements.txt

# List installed packages
python3 -m pip list

# Generate requirements file
python3 -m pip freeze > requirements.txt
```

### Common pip Commands

```python
# requirements.txt example
requests==2.28.1
pandas>=1.4.0
numpy==1.23.0
python-dotenv>=0.19.0
```

### Best Practices

1. Always use virtual environments for projects (Outside of this course ;)
2. Keep requirements.txt updated
3. Specify version numbers for critical dependencies
4. Use pip freeze to capture exact versions
5. Consider using pip-tools for dependency management

??? Danger "The `uv` Package Manager (Optional)"

    ```bash
    # AFTER SSH'ING TO THE VTA WORKSTATION as your user

    # Install uv for your user
    curl -LsSf https://astral.sh/uv/install.sh | sh
    cd ~/170d/cybbh
    git clone https://git.cybbh.space/kevin.m.crotty/delta-students.git
    ```

    `uv` is one of the most modern and capable environments out there. The full capabilities of `uv` are beyond the scope of this course, but we'll cover some of the basics to get you on your way.

    `uv` is written in rust and is a drop in replacement for `pip`. It is extremely optomized for speed and performance.
    It uses the established standard of a `pyproject.toml` file to manage dependencies and the python version.  This allows you to save
    your environment to VCS, but has way more control logic then a regular `requirements.txt` file.

    !!! Tip "Take a look at the `pyproject.toml` file in the `delta-students` repo. It's a good example of how to use `uv` to manage dependencies."

    The main take away from this `pyproject.toml` file are:

    - Setting the python version
    - Setting the application specific dependencies (*not used in this course*)
    - Setting the development dependencies (*used in this course*)

    It's important to mentally seperate the concepts of application dependencies and development dependencies, not only in Python
    but in any language.

    !!! Note "We'll come back to `uv` a little later once we've travelled a more traditional path."

    !!! Success "PRO TIP:"
        The `uv python list` command is one of the best ways to see all the python versions installed on your system. If you use it for nothing else, this would be worth it.

## Terminology

| Term          | Definition |
| ---           | ---        |
| Literal       | A value that is written directly in the code. eg `42` or `"Hello, World!"` |
| Comments      | Used to explain the code. They are not executed by the interpreter. eg `# This is a comment` |
| Keywords      | Reserved words in Python that have special meanings and cannot be used as variable names. eg `def`, `if`, `else`, `while` |
| Objects       | An of a class(template). eg `42` is an instance of the `int` class. |
| Modules       | A file containing Python code. eg `math` is a module that contains mathematical functions. |
| Packages      | A collection of modules. eg `numpy` is a package that contains the `numpy` module and others as well. |
| Functions     | A block of code that performs a specific task. eg `print("Hello, World!")` is a function that prints *"Hello, World!"* to the console. |
| Variables     | A name that refers to a value. eg `x = 42` is a variable that refers to the value `42`. |
| Expressions   | A combination of values, variables, and operators that evaluates to a single value. eg `2 + 2` is an expression that evaluates to `4`. |
| Statements    | A complete unit of execution. eg `print("Hello, World!")` is a statement that prints *"Hello, World!"* to the console. |
| Arguments     | A value passed to a function. eg `print("Hello, World!")` is a function that takes one argument, the string *"Hello, World!"*. |
| Parameters    | A variable that is part of a function's definition. eg `def greet(name):` is a function that takes one parameter, `name`. |
| Return        | The return statement is used to return a value from a function. eg `return x` is a return statement that returns the value of `x`. |
| Docstrings    | A string literal that appears right after the definition of a function, method, class, or module. eg `def greet(name):` is a function that takes one parameter, `name`. |


## Review Questions

!!! Info "Basic Concepts"

    1. What version of Python is currently the most widely used?
    2. What does PEP stand for in Python, and what is its purpose?
    3. Name three or more popular Python package managers.
    4. What is the difference between `pip list` and `pip freeze`?
    5. How can you check which version of Python you're currently using from the command line?

??? Success "Answers"
    1. Python 3 (or 3.12) and there's no plans for a Python 4
    2. Python Enhancement Proposal - formal design documents for Python features
    3. `pip`, `conda`, `poetry`, `uv`
    4. `pip list` shows all installed packages, `pip freeze` shows installed packages in requirements format
    5. `python --version` or `python -V`

!!! Info "Environment and Tools"

    1. What is a virtual environment and why should you use one?
    2. What file extension is commonly used for Python requirements files?
    3. What's the difference between `python3` and `python` commands on Linux systems?
    4. How do you exit the Python REPL?
    5. What is the purpose of `__init__.py` files in Python projects?

??? Success "Answers"
    1. Isolated environment for Python projects with their own dependencies
    2. .txt (requirements.txt) or .toml (pyproject.toml)
    3. python3 explicitly calls Python 3.x, python might call Python 2.x on older systems
    4. exit() or Ctrl+D (Unix) / Ctrl+Z (Windows)
    5. Marks a directory as a Python package

!!! Info "Standard Library and Modules"

    1. What module would you use to work with JSON data in Python?
    2. Which module helps you work with regular expressions in Python?
    3. Name three modules used for working with dates and times in Python.
    4. What module would you use to generate random numbers in Python?
    5. What's the difference between `import os` and `from os import path`?

??? Success "Answers"
    1. json module
    2. re module
    3. datetime, time, calendar
    4. random module
    5. import os imports the entire module, from os import path only imports the path submodule

## Practical Exercises

- Today we'll be working through the examples found in the Book "Learn Python 3 the Hard Way" found [HERE](https://git.cybbh.space/kevin.m.crotty/delta-students/-/blob/main/resources/lpthw.pdf?ref_type=heads), which will be good preperation for the rest of the Python Course.
- The book is very straight forward and the authors instructions should be followed closely.
- After today they'll be coding challenges accompanying each lesson, local to the `delta-students` repo.


## Additional Resources

- Official Python Documentation: [Python Documentation](https://docs.python.org/3/)
    - I reccomend the tutorial section as a good starting point.
- Great Python Blogs: [Real Python](https://realpython.com/)
    - Once we get into more complex topics, I'd reccomend finding a free blog on the subject.


