# Python Basics

## Objectives

- Python Style Guide and PEPs
- Variables and Assignment
- Dynamic vs Static Typing
- Basic Data Types Overview
- Strings (Basic)
- Type Conversion
- Operators
- Python Keywords


## Python Style Guide and PEPs

Python's philosophy on code style is documented in PEP 8 (Python Enhancement Proposal 8). PEPs are design documents that provide information to the Python community, typically covering new features or processes. For a full list of PEPs, see [Python's PEP Index](https://peps.python.org/).

### Important PEPs to Know

- **PEP 8**: Style Guide for Python Code
    - For all code in this course we will referene the [Google Python Style Guide](https://google.github.io/styleguide/pyguide.html) and it's accompanying [pylint RC file](https://google.github.io/styleguide/pylintrc)
- **PEP 20**: The Zen of Python
- **PEP 257**: Docstring Conventions

??? Success "Zen of Python"

    ```
    The Zen of Python, by Tim Peters

    -- Beautiful is better than ugly.
    Explicit is better than implicit.
    -- Simple is better than complex.
    Complex is better than complicated.
    -- Flat is better than nested.
    Sparse is better than dense.
    -- Readability counts.
    Special cases aren't special enough to break the rules.
    Although practicality beats purity.
    Errors should never pass silently.
    Unless explicitly silenced.
    In the face of ambiguity, refuse the temptation to guess.
    There should be one -- and preferably only one -- obvious way to do it.
    Although that way may not be obvious at first unless you're Dutch.
    Now is better than never.
    Although never is often better than right now.
    If the implementation is hard to explain, it's a bad idea.
    If the implementation is easy to explain, it may be a good idea.
    Namespaces are one honking great idea -- let's do more of those!
    ```

### PEP 8 Style Guidelines

### Naming Conventions

Python has well-established naming conventions that help make code more readable and maintainable. Following these conventions helps other Python developers understand your code more easily.

As always, reference the Google Style Guide to make sure you are compliant for the course.

```python
# GOOD variable names - clear and descriptive
first_name = "John"
total_count = 0
is_valid = True
MAX_CONNECTIONS = 100  # Constants in ALL_CAPS
http_response_code = 200

# BAD variable names - unclear or confusing
fn = "John"                 # Too short, meaning unclear
a = 0                       # Single letters (except for counters)
isvalid = True              # Missing underscore
DATA = "changes"            # ALL_CAPS but not a constant
_private = "hidden"         # Leading underscore without class
__name__ = "don't do this"  # Magic variable (Don't ever modify or use this notation)

# GOOD function names - verb phrases
def calculate_total(items):
    pass

def get_user_info():
    pass

def is_valid_email(email):
    pass

# BAD function names - unclear or inconsistent
def stuff():         # Vague
def Process_data():  # CamelCase (not Python style)
def return_value():  # Doesn't describe action

# GOOD class names - CamelCase
class UserAccount:
    pass

class DatabaseConnection:
    pass

# BAD class names - not CamelCase
class user_account:  # Should be UserAccount
class databaseConnection:  # Inconsistent capitalization

# Special cases
for i in range(10):  # 'i' is okay for simple loop counters
    pass

class Math:
    PI = 3.14159  # Constants in ALL_CAPS
    E = 2.71828
```

!!! Note "Naming Convention Rules"
    - Variables and functions: `lowercase_with_underscores`
    - Classes: `CamelCase`
    - Constants: `ALL_CAPS_WITH_UNDERSCORES`
    - Protected instance attributes: `_leading_underscore`
    - Private instance attributes: `__double_leading_underscore`
    - Magic methods: `__double_underscores__`

!!! Warning "Common Naming Mistakes"
    ```python
    # Don't use these patterns:
    l = [1, 2, 3]      # 'l' looks like '1'
    O = "hello"        # 'O' looks like '0'
    I = "world"        # 'I' looks like '1'

    # Don't use reserved keywords
    class = "Python"   # 'class' is a keyword

    # Don't use built-in function names
    list = [1, 2, 3]   # Shadows built-in 'list'
    str = "hello"      # Shadows built-in 'str'
    ```

!!! Info "Google Python Style Guide"
    Rather then go through all the style choices here (which would make this a very long document) please read the guide linked above and follow the applicable rules as you learn each new concept through the course.

## Variables and Assignment

Variable names are labels we apply to objects in memory. Just like a physical label on a box, we could peel it off and stick it on a different box whenever we want.


```python
# Basic assignment
x = 42
name = "Alice"
is_valid = True

# Multiple assignment
a, b, c = 1, 2, 3

# Augmented assignment
count = 0
count += 1  # Increment by 1

# Variable reassignment (dynamic typing)
value = 42        # value is an integer
value = "Hello"   # now value is a string
value = [1, 2, 3] # now value is a list
```

## Dynamic vs Static Typing


In statically typed languages (like C++, Java, Rust), variable types are checked at compile time and must be declared explicitly. Once a variable is declared as a certain type, it cannot change types during program execution. This provides several benefits: early error detection, better performance (since the compiler can optimize based on known types), and clearer code documentation through type annotations. However, it can also lead to more verbose code and less flexibility when working with data of different types.

```java
// Static typing in Java
String name = "Alice";  // Must be a String
name = 42;  // Compiler error! Can't change type to integer
```


In contrast, dynamically typed languages (like Python, JavaScript, Ruby) perform type checking at runtime, and variables can change types freely during execution. Variables are just names that can reference any type of object. This flexibility allows for more concise code and faster development cycles, especially when prototyping. However, it can also lead to runtime errors that would have been caught at compile time in a statically typed language, and can make code harder to maintain in larger projects without explicit type documentation.

```python
# Dynamically typed in Python
name = "Alice"  # Can be a string or any other type
name = 42  # No error, name now references an integer
name = [1, 2, 3]  # No error, name now references a list
# All of these are valid!
```

!!! Note "Modern Python Development"
    As we'll discuss in later lessons, the concept of "type hinting" has become the standard development practice for Python, which allows a static type checker to be used to verify that the types of variables are correct.  This is a compromise between the static and dynamic typing paradigms.

### Key Type System Concepts

1. **Type Inference**
```python
# Python infers types automatically
x = 42          # Python infers int
y = "hello"     # Python infers str
z = 3.14        # Python infers float

# The type() function reveals the inferred type
print(type(x))  # <class 'int'>
print(type(y))  # <class 'str'>
print(type(z))  # <class 'float'>
```

2. **Type Checking**
```python
# Python checks types at runtime
x = "5"
y = 10
# print(x + y)  # TypeError: can't concat str to int

# But allows explicit conversions
print(int(x) + y)  # 15
print(x + str(y))  # "510"
```

3. **Type Flexibility**
```python
# Variables can change types
x = 42          # int
x = "hello"     # now a str
x = [1, 2, 3]   # now a list

# This flexibility can be both powerful and dangerous
def add_something(x, y):
    return x + y  # Works for numbers AND strings, but can cause runtime errors
```

!!! Warning "Type System Gotchas"
    - Python's dynamic typing means type errors are often discovered at runtime
    - Always verify types when converting between them
    - Be careful with operations that work differently for different types (like `+`)



## Basic Python Data types

- An `int` is a type
- A `str` is a type
- A `bool` is a type

- A `Dog` is a type
- A `UpstairsWindowSmall` is a type

The difference between the *first three* and the *last two* is that the first three are `built-in` types, and the last two are user defined types.

You've probably heard the saying *"everything in Python is an object"*. This is because everything in Python is an instance of a `class`.
Let's look at the output of `dir()` for the `int` and `str` types (abbreviated).

**Methods unique to `int`:**

`['__abs__', '__neg__',  '__pos__', '__pow__', 'as_integer_ratio',  'denominator', 'numerator', 'real', ]`

**Methods unique to `str`:**

`['__contains__', '__getitem__', '__iter__', '__len__', 'capitalize', 'isascii', 'isdecimal', 'replace', 'split', 'upper', 'zfill']`

**Common methods:**

`['__add__', '__dir__', '__doc__', '__eq__', '__format__', '__ge__', '__repr__',]`


Without actually knowing the technical implementation of any of these methods, we can deduce just from their names what they might or probably do and start to play with them.


As we can see from the above example, the "type" of an object determines the characteristics(attributes and methods) of that object.

If two objects have the same exact characteristics, but different names, they are effectively the same type

??? Note "Duck Typing"
    We'll cover duck typing in more detail later, but at a high level, it's the idea that the characteristics of an object are more important than the expicit class definiton or hierarchy.  In fact, only "some" of the characteristics of obects have to be the same to be considered "duck typed".

### Number Data Types

The most common number data type is the `int` type.  The `int` type is used to represent whole numbers, and is only bounded by the size of the memory allocated to the object.

```python
a = 12
b = 12
print(a == b) # True
print(a + b) # 24
c = 230957829304982305723894280935738472 # Totally fine in Python
```

??? Warning "Integers in C"
    In C, you can save an integer value of say 12 into what's called a `char` (short for character).  You could also save it to a `short` or `long` type.  The difference would only be how much memory is wasted for a small value like 12.  As mentioned above, Python dynamically allocates memory for objects based on the size of the data they contain.  There is overhead that causes a speed penalty for small values, but the benefit is that you can use the same variable for any size of number and iterate your code quickly.

Python also has a built in `float` type, which is used to represent floating point numbers.  The `float` type is used to represent decimal numbers. However, there are some traps with floating point numbers as illustrated below.

```python
from decimal import Decimal, getcontext
import math

# Classic floating point problem
print("Float arithmetic:")
print(0.1 + 0.2)                  # Output: 0.30000000000000004
print(0.1 + 0.2 == 0.3)          # Output: False

# Decimal helps with DECIMAL numbers
print("\nDecimal arithmetic:")
print(Decimal('0.1') + Decimal('0.2'))  # Output: 0.3
print(Decimal('0.1') + Decimal('0.2') == Decimal('0.3'))  # Output: True

# BUT Decimal doesn't solve all floating point issues
# Example 1: Irrational numbers
print("\nIrrational numbers:")
print(Decimal(math.pi))  # Still an approximation!

# Example 2: Division problems
print("\nDivision:")
print(Decimal(1) / Decimal(3))  # Default precision
getcontext().prec = 50  # Increase precision
print(Decimal(1) / Decimal(3))  # Still can't represent 1/3 exactly

# Example 3: Binary to Decimal conversion issues
print("\nBinary to Decimal conversion:")
print(Decimal(0.1))
# Converting float to Decimal inherits float's imprecision
0.1000000000000000055511151231257827021181583404541015625
print(Decimal('0.1'))  # Using string input is precise
0.1
```

Key points about `decimal.Decimal`:

-   What it DOES solve:
    -   Exact decimal representation of decimal numbers
    -   Precise control over precision/rounding
    -   Better for financial calculations where exact decimal representation matters

-   What it DOESN'T solve:
    -   Irrational numbers (π, √2, etc.)
    -   Fractions that can't be represented in decimal (1/3, 1/7, etc.)
    -   Binary to decimal conversion issues when creating Decimals from floats


??? Tip "Don't forget about `fractions.Fraction`"
    For fractions, you can use the `fractions` module, which can represent fractions exactly.

    ```python
    from fractions import Fraction

    # Create a Fraction
    x = Fraction(1, 3)
    print(x)  # Output: 1/3

    # Perform arithmetic
    print(x + Fraction(1, 2))  # Output: 5/6
    ```

!!! Note "Floating Point Standards"
    - Floats are defined in the IEEE 754 standard, which is a way to represent floating point numbers in binary.
    - Floating point operations(Flops) are executed in hardware via the FPU (Floating Point Unit) and are very fast.

The logical next question might be *"When should I use floats over Decimal and vice versa?"* Great Question!

**Use Floats for:**

-   Scientific calculations
    - Decimal numbers like 98.6 (body temperature) or 9.81 (gravity) already have a level of inherent uncertainty. The 15-17 decimal places of float precision are more than enough to cover the **Significant Figures** of most scientific measurements.
-   Performance-Critical Code
    - Because Flops are executed in hardware via the FPU, they have a significant computational advantage over Decimal operations, which are executed at the software level.
-   Memory-Efficient Code
    - *See example below*

**Use Decimal for:**

-   Financial calculations
    - Clearly we can't have an "Office Space" situation where fractions of a penny are being misallocated.
-   Exact decimal representation
-   Precise control over precision/rounding

```python
import math
from decimal import Decimal
from time import perf_counter


# Floats are perfect for:
velocity = 9.81 * time          # Physics calculations
wave_function = math.sin(x)     # Trigonometry
temperature = 98.6              # Scientific measurements
distance = math.sqrt(x**2 + y**2) # Geometric calculations

# Float operations are much faster
numbers = range(1000000)

# perf_counter() is just a very accurate timer

start = perf_counter()
float_sum = sum(float(x) / 100 for x in numbers)
print(f"Float time: {perf_counter() - start}")
# Float time: 0.140158822061494

start = perf_counter()
decimal_sum = sum(Decimal(x) / Decimal(100) for x in numbers)
print(f"Decimal time: {perf_counter() - start}")
# Decimal time: 1.4788967268541455 - An order of magnitude slower

import sys

x = 3.14159
d = Decimal('3.14159')
f = Fraction(314159, 100000)

print(f"Size of float: {sys.getsizeof(x)} bytes")
print(f"Size of Decimal: {sys.getsizeof(d)} bytes")
print(f"Size of Fraction: {sys.getsizeof(f)} bytes")
# Size of float: 24 bytes
# Size of Decimal: 104 bytes - Software overhead sucks!
# Size of Fraction: 48 bytes

```

??? Tip "Floating Point Deep Dive"
    Check out [this](https://jvns.ca/blog/2023/01/13/examples-of-floating-point-problems/) site for a great explanation of floating point issues.

### Boolean Data Types

The `bool` type in Python represents one of two values: `True` or `False`. Booleans are essential for control flow and logical operations in programming.

```python
# It's good practice to name your boolean
# variables to match a specific state for easier debugging
is_valid = True
has_error = False


# Booleans can be created from other types
print(bool(1))       # True
print(bool(0))       # False
print(bool(""))      # False
print(bool("text"))  # True
print(bool([]))      # False
print(bool([1, 2]))  # True
```

The concept of "truthiness" is a bit of a gotcha for new Python developers.  In short, Python considers the following values to be `False`:

- `False`
- `None`
- Zero of any numeric type: `0`, `0.0`
- Empty sequences: `""`, `[]`, `()`
- Empty mappings: `{}`

Everything else evaluates to `True`.

This is very powerful when used in control flow.

```python
my_list = [1, 2, 3]
# Note I'm not counting or checking the
# length of the list directly with explicit code
if my_list:
    print("List is not empty")
else:
    print("List is empty")
```

Boolean operators include:
- `and`: Returns `True` if both operands are `True`
- `or`: Returns `True` if at least one operand is `True`
- `not`: Returns the opposite boolean value

```python
x = True
y = False

print(x and y)  # False
print(x or y)   # True
print(not x)    # False

# Short-circuit evaluation

x = 0
y = 10
#  The second part (y / x) is *not* evaluated if the first part is False.
if x != 0 and y / x > 2:
    print("This won't cause a division by zero error")

# Similarly, with 'or', if the first part is True, the second is skipped.
if x == 0 or y / x > 2:
    print("This also avoids division by zero")
## Strings (Basic)

!!! Note "Strings will be covered in more detail later in a future lesson"

Strings are sequences of characters:

```python
# String creation
single_quotes = 'Hello'
double_quotes = "World"
multi_line = """This is a
multi-line string"""

# String concatenation
full_name = "John" + " " + "Doe"

# String methods
text = "Python Programming"
print(text.upper())
print(text.lower())
print(text.split())

# String formatting
name = "Alice"
age = 25
message = f"{name} is {age} years old"
```

## Type Conversion

Python provides several built-in functions to convert between data types:

```python
# String conversions
num_str = str(123)       # "123"
float_str = str(3.14)    # "3.14"
bool_str = str(True)     # "True"

# Numeric conversions
integer = int("123")     # 123
integer = int(3.14)      # 3 (truncates decimal)
float_num = float("3.14") # 3.14
float_num = float(123)   # 123.0

# Boolean conversions
bool(1)        # True
bool(0)        # False
bool("False")  # True (non-empty string)
bool("")       # False (empty string)

# Special cases
int("FF", 16)  # 255 (hex to decimal)
bin(15)        # "0b1111" (to binary string)
hex(255)       # "0xff" (to hex string)
```

!!! Warning "Type Conversion Gotchas"
    - `int()` will raise ValueError for non-numeric strings
    - `float()` will raise ValueError for non-numeric strings
    - `bool("False")` is `True` because it's a non-empty string
    - **Always validate user input before conversion**


## Operators

Python provides a rich set of operators for working with numbers and other data types.

### Arithmetic Operators
```python
# Basic arithmetic
print(10 + 3)    # 13    Addition
print(10 - 3)    # 7     Subtraction
print(10 * 3)    # 30    Multiplication
print(10 / 3)    # 3.333 Division (always returns float)
print(10 // 3)   # 3     Floor division (truncates decimal)
print(10 % 3)    # 1     Modulus (remainder)
print(10 ** 3)   # 1000  Exponentiation

# Order of operations follows PEMDAS
print(2 + 3 * 4)     # 14    (multiplication before addition)
print((2 + 3) * 4)   # 20    (parentheses first)
```

### Comparison Operators
```python
x, y = 5, 10
print(x < y)     # True   Less than
print(x <= y)    # True   Less than or equal
print(x > y)     # False  Greater than
print(x >= y)    # False  Greater than or equal
print(x == y)    # False  Equal to
print(x != y)    # True   Not equal to

# Chained comparisons
print(1 < x < 10)    # True  (equivalent to 1 < x and x < 10)
print(1 < y < 10)    # False
```

### Bitwise Operators

!!! Note "Bitwise Operators"
    There are some very powerful bitwise operations that are beyond the scope of this course, but if you are intersted, look up:
    - Flags and Permissions
    - Network Programming (this is actually in the course ... eventually)
    - Memory Optimization/Packing
    - Faster Math
    - GPU Programming
    - Cryptography

    As a simple example, if your making a game, you could use booleans like `is_equipped`, `is_dead`, `is_alive`, etc.  Save them to a list or dictionary, but if you're using a bitwise operation, you can save a lot of memory by just using a single bit to represent each of these states.

```python
a, b = 60, 13        # 60 = 0011 1100, 13 = 0000 1101
print(a & b)         # 12   AND (0000 1100)
print(a | b)         # 61   OR  (0011 1101)
print(a ^ b)         # 49   XOR (0011 0001)
print(~a)           # -61  NOT (inverts all bits)
print(a << 2)        # 240  Left shift  (1111 0000)
print(a >> 2)        # 15   Right shift (0000 1111)
```

!!! Warning "Bitwise Operators"
    It's much easier to think about bitwise operators when comparing two binary numbers vertically like a traditional addition problem.

    Example:
    ```
    10101010
    01110111
    AND
    00100010
    ```

    There are some

### Assignment Operators
```python
# Simple assignment
x = 5

# Augmented assignment
x += 3      # x = x + 3 (addition) result: 8
x -= 2      # x = x - 2 (subtraction) result: 3
x *= 4      # x = x * 4 (multiplication) result: 20
x /= 2      # x = x / 2 (division) result: 2.5
x //= 2     # x = x // 2 (floor division) result: 2
x %= 3      # x = x % 3 (modulus) result: 2
x **= 2     # x = x ** 2 (exponentiation) result: 25

# Multiple assignment (will cover tuple unpacking in future lessons)
a, b = 1, 2     # a = 1, b = 2
x = y = z = 0   # All variables get same value
```

!!! Note "Division in Python 3"
    - Regular division (`/`) always returns a float
    - Floor division (`//`) returns an integer for integer operands
    - Use `round()` for specific decimal places
    ```python
    print(10 / 4)        # 2.5
    print(10 // 4)       # 2
    print(round(10/4, 2))  # 2.50
    ```

!!! Warning "Common Operator Gotchas"
    ```python
    # Integer division truncates towards negative infinity
    print(10 // 3)       # 3
    print(-10 // 3)      # -4 (not -3)

    # Modulo with negative numbers can be surprising
    print(-10 % 3)       # 2 (not -1)

    # Floating point arithmetic isn't exact
    print(0.1 + 0.2)     # 0.30000000000000004

    # Identity vs Equality
    a = [1, 2, 3]
    b = [1, 2, 3]
    print(a == b)        # True  (same value)
    print(a is b)        # False (different objects)
    ```

### Comparison Operators

Python provides several operators for comparing values:

```python
# Basic comparisons
x = 5
y = 10

print(x == y)    # Equal to (False)
print(x != y)    # Not equal to (True)
print(x < y)     # Less than (True)
print(x <= y)    # Less than or equal to (True)
print(x > y)     # Greater than (False)
print(x >= y)    # Greater than or equal to (False)

# Identity comparison
a = [1, 2, 3]
b = [1, 2, 3]
c = a

print(a is c)    # True (same object in memory)
print(a is b)    # False (different objects with same value)
print(a == b)    # True (same value)

# Chained comparisons
x = 5
print(1 <= x <= 10)  # True (equivalent to: 1 <= x and x <= 10)

# None comparison
value = None
print(value is None)      # Correct way to check for None
print(value == None)      # Works but not recommended
```

- Key points:
    - Use `==` for value comparison
    - Use `is` for identity comparison (especially with `None`)
    - Chained comparisons are more readable than using `and`
    - `is` checks if objects are the same in memory
    - `==` checks if objects have the same value


## Python Keywords

Python keywords are reserved words within the language. In algebra class,
students learn to name values with names like x, y, and 𝛑 (pi).
Likewise, a programmer can assign names to values in Python as well;
however, it would be wise to avoid any Python keyword.

- These words cannot be used as the names of variables or functions.
- They are listed here for reference.

```python
False               class               from                or
None                continue            global              pass
True                def                 if                  raise
and                 del                 import              return
as                  elif                in                  try
assert              else                is                  while
async               except              lambda              with
await               finally             nonlocal            yield
break               for                 not
```

## Check on Learning / Challenges

!!! Warning "Complete challenges on your VTA workstation and dev environment."

- Today's challenges can be found at `~/170D/cybbh/challenges/py_1_1/03_python_basics/{challenge_number}`
- All challenge prompts are contained in their respective `main.py` file as comments.
- The `solution.py` file exists primarily as a last resort after you've exhausted getting help from the instructor, a classmate, or the web.
