# Strings and IO

## Objectives

- Strings
    - String Slicing
    - String Methods
    - String Formatting
    - Special Characters
    - Encoding
- Input/Output
    - print()
    - input()
- Introduction to File System Operations



## Strings

Strings in Python are immutable sequences of Unicode characters. They can be created using single quotes (`'`), double quotes (`"`), or triple quotes (`'''` or `"""`) for multi-line strings.

!!! Note "Follow the google style guide for consistent quote usage recommendations"


```python
# Different ways to create strings
single_quoted = 'Hello'
double_quoted = "World"
multi_line = """This is a
multi-line string"""
# String operations
name = "Python"
print(len(name))          # 6
print(name[0])           # 'P'
print(name[-1])          # 'n' Note that there isn't a negative zero index
print(name[1:4])         # 'yth'
print("th" in name)      # True
print(name.upper())      # 'PYTHON'
```

!!! Note "String Indexing"
    String indexing is zero-based, meaning the first character is at index 0, the second at index 1, and so on. Negative indexing is also supported, where `-1` refers to the last character, `-2` to the second last, and so forth.

### String Slicing

Strings in Python are sequences, which means they support the sequence protocol - including slicing operations. A slice lets you extract a portion of a string using the syntax `string[start:stop:step]`, where all parameters are optional.

```python
text = "Python Programming"

print(text[2]) # 't' - access the 3rd character like a traditional array

# Basic slicing [start:stop] - stop is exclusive
print(text[0:6])      # "Python"
print(text[:6])       # "Python" (start defaults to 0)
print(text[7:])       # "Programming" (stop defaults to end)
print(text[:])        # "Python Programming" (full string copy)

# Slicing with step [start:stop:step]
print(text[::2])      # "Pto rgamn" (every 2nd character)
print(text[::-1])     # "gnimmargorP nohtyP" (reverse string)
print(text[7::2])     # "Pormig" (start at 7, every 2nd char)

# Negative indices in slices
print(text[-11:-2])   # "Programmi"
print(text[-11:])     # "Programming" same as text[7:]
```

The slice notation works because strings implement the sequence protocol, which includes:

- `__len__` - Returns length of sequence
- `__getitem__` - Handles both index and slice access
- `__iter__` - Makes the object iterable

!!! Note "Slice Objects"

    Slices are actually objects in Python. When you use slice notation, Python creates a `slice` object:

    ```python
    # These are equivalent:
    text[1:5]
    text[slice(1, 5)]

    # You can store and reuse slices
    first_word = slice(0, 6)
    print(text[first_word])  # "Python"
    ```

**Common slicing patterns:**

```python
text = "Hello World"
first_five = text[:5]    # Get first n characters
last_five = text[-5:]    # Get last n characters
middle = text[2:-2]      # Remove first/last n characters
reversed = text[::-1]    # Reverse the string
every_other = text[::2]  # Get every nth character
```

!!! Success "Rememeber!!! slicing works on all sequences, not just strings"


!!! Warning "String Immutability"
    While you can slice strings to create new strings, you cannot modify the original string through slicing:
    ```python
    text = "Hello"
    text[1:4] = "i"  # TypeError: 'str' object does not support item assignment
    ```

### String Methods

Python strings come with many built-in methods:

```python
text = " Hello, World! "
# "Hello, World!" removes whitespace from the beginning and end
print(text.strip())
# " hello, world! "
print(text.lower())
# [" Hello", " World! "] splits on delimiter
print(text.split(","))
# "A,B,C" joins a list of strings with a delimiter
print(",".join(["A", "B", "C"]))
# "Hewwo" replace all instances of a substring with another substring
print("Hello".replace("l", "w"))
```

Some sequence methods are shared across all sequences, but some are specific to the type of sequence.

```python
# Methods shared across sequences (strings, lists, tuples):
sequence = "Hello"  # Could also be a list or tuple
print(len(sequence))        # Length works for all sequences
print(sequence.count('l'))  # Count works for all sequences
print(sequence.index('e'))  # Index works for all sequences

# String-specific methods:
text = "Hello, World"
print(text.upper())         # String-only: converts to uppercase
print(text.strip())         # String-only: removes whitespace
print(text.isdigit())       # String-only: checks if string is all digits

# List-specific methods (won't work on strings):
my_list = [1, 2, 3]
my_list.append(4)           # Lists can be modified, strings can't
my_list.extend([5, 6])     # List-specific method
my_list.pop()              # List-specific method

# Common operations that work on all sequences:
print('e' in "Hello")      # Membership testing
print(sequence[1:3])       # Slicing
for item in sequence:      # Iteration
    print(item)
```

- Key differences:
    - Strings are immutable, so they don't have methods that modify in place
    - String methods often deal with text-specific operations (case, whitespace, etc.)
    - List methods often deal with modification (append, remove, etc.)
    - Some operations like len(), slicing, and iteration are universal to all sequences

### String Formatting

Python offers multiple ways to format strings:

```python
name = "Alice"
age = 25

# %-formatting (old style)
print("Name: %s, Age: %d" % (name, age))

# str.format()
print("Name: {}, Age: {}".format(name, age))

# f-strings (Python 3.6+)
print(f"Name: {name}, Age: {age}")
print(f"Age in 5 years: {age + 5}")
```

### Special Characters

Python strings can include special characters using escape sequences. These sequences start with a backslash (`\`) followed by a character that gives them special meaning:

```python
# Common escape sequences
print("Hello\nWorld")     # \n - newline
print("Tab\there")        # \t - tab
print("Quote\"inside\"")  # \" - double quote
print('Quote\'inside\'')  # \' - single quote
print("Path\\to\\file")   # \\ - backslash
print("\u2764")          # \u - Unicode (heart: ❤)
print("\N{SNAKE}")       # \N - Named Unicode (🐍)

# Raw strings ignore escape sequences
print(r"C:\Users\name")  # Prints exactly as written
print(r"This\nis\not\formatted")
```

??? Tip "Common Special Characters"
    - `\n` - Newline
    - `\t` - Tab
    - `\r` - Carriage return
    - `\b` - Backspace
    - `\"` - Double quote
    - `\'` - Single quote
    - `\\` - Backslash
    - `\u```X` - Unicode character (4 hex digits)
    - `\U```XX` - Unicode character (8 hex digits)
    - `\N{name}` - Named Unicode character

### Encoding

Strings in Python are Unicode by default, but they can be encoded into different formats. Python also provides functions to work with the numeric values of characters:

```python
# Character to number conversion
print(ord('A'))          # 65 - Get Unicode code point
print(chr(65))           # 'A' - Get character from code point
print(ord('世'))         # 19990 - Unicode supports international characters
print(chr(19990))        # '世'

# Working with ranges of characters
for i in range(65, 71):  # Print A through F
    print(chr(i), end=' ')  # A B C D E F

# Converting between strings and bytes
text = "Hello, 世界"

# Different encoding schemes
utf8_bytes = text.encode('utf-8')
ascii_bytes = text.encode('ascii', errors='ignore')  # Loses non-ASCII chars
latin1_bytes = text.encode('latin-1', errors='replace')  # Replaces with ?

print(utf8_bytes)        # b'Hello, \xe4\xb8\x96\xe7\x95\x8c'
print(ascii_bytes)       # b'Hello, '
print(latin1_bytes)      # b'Hello, ??'

# Converting back to string
print(utf8_bytes.decode('utf-8'))  # "Hello, 世界"

# Common scenarios where you work with bytes directly:

# 1. Reading binary files (like images)
with open('image.jpg', 'rb') as f:
    image_bytes = f.read()  # Returns bytes object
    # First few bytes often indicate file type
    if image_bytes.startswith(b'\xFF\xD8\xFF'):
        print("This is a JPEG file")

# 2. Network protocols
def handle_network_data(data: bytes):
    # Many network protocols specify exact byte patterns
    if data.startswith(b'GET '):
        print("This is an HTTP GET request")

# 3. Working with fixed-width data
record = bytearray(10)  # Create 10-byte record
record[0:4] = b'USER'   # Set first 4 bytes
record[4:] = b'123456'  # Set remaining 6 bytes

# 4. Memory-efficient string modifications
text = bytearray(b'Hello World')
text[0] = ord('h')  # Modify single byte
print(text.decode())  # 'hello World'

# 5. Cryptography and hashing
import hashlib
message = b"Hello, World!"  # Must be bytes for hashing
hash_value = hashlib.sha256(message).hexdigest()
```

!!! Note "When to Use Bytes"

    **You'll work with bytes directly when:**

    - Reading/writing binary files (images, executables)
    - Implementing network protocols
    - Processing fixed-width data formats
    - Working with cryptography
    - Optimizing memory usage for large data
    - Interfacing with C libraries or system calls

!!! Warning "Encoding Errors"

    When working with encodings, you might encounter errors. Python provides several error handlers:

    ```python
    # Different error handling strategies
    text = "Hello, 世界"
    print(text.encode('ascii', errors='ignore'))   # Drops non-ASCII
    print(text.encode('ascii', errors='replace'))  # Replaces with ?
    print(text.encode('ascii', errors='xmlcharrefreplace'))  # Uses XML entities
    ```

## Input/Output

Python provides several ways to handle input and output operations, from basic console I/O to file operations.

### print()

The print() function is Python's primary way to output information:

```python
# Basic printing
print("Hello, World!")

# Multiple items
name = "Alice"
age = 25
print(name, age)  # Prints: Alice 25

# Separator and end parameters
print("Hello", "World", sep="-")  # Hello-World
print("One", end=" ")  # No newline
print("Two")  # Prints: One Two

# Formatting output
price = 19.99
quantity = 3
# The :.2f is a format specifier that tells Python to format the output as a float with 2 decimal places
print(f"Total: ${price * quantity:.2f}")

# File output (redirect to file)
with open('output.txt', 'w') as f:
    print("This goes to a file", file=f)
```

!!! Note "`with`"

    We'll cover the `with` statement in more detail in future lessons. For now, just know that it's a convenience feature that ensures resources are released when they are no longer needed.

### input()

The input() function reads user input from the console:

```python
# Basic input
name = input("What is your name? ")
print(f"Hello, {name}!")

# Converting input to numbers
age = int(input("How old are you? "))
height = float(input("Enter your height in meters: "))

# Input validation example
while True:
    try:
        number = int(input("Enter an integer: "))
        break
    except ValueError:
        print("That's not a valid integer. Try again.")
print(f"Thank you, {number} is a valid integer.")
```

### Introduction to File Operations

Python makes it easy to work with files. Here's a high-level overview:


=== "Reading a file"
    ```python
    with open('example.txt', 'r') as file:
        # Read entire file
        content = file.read()

        # Or read line by line
        file.seek(0)  # Go back to start of file
        for line in file:
            print(line.strip())

    # NOTE: `file.seek(0)` is used to reset
    # the file pointer to the beginning of the file.
    # This is important when reading a file multiple times,
    # as the file pointer will be at
    # the end of the file after the first read.
    ```
=== "Writing to a file"
    ```python
    with open('output.txt', 'w') as file:
        file.write("Hello, World!\n")
        file.write("Another line\n")

    # Appending to a file
    with open('log.txt', 'a') as file:
        file.write("New log entry\n")

    # NOTE: If you don't specify the append mode,
    # the file.write() method will start at the beginning
    # of the file and overwrite the existing content.
    ```
=== "Multiple Files in a context manager"
    ```python
    def process_data(input_file, output_file):
        with open(input_file, 'r') as infile, \
            open(output_file, 'w') as outfile:
            for line in infile:
                # Process the line
                processed = line.upper()
                outfile.write(processed)

    # NOTE: Multiple files can be opened in the same `with` statement
    # by separating them with a comma.
    # This function effectively converts the input file
    # to uppercase and writes it to the output file.
    ```

### Basic File System Operations

Python's `os` and `pathlib` modules provide file system functionality:

```python
import os
from pathlib import Path

# Check if file exists
if os.path.exists('file.txt'):
    print("File exists!")

# Create directory
os.makedirs('new_folder', exist_ok=True)

# List directory contents
for item in os.listdir('.'):
    print(item)

# Modern approach using pathlib
current_dir = Path('.')
for item in current_dir.iterdir():
    if item.is_file():
        print(f"File: {item}")
    elif item.is_dir():
        print(f"Directory: {item}")

# Working with file paths
path = Path('documents/example.txt')
print(path.parent)  # documents
print(path.name)    # example.txt
print(path.suffix)  # .txt
```

!!! tip "File Handling Best Practices"
    - Always use context managers (with statements) when working with files
    - Close files explicitly if not using context managers
    - Use appropriate error handling for file operations
    - Check file existence before operations
    - Use pathlib for modern, object-oriented file operations

!!! warning "Common Pitfalls"
    - Using hardcoded file paths
    - Not considering different operating systems' path separators
    - Not handling encoding issues with text files

## Check on Learning / Challenges

!!! Warning "Complete challenges on your VTA workstation and dev environment."

- Today's challenges can be found at `~/170D/cybbh/challenges/py_1_1/04_strings_and_io/{challenge_number}`
- All challenge prompts are contained in their respective `main.py` file as comments.
- The `solution.py` file exists primarily as a last resort after you've exhausted getting help from the instructor, a classmate, or the web.
