# Lists

## Objectives

- Lists
    - Creation and Access
    - List Methods
    - List Operations
    - List Slicing
    - Sorting Lists
    - Membership and all/any


## Lists

List objects are ordered, mutable sequences of elements. They can contain items of different types and are defined using square brackets.

- **List Characteristics**
    - Ordered - collection of 0 or more elements(any type)
    - Mutable - can be changed after creation (same object id)
    - Heterogeneous - can contain items of different types
    - Indexable - can be accessed by index
    - Sequenced - can be iterated over

!!! Warning "IndexError - Raised when an index is out of range"


### Creation and Access

```python
# Creating lists
empty_list = list()
empty_list = [] # Technically the square brackets are just a shortcut for calling list()
numbers = [1, 2, 3, 4, 5]
mixed = [1, "hello", 3.14, True]

# Accessing elements (indexing starts at 0)
first = numbers[0]      # 1
last = numbers[-1]      # 5
subset = numbers[1:4]   # [2, 3, 4]

# Nested lists
matrix = [
    [1, 2, 3],
    [4, 5, 6],
    [7, 8, 9]
]
print(matrix[1][1])  # 5
```

### List Methods

```python
# Common list methods
numbers = [1, 2, 3]
numbers_length = len(numbers)

# Adding elements
numbers.append(4)       # [1, 2, 3, 4]
numbers.insert(0, 0)    # [0, 1, 2, 3, 4]
numbers.extend([5, 6])  # [0, 1, 2, 3, 4, 5, 6]

# Removing elements
numbers.pop()          # Removes and returns last element
numbers.pop(0)         # Removes and returns element at index 0
numbers.remove(2)      # Removes first occurrence of 2
numbers.clear()        # Removes all elements

# Other useful methods
nums = [3, 1, 4, 1, 5, 9, 2, 6, 5, 3]
nums.sort()            # Sorts in place
nums.reverse()         # Reverses in place
count = nums.count(5)  # Counts occurrences of 5
index = nums.index(4)  # Finds first index of 4
```


```python
# List concatenation
list1 = [1, 2, 3]
list2 = [4, 5, 6]
combined = list1 + list2  # [1, 2, 3, 4, 5, 6]

# List unpacking
weekdays = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday"]
mon, tue, wed, thu, fri = weekdays
first, *rest = weekdays  # first = "Monday", rest = ["Tuesday", "Wednesday", "Thursday", "Friday"]
first, *middle, last = weekdays  # first = "Monday", middle = ["Tuesday", "Wednesday", "Thursday"], last = "Friday"

# List multiplication
zeros = [0] * 4  # [0, 0, 0, 0]

# Checking membership
numbers = [1, 2, 3, 4, 5]
print(3 in numbers)     # True
print(6 not in numbers) # True
```

### List Slicing

```python
# Basic slicing [start:stop:step]
numbers = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

# Get subset of list
first_three = numbers[0:3]    # [0, 1, 2]
middle = numbers[3:7]         # [3, 4, 5, 6]

# Negative indices
last_three = numbers[-3:]     # [7, 8, 9]
except_last_two = numbers[:-2]  # [0, 1, 2, 3, 4, 5, 6, 7]

# Using step
every_second = numbers[::2]   # [0, 2, 4, 6, 8]
every_third = numbers[::3]    # [0, 3, 6, 9]

# Negative step (reverse)
reversed_list = numbers[::-1]  # [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
reverse_by_two = numbers[::-2] # [9, 7, 5, 3, 1]

# Modifying slices
numbers[1:4] = [10, 20, 30]   # Replace multiple elements
numbers[::2] = [0] * 5        # Replace every second element
```

### Sorting Lists

```python
numbers = [3, 1, 4, 1, 5, 9, 2, 6, 5, 3]
numbers.sort()  # Sorts in place
numbers.sort(reverse=True)
sorted_numbers = sorted(numbers)  # Creates new sorted list

# Sorting with key function
words = ["bananas", "Apple", "cherry"]
sorted_by_len = sorted(words, key=len)  # Sort by length
sorted_by_alpha = sorted(words, key=str.lower)  # Sort by lowercase
```

!!! Note ".sort()" vs "sorted()"
    - `.sort()` - instance method that sorts the list in place
    - `sorted()` - function that returns a new sorted list
    - `sorted()` can take any sequence, not just a list

!!! Note "Alphabetical Sorting"
    - Only like objects can be sorted, so 5 and "5" are not comparable
    - If you have a list of strings, .sort() will sort them alphabetically
    - In the case of "a" vs "A", the ascii value is used, and therefor "A" comes first (65 vs 97 in ascii)

### Membership and all/any

```python
# Membership
numbers = [1, 2, 3, 4, 5]
print(3 in numbers)  # True
print(6 not in numbers)  # True

# all() and any()
my_list = [1, "2", (3, 4), [5, 6]]
all(my_list)  # True

my_list_2 = [2, False, None, 0, ""]
any(my_list_2)  # True

```

### The Power of Lists and Sequences

Lists and sequences are fundamental building blocks in Python programming. They're so essential that you'll find them in virtually every Python program you write or encounter. Here's why:

1. **Data Organization**: Lists help organize related items into a single container, making data management more intuitive and efficient.
2. **Iteration**: Most data processing involves working with sequences of values
3. **Data Transformation**: Lists enable powerful data transformations through methods like map, filter, and list comprehensions
4. **Algorithm Implementation**: Many algorithms rely on list operations
5. **Data Collection and Analysis**: Lists are crucial for collecting and analyzing data

In the next lesson, we'll learn about comprehensions, which will unlock the full capabilities of lists. For now, just get familiar with the syntax and methods.

!!! tip "List Performance"
    - Accessing elements by index: O(1)
    - Appending elements: O(1) amortized
    - Inserting/Deleting elements: O(n)
    - Searching unsorted list: O(n)
    - Slicing: O(k) where k is slice size

!!! note "Sequence Types"
    Lists are just one type of sequence in Python. Others include:
    - Tuples (immutable)
    - Strings (immutable sequences of characters)
    - Range objects (immutable sequences of numbers)
    - Bytes and ByteArrays (for binary data)

    Understanding how to work with sequences is crucial for effective Python programming.

