# Dictionaries

## Objectives

- Dictionaries
    - Key-Value Pairs
    - Dictionary Methods
    - Common Operations
    - Dictionary Sorts


## Dictionaries

Dictionaries are mutable mappings of key-value pairs. Keys must be immutable and unique.

Dictionaries are like lists, but instead of being indexed by a range of numbers, they are indexed by a range of user-defined keys.

### Key-Value Pairs

```python
# Creating dictionaries
empty_dict = dict()
empty_dict = {}
person = {
    "name": "Alice",
    "age": 25,
    "city": "New York"
}

# Accessing values
print(person["name"])        # Using square brackets
print(person.get("age"))     # Using get() method
print(person.get("phone", "Not found"))  # With default value

# Modifying dictionaries
person["age"] = 26           # Update value
person["phone"] = "555-1234" # Add new key-value pair
```

### Dictionary Methods

```python
# Common dictionary methods
person = {"name": "Alice", "age": 25}

# Getting information
keys = person.keys()       # dict_keys(['name', 'age'])
values = person.values()   # dict_values(['Alice', 25])
items = person.items()     # dict_items([('name', 'Alice'), ('age', 25)])

# NOTE: These methods return views, which look like lists but are not. If we modify the dictionary, the views will reflect the changes. Having said that, you can easily convert them to lists if needed.

# Modifying dictionary
person.update({"city": "New York", "age": 26})
removed = person.pop("age")     # Removes and returns value
person.clear()                  # Removes all items

# Creating with dict methods
names = dict(first="Alice", last="Smith")
pairs = dict([("name", "Bob"), ("age", 30)])


#TODO: .get method
print(person.get("age"))
print(person.get("phone", "Not found"))

#TODO: defaultdict
from collections import defaultdict

# Create a defaultdict with a default value of 0
dd = defaultdict(int)


```


!!! tip "defaultdict"
    You'll commonly see `from collections import defaultdict` in codebases instead of using the vanilla `dict` type. This is a dictionary subclass that provides a default value for any key that hasn't been set yet.

    This:

    - Provides error prevention, avoiding `KeyError` when accessing a key that doesn't exist
    - Simplifies code by eliminating the need for default values in dictionaries
    - Improves efficiency by reducing the need for checking if a key exists before accessing it
    - Matches common patterns in data that lends itself to using default values

### Common Operations

```python
# Dictionary operations
inventory = {
    "apple": 5,
    "banana": 8,
    "orange": 12
}

# Checking membership
print("apple" in inventory)         # True
print("grape" not in inventory)     # True

# NOTE: When checking membership, the key is checked, not the value.
# If you need to check the value, you can use the `in` operator on the values() method.

print(5 in inventory.values()) # True

# Looping through dictionaries
for key in inventory: # Equivalent to inventory.keys()
    print(f"{key}: {inventory[key]}")

for key, value in inventory.items():
    print(f"{key}: {value}")

```

!!! Warning "The examples below employ features not yet covered. Be sure to come back to them in a few days if you're not sure what's going on."

### Dictionary Sorts

Dictionaries cannot be sorted; however, the keys and values can be passed to
the `sorted` function. A sorted `list` will be returned.

*   The keys and values of a dictionary can be converted to a `list`, which
    could then be sorted by using the `sort` method of the `list`.

*   Alternatively, the `sorted` function can be used.

*   To sort the resulting `list` objects by the values in the dictionary, use
    the `get` method of the dictionary as the `key` of the `sort` method and the
    `sorted` function.

```python
#!/usr/bin/env python3

states = {'Nevada': 'NV', 'Maine': 'ME',
          'Maryland': 'MD', 'Massachusetts': 'MA',
          'New Hampshire': 'NH'}

states_by_long_name = list(states.keys())  # Convert to list
states_by_long_name.sort()
print(f'States by long name: {states_by_long_name}')
# Output: States by long name: ['Maine', 'Maryland', 'Massachusetts', 'Nevada', 'New Hampshire']


states_by_short_name = list(states.keys()) # Convert to list
states_by_short_name.sort(key=states.get) # Sort by dictionary values
print(f'States by short name: {states_by_short_name}')
# Output: States by short name: ['Massachusetts', 'Maryland', 'Maine', 'New Hampshire', 'Nevada']

print()

states_by_long_name = sorted(states.keys())
print(f'States by long name: {states_by_long_name}')
# Output: States by long name: ['Maine', 'Maryland', 'Massachusetts', 'Nevada', 'New Hampshire']

states_by_short_name = sorted(states.keys(), key=states.get)
print(f'States by short name: {states_by_short_name}')
# Output: States by short name: ['Massachusetts', 'Maryland', 'Maine', 'New Hampshire', 'Nevada']

states_by_short_name_reversed = sorted(states.keys(), key=states.get, reverse=True)
print(f'States by short name reversed: {states_by_short_name_reversed}')
# Output: States by short name reversed: ['Nevada', 'New Hampshire', 'Maine', 'Maryland', 'Massachusetts']
```


## Basic Collections Operations

### Sorting

```python
# Sorting lists
numbers = [3, 1, 4, 1, 5, 9, 2, 6]
sorted_nums = sorted(numbers)           # Creates new sorted list
numbers.sort()                         # Sorts in place

# Sorting with key function
words = ["banana", "apple", "cherry"]
sorted_by_len = sorted(words, key=len)  # Sort by length
```

## Real World Examples

### Network Security Log Analysis

```python
# Log entry processing using lists and dictionaries
log_entries = [
    {"timestamp": "2024-03-15 10:23:45", "ip": "192.168.1.100", "event": "LOGIN_FAILED"},
    {"timestamp": "2024-03-15 10:23:47", "ip": "192.168.1.100", "event": "LOGIN_FAILED"},
    {"timestamp": "2024-03-15 10:23:50", "ip": "192.168.1.100", "event": "LOGIN_FAILED"},
    {"timestamp": "2024-03-15 10:24:01", "ip": "192.168.1.101", "event": "LOGIN_SUCCESS"},
]

# Count login attempts by IP
login_attempts = {}
for entry in log_entries:
    ip = entry["ip"]
    login_attempts[ip] = login_attempts.get(ip, 0) + 1

# Identify potential brute force attacks
suspicious_ips = {
    ip: count for ip, count in login_attempts.items()
    if count >= 3
}

print("Suspicious IPs:", suspicious_ips)
```

### Port Scanner Results Processing

```python
# Storing and analyzing port scan results
scan_results = {
    "192.168.1.1": {
        "ports": [80, 443, 22, 3389],
        "os": "Windows",
        "status": "active"
    },
    "192.168.1.2": {
        "ports": [22, 80],
        "os": "Linux",
        "status": "active"
    },
    "192.168.1.3": {
        "ports": [],
        "os": "unknown",
        "status": "inactive"
    }
}

# Find all machines with specific open ports
def find_machines_with_port(results, target_port):
    return [
        ip for ip, data in results.items()
        if target_port in data["ports"]
    ]

# Find machines with SSH (port 22) open
ssh_enabled = find_machines_with_port(scan_results, 22)
print("Machines with SSH enabled:", ssh_enabled)

# Count machines by OS
os_count = {}
for data in scan_results.values():
    os = data["os"]
    os_count[os] = os_count.get(os, 0) + 1

print("OS distribution:", os_count)
```

### Packet Capture Analysis

```python
# Analyzing network packets
packets = [
    {"source": "10.0.0.1", "dest": "8.8.8.8", "size": 64, "protocol": "DNS"},
    {"source": "10.0.0.2", "dest": "192.168.1.100", "size": 1500, "protocol": "HTTP"},
    {"source": "10.0.0.1", "dest": "8.8.8.8", "size": 64, "protocol": "DNS"},
    {"source": "10.0.0.3", "dest": "192.168.1.100", "size": 550, "protocol": "HTTP"},
]

# Group packets by protocol
protocol_packets = {}
for packet in packets:
    proto = packet["protocol"]
    if proto not in protocol_packets:
        protocol_packets[proto] = []
    protocol_packets[proto].append(packet)

# Calculate total data transferred per source IP
data_by_source = {}
for packet in packets:
    source = packet["source"]
    data_by_source[source] = data_by_source.get(source, 0) + packet["size"]

# Find potential DNS tunneling (frequent small DNS packets)
dns_sources = {}
for packet in packets:
    if packet["protocol"] == "DNS":
        source = packet["source"]
        dns_sources[source] = dns_sources.get(source, 0) + 1

suspicious_dns = {
    ip: count for ip, count in dns_sources.items()
    if count > 1
}
```

### System User Management

```python
# Managing system users and permissions
users = {
    "admin": {
        "permissions": ["read", "write", "execute"],
        "groups": ["sudo", "staff"],
        "last_login": "2024-03-15 09:00:00"
    },
    "analyst": {
        "permissions": ["read", "write"],
        "groups": ["staff"],
        "last_login": "2024-03-15 10:30:00"
    },
    "guest": {
        "permissions": ["read"],
        "groups": ["visitors"],
        "last_login": "2024-03-14 15:45:00"
    }
}

# Check user permissions
def has_permission(username, permission):
    if username not in users:
        return False
    return permission in users[username]["permissions"]

# Find all users in a specific group
def users_in_group(group):
    return [
        username for username, data in users.items()
        if group in data["groups"]
    ]

# List users with elevated permissions
sudo_users = users_in_group("sudo")
print("Users with sudo access:", sudo_users)

# Remove expired guest accounts (not logged in for 24 hours)
from datetime import datetime, timedelta
current_time = datetime.now()
expired_users = [
    username for username, data in users.items()
    if (datetime.strptime(data["last_login"], "%Y-%m-%d %H:%M:%S")
        < current_time - timedelta(days=1))
    and "visitors" in data["groups"]
]
```

### Firewall Rule Management

```python
# Managing firewall rules
firewall_rules = [
    {"id": 1, "source": "192.168.1.0/24", "dest": "ANY", "port": 80, "action": "ALLOW"},
    {"id": 2, "source": "ANY", "dest": "192.168.1.100", "port": 22, "action": "DENY"},
    {"id": 3, "source": "10.0.0.0/8", "dest": "192.168.1.0/24", "port": 443, "action": "ALLOW"},
]

# Sort rules by ID
sorted_rules = sorted(firewall_rules, key=lambda x: x["id"])

# Find all rules affecting a specific port
def find_rules_for_port(rules, port):
    return [rule for rule in rules if rule["port"] == port]

# Group rules by action
rules_by_action = {}
for rule in firewall_rules:
    action = rule["action"]
    if action not in rules_by_action:
        rules_by_action[action] = []
    rules_by_action[action].append(rule)

# Check if traffic would be allowed
def check_traffic(source_ip, dest_ip, port):
    for rule in firewall_rules:
        # This is a simplified check - real implementation would need
        # proper IP/subnet matching
        if (rule["source"] in [source_ip, "ANY"] and
            rule["dest"] in [dest_ip, "ANY"] and
            rule["port"] == port):
            return rule["action"] == "ALLOW"
    return False  # Default deny
```

## Check on Learning / Challenges

!!! Warning "Complete challenges on your VTA workstation and dev environment."

- Today's challenges can be found at `~/170D/cybbh/challenges/py_1_1/05_data_structures/{challenge_number}`
- All challenge prompts are contained in their respective `main.py` file as comments.
- The `solution.py` file exists primarily as a last resort after you've exhausted getting help from the instructor, a classmate, or the web.
