# Conditionals

## Objectives

- Conditional Statements
    - if/elif/else
    - Comparison Operators
    - Logical Operators
    - Ternary Operators
    - Switch Statements (3.10+)


## Conditional Statements

Python uses conditional statements to make decisions in code. The most basic form is the `if` statement.

### if/elif/else

The `if` statement executes code when a condition is True. You can add `elif` (else if) for additional conditions, and `else` as a catch-all:

```python
# Basic if statement
temperature = 75

if temperature > 80:
    print("It's hot!")
elif temperature > 60:
    print("It's nice out!")
else:
    print("It's cold!")
```

??? Note "Truthiness"
    Recall that in Python, `0`, `""`, `[]`, `{}`, and `None` are all considered `False`.  Everything else is considered `True`.

    So the following is valid:

    ```python
    my_list = ["not", "an", "empty", "list"]
    if my_list:
        print("List is not empty")
    else:
        print("List is empty")
    ```

```python
# Basic Input Validation
number = input("Enter a number between 1 and 100")
if number.isnumeric():
    number = int(number)
else:
    print("Invalid input")
```

```python
# pass keyword
if some_condition:
    pass # keyword used to indicate that no code should be executed
else:
    print("Condition was false")

```

### Comparison Operators (Refresher)

Python provides several operators for comparing values:

```python
x = 5
y = 10

print(x == y)  # Equal to (False)
print(x != y)  # Not equal to (True)
print(x < y)   # Less than (True)
print(x > y)   # Greater than (False)
print(x <= y)  # Less than or equal to (True)
print(x >= y)  # Greater than or equal to (False)
print(x is y)  # Identity (False)
print(x is not y)  # Negated identity (True)
```

```python
value = int(input('Please enter a whole number: '))

print(value, end=' is ')
if value <= 5:
    print('less than or equal to 5')
elif value <= 10:
    print('between 6 and 10 inclusively')
elif value <= 15:
    print('between 11 and 15 inclusively')
else:
    print('greater than 15')

print('This statement is not part of the above if else')
```

### Logical Operators

Logical operators (`and`, `or`, `not`) combine or modify boolean conditions:

```python
# AND operator - both conditions must be True
sunny = True
warm = True
good_beach_day = sunny and warm  # True

# OR operator - at least one condition must be True
has_cash = False
has_credit_card = True
can_pay = has_cash or has_credit_card  # True

# NOT operator - inverts a boolean value
is_raining = True
is_not_raining = not is_raining  # False

# Complex combinations
age = 25
income = 50000
has_license = True

# All conditions must be True
can_rent_car = (age >= 21) and has_license and (income > 30000)

# NOTE: the parentheses help group the expressions for readability, but are not required.

# At least one condition must be True
needs_umbrella = is_raining or (not sunny)
```

### Ternary Operators

A ternary operator is a one-line conditional expression that returns one of two values based on a condition. Think of it as a compressed if/else statement:

General syntax: `[true_part] if [condition] else [false_part]`

```python
# Traditional if/else
if age >= 18:
    status = "adult"
else:
    status = "minor"

# Same logic using ternary operator
status = "adult" if age >= 18 else "minor"

# More examples
price = 100
discount = 0.9 if is_student else 1.0
final_price = price * discount

message = "Allowed" if age >= 21 else "Not allowed"

# Can be nested (but avoid for readability)
grade = "A" if score >= 90 else "B" if score >= 80 else "C"

# Don't do this! Use regular if/elif/else instead
result = ("High" if x > 100 else
          "Medium" if x > 50 else
          "Low" if x > 0 else
          "Invalid")
```

**Key Differences:**

1. **Purpose**:
   - Logical operators combine boolean conditions
   - Ternary operators provide a shorthand for if/else value assignment

2. **Syntax**:
   - Logical: `condition1 and/or condition2`
   - Ternary: `value_if_true if condition else value_if_false`

3. **Output**:
   - Logical operators return `True` or `False`
   - Ternary operators can return any type of value

4. **Use Case**:
   - Logical operators for decision making
   - Ternary operators for concise value assignment

```python
# Example combining both
age = 20
has_id = True

# Logical operator in the condition
# Ternary operator for the assignment
status = "Welcome" if (age >= 21 and has_id) else "Not allowed"

# Breaking this down:
# 1. (age >= 21 and has_id) is evaluated first (logical)
# 2. Based on that result, either "Welcome" or "Not allowed"
#    is assigned (ternary)
```

!!! Warning "Do not use ternary operators for complex logic where readability will suffer."


### Switch Statements (3.10+)

Switch statements, formally known as structural pattern matching, are a new feature in Python 3.10 using the `match/case` syntax.  They provide a powerful way to handle different code paths based on the structure and value of a variable.  While similar in concept to switch statements in other languages, Python's `match` statement offers more flexibility through pattern matching.  It allows you to not only match against specific values but also against sequences, classes, and more complex data structures.  Importantly, Python's `match` does **not** have fallthrough; only the first matching case is executed.


??? Warning "Implied Break"
    Unlike other languages, the `match` statement has an implicit `break` if the case is matched.  This means that once a case is matched, the execution will break out of the `match` statement. In C, if you don't include the break statements, all the cases will be executed.

**Example 1 (Matching Values):**

```python
http_status = 404

match http_status:
    case 200:
        print("OK")
    case 404:
        print("Not Found")
    case 500:
        print("Internal Server Error")
    case _:  # Default case
        print("Unknown Status Code")
```

**Example 2 (Matching Data Structures):**

```python
user_data = ("alice", 30, "admin")  # Tuple representing user data

match user_data:
    case (name, age, "admin"):
        print(f"Admin user: {name}, age {age}")
    case (name, age, role):
        print(f"Regular user: {name}, age {age}, role {role}")
    case _:
        print("Invalid user data")
```

**Example 3 - Firewall Rules:**

```python
import ipaddress

def ip_in_network(ip_str, network_str):
    try:
        ip = ipaddress.ip_address(ip_str)
        network = ipaddress.ip_network(network_str)
        return ip in network
    except ValueError:  # Handle invalid IP addresses or network strings
        return False

packet = {  # Example packet
    "source_ip": "192.168.1.10",
    "destination_ip": "10.0.0.5",
    "port": 80,
    "protocol": "TCP"
}

match packet:
    case {"source_ip": src,
          "destination_ip": dst,
          "port": 80,
          "protocol": "TCP"} if ip_in_network(dst, "10.0.0.0/24"):
        print(f"Allowed HTTP traffic from {src} to internal network ({dst})")
    case {"source_ip": src,
          "destination_ip": "192.168.2.100",
          "port": 22,
          "protocol": "TCP"}:
        print(f"Allowed SSH from {src} to DMZ server")
    case {"protocol": "ICMP"}:
        print("Blocked ICMP (ping)")
    case {"source_ip": src,
          "destination_ip": dst,
          "port": port,
          "protocol": proto}:
        print(f"Traffic from {src} to {dst} on port {port} using {proto} blocked by default")
    case _: # Default case
        print("Corrupt or improperly formatted packet")
```



## Loops

Loops allow you to repeat code multiple times.

### for loops

The for loop is used to iterate over sequences:

!!! Note "Sequences"
    Remember that the term "sequence" is used to describe a list, tuple, string, or custom data structure that can be iterated over. What determines whether an object can be iterated over will be covered in a later section.

```python

# String
for char in "hello":
    print(char)

# Tuple
my_tuple = (1, 2, 3)
for num in my_tuple:
    print(num)

# Dictionary (keys)
my_dict = {"a": 1, "b": 2}
for key in my_dict:  # Iterates over keys by default
    print(key)

# Dictionary (values)
for value in my_dict.values():
    print(value)

# Dictionary (key-value pairs)
for key, value in my_dict.items():
    print(f"Key: {key}, Value: {value}")

# Iterating over a range
for i in range(3):
    print(i)  # Outputs: 0, 1, 2

# Iterating over a list
fruits = ["apple", "banana", "cherry"]
for fruit in fruits:
    print(fruit)

# NOTE: how easy it is to iterate over a list without specifying an index, counter, or length as in other languages.

# Enumerate for index and value
for index, fruit in enumerate(fruits):
    print(f"{index}: {fruit}")
```

### while loops

While loops repeat as long as a condition is True:

!!! Warning "DANGEROUS - Infinite loop!"
    ```python
    while True:
        print("This will run forever!")
    ```

```python
# Always True loop
while game_running == True:
    # game loop code
    show_game_screen()
    get_mouse_input()
    update_game_state() # One of these functions will typically set game_running to False

# Counting loop
count = 0
while count < 3:
    print(count)
    count += 1

# With user input
password = ""
while password != "secret":
    password = input("Enter password: ")
```

### Loop Control

Break and continue help control loop execution:

```python
# break example
for i in range(10):
    if i == 5:
        break  # Exit loop when i is 5
    print(i)

# NOTE: Every line below the break, but inside the outer loop block will be skipped
# and no further loops will be executed.

# continue example
for i in range(5):
    if i == 2:
        continue  # Skip printing 2
    print(i)

# NOTE: continue will skip the rest of the loop body like break,
# but will continue to execute the next iteration.
```

### Loop with else

Python allows an else clause with loops that executes when the loop completes normally:

```python
# Finding a number
numbers = [1, 3, 5, 7, 9]
search_for = 4

for num in numbers:
    if num == search_for:
        print("Found it!")
        break
else:
    print("Number not found")  # This will execute

# NOTE: The else clause will execute if the loop reaches the end of the sequence without breaking.
```

### Comprehensions

List and Dictionary comprehensions provide a concise way to create lists and dictionaries:

!!! Warning "New Developer Note"

    **Warning 1:** Remeber that the return value of a comprehension is a new list or dictionary. If you are just trying to compress a for loop into a single line or something else weird, just use a regular for loop.

    **Warning 2:** Novice programmers often "fall in love" with comprehensions and use them for everything. This is a bad idea. Comprehensions are a powerful tool and should be used when it makes sense to do so.

    **Warning 3:** KISS (Keep It Simple Stupid). If you think a novice programmer would struggle to understand the comprehension, then be assured that experienced programmers will to. There are plenty of professional developers who strictly prohibit the use of comprehensions, believing that they make code harder to read and understand.

**List Comprehensions**

```python
# Traditional way
squares = []
for x in range(5):
    squares.append(x**2)

# Using list comprehension
squares = [x**2 for x in range(5)]
print(squares)  # Output: [0, 1, 4, 9, 16]

# With conditional
even_squares = [x**2 for x in range(10) if x % 2 == 0]
print(even_squares)  # Output: [0, 4, 16, 36, 64]
```

**Dictionary Comprehensions**

```python
# Traditional way
squares = {}
for x in range(5):
    squares[x] = x**2

# Using dictionary comprehension
squares = {x: x**2 for x in range(5)}
print(squares)  # Output: {0: 0, 1: 1, 2: 4, 3: 9, 4: 16}

# NOTE: The main difference is that 2 expressions are used to the left of the for loop instead of 1. The first is the key, the second is the value.
```

## Check on Learning / Challenges

!!! Warning "Complete challenges on your VTA workstation and dev environment."

- Today's challenges can be found at `~/170D/cybbh/challenges/py_1_1/06_control_flow/{challenge_number}`
- All challenge prompts are contained in their respective `main.py` file as comments.
- The `solution.py` file exists primarily as a last resort after you've exhausted getting help from the instructor, a classmate, or the web.
