
## Practical Lab: Network Configuration Auditor

**Overall Goal:**  We're going to build a script that audits the configuration of network devices.  It will read device information from a file, check for specific configuration settings (which we'll define as we go), and report on any discrepancies or potential issues. This will touch on variables, data types, file I/O, string manipulation, control flow, and data structures.

---

```python
"""
# Objectives

* Reinforce core Python programming concepts through a practical, hands-on exercise.
* Build a functional script from the ground up, demonstrating the power of combining
  different programming techniques.
* Understand how Python can be used to automate tasks related to network configuration
  management.

# Introduction

You are a network engineer tasked with ensuring the consistency and security of your
organization's network devices. You'll develop a Python script to automate the
process of checking device configurations against a set of predefined rules.

"""
```

## Challenge 1: Device Data Representation

```python
"""
# Instructions

1.  Create variables to store the following information for a *single* network device:
    *   `hostname`: (string) The device's hostname (e.g., "router-01").
    *   `ip_address`: (string) The device's primary IP address (e.g., "192.168.1.1").
    *   `model`: (string) The device model (e.g., "Cisco ISR4331").
    *   `is_enabled`: (boolean) Whether the device's administrative status is enabled (True) or disabled (False).
    *   `services`: (list of strings) A list of services running on the device (e.g., ["ssh", "telnet", "http"]).

2.  Assign appropriate values to each variable. You can make up the specific values,
    but they should be realistic for a network device.

3. Print out each variable and its value on separate lines.  Use f-strings for clean output.
    Example:
    Hostname: router-01
    IP Address: 192.168.1.1
    ...
"""
```

## Challenge 2: Storing Multiple Devices

```python
"""
# Instructions

1.  Create a *list* called `devices`.

2.  Create *three* dictionaries, each representing a different network device.
    Each dictionary should have the following keys (using the variable names from
    Challenge 1 as a guide):
    *   `hostname`
    *   `ip_address`
    *   `model`
    *   `is_enabled`
    *   `services`

    Populate each dictionary with data for a different device.  Make sure the
    `services` key contains a list of services (which can be empty).

3.  Append each of the three device dictionaries to the `devices` list.

4.  Use a `for` loop to iterate through the `devices` list.  Inside the loop,
    print the `hostname` and `ip_address` of each device, formatted like this:

    ```
    Device: router-01, IP: 192.168.1.1
    Device: switch-02, IP: 192.168.1.2
    ...
    ```
"""
```

## Challenge 3: Reading Device Data from a File

```python
"""
# Instructions

1.  Create a text file named `devices.txt`.  Each line in the file should
    represent a single device and have the following format:

    `hostname,ip_address,model,is_enabled,service1;service2;service3`

    *   Fields are separated by commas.
    *   The `is_enabled` field should be either "True" or "False" (as strings).
    *   Services are separated by semicolons.  If a device has no services,
        the last field should still be present, but empty (e.g., `router-03,10.0.0.3,ModelX,True,`).

2.  Add at least three lines to `devices.txt`, representing different devices.

3.  Write Python code to:
    *   Open the `devices.txt` file for reading.
    *   Read the file line by line.
    *   For each line:
        *   Split the line into its individual fields.
        *   Create a dictionary representing the device (using the same keys as before).
        *   Convert the `is_enabled` field to a boolean value (True or False).
        *   Split the services string into a list of services.
        *   Append the device dictionary to a list called `devices` (reusing the name from Challenge 2 - we're replacing the hardcoded list).
        *   Be sure to strip() any extra whitespace.

4.  After reading the file, use a `for` loop to print the contents of the `devices`
    list (as dictionaries), verifying that the data was read and processed correctly.
"""
```

## Challenge 4: Configuration Checks - Part 1 (Basic Checks)

```python
"""
# Instructions

Now we'll start adding configuration checks.  These checks will be functions
that take a device dictionary as input and return `True` if the device passes
the check, and `False` otherwise.

1.  Create a function called `check_ip_address_format(device)`.  This function
    should:
    *   Take a device dictionary as input.
    *   Check if the `ip_address` field is in a valid IPv4 format (four numbers
        separated by dots, each number between 0 and 255). You don't need a
        complex regular expression; simple splitting and checking is sufficient.
    *   Return `True` if the format is valid, `False` otherwise.

2.  Create a function called `check_device_enabled(device)`. This function should:
    *   Take a device dictionary as input.
    *   Check if the `is_enabled` field is `True`.
    *   Return `True` if it is, `False` otherwise.

3.  Create a function called `check_service_running(device, service_name)`. This function should:
    *   Take a device dictionary as input, *and* a `service_name` (string).
    *   Check if the `service_name` is present in the device's `services` list.
    *   Return `True` if the service is running, `False` otherwise.

4.  Test your functions:
    *   Call each function with one of the device dictionaries you created in
        Challenge 3.
    *   Print the results of each function call to verify they work correctly.
    *   Test `check_service_running` with both a service that *is* running and
        a service that is *not* running.

"""
```

## Challenge 5: Configuration Checks - Part 2 (Report Generation)

```python
"""
# Instructions

Now we'll combine the checks and generate a report.

1.  Create a function called `audit_devices(devices, disallowed_services)`.  This
    function should:
    *   Take a list of device dictionaries (`devices`) as input.
    * Take a list of strings called `disallowed_services`
    *   Iterate through the `devices` list.
    *   For each device:
        *   Check if the IP address format is valid (using `check_ip_address_format`).
        *   Check if the device is enabled (using `check_device_enabled`).
        *   Check if any of the `disallowed_services` are running on the device (using `check_service_running`).
        *   Print a report for *each* device, indicating:
            *   The device's hostname.
            *   Whether the IP address format is valid.
            *   Whether the device is enabled.
            *   A list of any disallowed services found running on the device.  If none are found, print "No disallowed services found."

2.  Call `audit_devices` with the `devices` list you created in Challenge 3,
    and a list of disallowed service names, that you create for this function call.  For example:
    ```python
    disallowed_services = ["telnet", "ftp"]
    audit_devices(devices, disallowed_services)
    ```

3.  Run your script and observe the output.

"""
```

## Challenge 6: Writing the Audit Report to a File

```python
"""
# Instructions

Finally, let's write the audit report to a file instead of just printing it
to the console.

1.  Modify the `audit_devices` function to take an additional argument:
    `output_file` (a string representing the filename).

2.  Inside `audit_devices`, open the `output_file` for writing *before* you
    start iterating through the devices.  Use a `with` statement.

3.  Instead of printing the report to the console, *write* the report to the
    output file. You'll need to convert the report for each device into a string.

4.  After the loop finishes, the file will be automatically closed (thanks to
    the `with` statement).

5.  Call `audit_devices` again, but this time, pass in a filename (e.g.,
    "audit_report.txt") as the `output_file` argument.

6. Run your script, and verify a new file, matching the passed filename, is created, and that the file
    contains the audit report.
"""
```

