# Advanced Data Concepts

## Objectives

- Mutability vs Immutability
    - Understanding object behavior
    - Identifying mutable and immutable types
    - Impact on function behavior
- Variable References
    - Memory addresses
    - Reference sharing
    - Identity vs equality
- Shallow vs Deep Copy
    - Copy module
    - Nested data structures
    - Independence of copies
- Memory Management Basics
    - Reference counting
    - Garbage collection
    - Resource management
- Advanced Collections
    - Tuples
    - Sets
    - Named Collections



## Mutability vs Immutability

Mutable objects can be changed after creation, while immutable objects cannot. This distinction is crucial for predicting how your data will behave, especially when passing to functions or sharing across your program.

```python
# Demonstrating mutable objects (lists)
my_list = [1, 2, 3]
another_list = my_list  # Both variables reference the same list
my_list[0] = 100        # Modifies the object that both variables reference
print(another_list)     # Output: [100, 2, 3]

# Demonstrating immutable objects (strings)
my_string = "hello"
another_string = my_string  # Both variables reference the same string
my_string = "hello world"   # Creates a new string, doesn't affect another_string
print(another_string)       # Output: "hello"
```

!!! Note "Types by Mutability"
    **Mutable types**: Lists, dictionaries, sets, user-defined classes

    **Immutable types**: Integers, floats, strings, tuples, frozensets, booleans

### Practical Implications

Mutability affects how objects behave when passed to functions:

```python
# Function behavior with mutable objects
def modify_list(a_list):
    a_list.append(5)  # Modifies the original list

test_list = [1, 2, 3]
modify_list(test_list)
print(test_list)  # Output: [1, 2, 3, 5]

# Contrasting methods: in-place vs return new
# Sort in-place (modifies original)
my_list = [3, 1, 2]
my_list.sort()
print(my_list)  # Output: [1, 2, 3]

# Sorted returns a new list
my_list = [3, 1, 2]
new_list = sorted(my_list)
print(my_list)   # Output: [3, 1, 2]
print(new_list)  # Output: [1, 2, 3]
```

!!! Warning "Common Pitfalls"
    - Modifying a mutable object that is referenced by multiple variables will affect all references
    - In-place methods modify the object rather than returning a new one
    - Collection elements maintain their original mutability characteristics

!!! Tip "Best Practices"
    - Favor immutability when possible for predictable behavior
    - Be aware of whether a function modifies objects in-place or returns new ones
    - Use the `id()` function to check if two variables reference the same object

??? "Advanced Topic: Object Identity"
    ```python
    # Checking object identity
    a = [1, 2, 3]
    b = a
    c = [1, 2, 3]  # Same content, different object

    print(id(a) == id(b))  # True - same object
    print(id(a) == id(c))  # False - different objects
    print(a == c)          # True - same content (equality)
    print(a is c)          # False - not the same object (identity)
    ```

## Variable References

In Python, variables don't store values directly; they store references to objects in memory. This reference system is fundamental to understanding how data behaves in Python.

```python
# Multiple references to the same object
a = [1, 2, 3]
b = a        # b refers to the same list object
print(id(a), id(b))  # Same memory address

# Modifying through one reference affects the object
b.append(4)
print(a)     # Output: [1, 2, 3, 4]

# Assignment creates a new reference, not a copy
c = a
a = [5, 6, 7]  # a now points to a new list
print(c)       # Output: [1, 2, 3, 4]
```

!!! Note "Reference vs Value"
    Python uses a "pass by object reference" mechanism, which means it passes references to objects rather than copying values or passing direct memory addresses.

### Object Identity and Equality

Understanding the difference between identity (`is`) and equality (`==`):

```python
a = [1, 2, 3]
b = [1, 2, 3]
c = a

print(a == b)  # True - same content
print(a is b)  # False - different objects
print(a is c)  # True - same object
```

!!! Warning "Integer Caching"
    Python caches small integers and frequently used strings for efficiency. This can lead to surprising identity behavior:

    ```python
    a = 5
    b = 5
    print(a is b)  # True - Python caches small integers

    x = 500
    y = 500
    print(x is y)  # Might be False - large integers may not be cached
    ```

## Shallow vs Deep Copy

When working with nested data structures, you need to understand the difference between shallow and deep copies.

```python
import copy

# Original nested list
original = [[1, 2, 3], [4, 5, 6]]

# Shallow copy - copies outer list but shares inner lists
shallow = copy.copy(original)
# or shallow = original[:]  # Slice creates a shallow copy

# Deep copy - recursively copies all nested objects
deep = copy.deepcopy(original)

# Modify inner list through original
original[0][0] = 'X'

print(original)  # Output: [['X', 2, 3], [4, 5, 6]]
print(shallow)   # Output: [['X', 2, 3], [4, 5, 6]] - affected by change
print(deep)      # Output: [[1, 2, 3], [4, 5, 6]] - independent copy
```

!!! Note "When to Use Each"
    - **Shallow Copy**: When you only need to modify the outer container without affecting the original
    - **Deep Copy**: When you need a completely independent copy of a nested structure

!!! Warning "Performance Considerations"
    Deep copying can be computationally expensive for large or deeply nested structures. Use it judiciously.

## Memory Management Basics

Python handles memory management automatically through reference counting and garbage collection.

### Reference Counting

```python
# Conceptual example of reference counting
x = [1, 2, 3]  # Reference count: 1
y = x          # Reference count: 2
del x          # Reference count: 1
del y          # Reference count: 0 (object is now eligible for garbage collection)
```

!!! Note "Garbage Collection"
    Python's garbage collector automatically reclaims memory from objects that are no longer reachable. This includes objects in reference cycles that reference counting alone can't handle.

```python
import gc

# Force garbage collection (rarely needed in practice)
gc.collect()
```

!!! Tip "Resource Management"
    Use context managers (with statements) for resources like files to ensure proper cleanup:

    ```python
    with open('example.txt', 'r') as file:
        data = file.read()
    # File is automatically closed when the block exits
    ```

## Advanced Collections

### Tuples

Tuples are immutable sequences similar to lists. Their immutability makes them useful for fixed collections of data.

```python
# Creating tuples
empty_tuple = ()
single_item = (1,)  # Note the comma
coordinates = (10, 20)

# Tuple unpacking
x, y = coordinates
print(x, y)  # Output: 10 20

# Multiple returns from functions use tuples
def get_dimensions():
    return (1920, 1080)

width, height = get_dimensions()

# Tuples as dictionary keys (lists cannot be keys)
locations = {
    (40.7128, -74.0060): "New York",
    (34.0522, -118.2437): "Los Angeles"
}
```

!!! Note "Tuple Advantages"
    - **Immutability**: Cannot be changed after creation
    - **Hashable**: Can be used as dictionary keys or set elements
    - **Performance**: Slightly more memory efficient than lists

### Sets

Sets are unordered collections of unique elements, optimized for membership testing and eliminating duplicates.

```python
# Creating sets
numbers = {1, 2, 3, 4, 5}
also_numbers = set([1, 2, 2, 3, 4])  # Duplicates are automatically removed
print(also_numbers)  # Output: {1, 2, 3, 4}

# Adding and removing elements
numbers.add(6)
numbers.remove(1)  # Raises KeyError if element not found
numbers.discard(10)  # No error if element not found

# Set operations
a = {1, 2, 3, 4}
b = {3, 4, 5, 6}

print(a.union(b))         # {1, 2, 3, 4, 5, 6}
print(a.intersection(b))  # {3, 4}
print(a.difference(b))    # {1, 2}
print(b.difference(a))    # {5, 6}
print(a.symmetric_difference(b))  # {1, 2, 5, 6}

# Set operations with operators
print(a | b)  # Union
print(a & b)  # Intersection
print(a - b)  # Difference
print(a ^ b)  # Symmetric difference

# Membership testing (very efficient)
print(3 in a)  # True
```

!!! Warning "Set Requirements"
    Set elements must be hashable (immutable). You cannot add lists or dictionaries to a set, but you can add tuples.

!!! Tip "Set Use Cases"
    - Removing duplicates from a collection
    - Membership testing for large collections
    - Mathematical set operations
    - Finding unique items across collections

### Named Collections

Named tuples provide a way to create simple classes with named fields.

```python
from collections import namedtuple

# Define a named tuple type
Point = namedtuple('Point', ['x', 'y'])

# Create instances
p1 = Point(10, 20)
p2 = Point(x=30, y=40)

# Access by name or position
print(p1.x, p1[0])  # Both ways work: 10 10

# Immutable but replaceable
p3 = p1._replace(x=15)
print(p3)  # Point(x=15, y=20)

# Conversion to dictionary
print(p1._asdict())  # {'x': 10, 'y': 20}

# Practical example: representing data records
Employee = namedtuple('Employee', 'name age department salary')
emp = Employee('Alice', 30, 'Engineering', 85000)
print(f"{emp.name} works in {emp.department}")
```

!!! Tip "When to Use Named Tuples"
    - When you need a lightweight immutable data container
    - When using a full class would be overkill
    - To make code more readable with named fields
    - For returning multiple values from functions with clear field names

