# Functions

## Objectives

- Functions
    - Function Definition
    - Parameters and Arguments
    - Return Values
    - Scope and Namespaces
    - Lambda Functions
    - Function Documentation
    - Type Hints (Introduction)
    - Best Practices

## Function Definition

Functions are the building blocks of reusable code. They allow you to break down complex tasks into smaller, manageable units, making your code more organized, readable, and easier to maintain.

```python
# Basic function definition
def greet(name):
    """Prints a greeting message."""
    print(f"Hello, {name}!")

# Function call
greet("Alice")  # Output: Hello, Alice!

# Function with multiple parameters
def calculate_rectangle_area(length, width):
    """Calculate the area of a rectangle."""
    area = length * width
    return area

# Calculate and print area
result = calculate_rectangle_area(5, 3)
print(f"The area is {result}")  # Output: The area is 15
```


## Parameters and Arguments

Parameters allow you to pass data into functions, making them flexible and adaptable. Understanding the different types of arguments gives you fine-grained control over how your functions receive and process data.

```python
# Positional Arguments
def describe_person(name, age):
    print(f"{name} is {age} years old.")

describe_person("Bob", 25)  # Output: Bob is 25 years old.

# Keyword Arguments
describe_person(age=30,*args (variable positional)  name="Charlie")  # Output: Charlie is 30 years old.

# Default Parameter Values
def greet(name, greeting="Hello"):
    print(f"{greeting}, {name}!")

greet("David")               # Output: Hello, David!
greet("Eve", "Good morning") # Output: Good morning, Eve!

# Variable-Length Arguments (*args)
def sum_numbers(*args):
    total = 0
    for num in args:
        total += num
    return total

print(sum_numbers(1, 2, 3))      # Output: 6
print(sum_numbers(4, 5, 6, 7))   # Output: 22

# Variable-Length Keyword Arguments (**kwargs)
def print_info(**kwargs):
    for key, value in kwargs.items():
        print(f"{key}: {value}")

print_info(name="Alice", age=30, city="New York")
# Output:
# name: Alice
# age: 30
# city: New York
```

!!! Warning "Mutable Default Arguments"
    Be careful with mutable default arguments like lists or dictionaries. They are created only once when the function is defined, not each time it's called.

    ```python
    # Incorrect way - list is shared between calls
    def append_to_list(value, my_list=[]):  # DON'T DO THIS!
        my_list.append(value)
        return my_list

    print(append_to_list(1))  # Output: [1]
    print(append_to_list(2))  # Output: [1, 2]  (Unexpected!)

    # Correct way
    def append_to_list(value, my_list=None):
        if my_list is None:
            my_list = []  # Create a new list each time
        my_list.append(value)
        return my_list
    ```

!!! Tip "Argument Order"
    When defining functions with multiple types of parameters, they must follow this order:

    1. Positional parameters
    2. *args (variable positional)
    3. Parameters with default values
    4. **kwargs (variable keyword)

## Return Values

Return values allow functions to send data back to the caller, enabling you to use the results of computations in other parts of your code.

```python
# Returning a single value
def add(x, y):
    return x + y

result = add(5, 3)
print(result)  # Output: 8

# Returning multiple values (as a tuple)
def calculate(x, y):
    sum_val = x + y
    diff_val = x - y
    return sum_val, diff_val

sum_result, diff_result = calculate(10, 4)  # Tuple unpacking
print(sum_result, diff_result)  # Output: 14 6

# Implicit None return
def do_something():
    print("Doing something...")
    # No return statement

result = do_something()
print(result)  # Output: None
```

!!! Note "Return Statement"
    The `return` statement immediately exits the function and returns the specified value(s) to the caller. A function without a `return` statement implicitly returns `None`.

## Scope and Namespaces

Understanding scope determines where variables are accessible within your code. Python uses namespaces to keep track of variables, and the scope rules (LEGB: Local, Enclosing, Global, Built-in) dictate how variables are resolved.

```python
x = 10  # Global variable

def my_function():
    x = 5  # Local variable (shadows the global x)
    print(f"Inside function: x = {x}")  # Output: Inside function: x = 5

my_function()
print(f"Outside function: x = {x}")  # Output: Outside function: x = 10

# Modifying global variables
y = 20  # Global

def modify_global():
    global y  # Tells Python to use the global y
    y = 30    # Modifies the global y
    print(f"Inside function: y = {y}")  # Output: Inside function: y = 30

modify_global()
print(f"Outside function: y = {y}")  # Output: Outside function: y = 30

# Nested functions and nonlocal
def outer_function():
    z = 15  # Enclosing scope variable

    def inner_function():
        nonlocal z  # Refers to z in the enclosing scope
        z = 25     # Modifies the z in outer_function's scope
        print(f"Inside inner_function: z = {z}")  # Output: Inside inner_function: z = 25

    inner_function()
    print(f"Inside outer_function: z = {z}")  # Output: Inside outer_function: z = 25

outer_function()
```

!!! Note "LEGB Rule"
    Python follows the LEGB rule to resolve variable names:

    1. **Local** - Variables defined within the current function
    2. **Enclosing** - Variables in the enclosing (parent) function's scope
    3. **Global** - Variables defined at the module level
    4. **Built-in** - Python's pre-defined names (like `print`, `len`)

!!! Warning "Use of global"
    Avoid using `global` excessively. It can make code harder to reason about and maintain. Pass variables as arguments and return values instead, whenever possible.

## Lambda Functions

Lambda functions (also called anonymous functions) are small, single-expression functions defined without a name. They are useful for short operations, particularly with functions like `map`, `filter`, and `sort`. Theses three functions take a function as an argument

??? Info "Function as an Argument"
    Functions can be passed as arguments to other functions. This is a powerful feature of Python that allows you to write more flexible and reusable code.

    ```python
    #map example
    numbers = [1, 2, 3, 4]


```python
# Regular function vs lambda equivalent
data = (1, 4)


# Using lambda with map()
numbers = [1, 2, 3, 4]
squared_numbers = list(map(lambda x: x**2, numbers))
print(squared_numbers)  # Output: [1, 4, 9, 16]

# Using lambda with filter()
even_numbers = list(filter(lambda x: x % 2 == 0, numbers))
print(even_numbers)     # Output: [2, 4]

# Using lambda with sort()
points = [(1, 5), (3, 2), (2, 8)]
sorted_points = sorted(points, key=lambda point: point[1])  # Sort by y-coordinate
print(sorted_points)    # Output: [(3, 2), (1, 5), (2, 8)]
```

!!! Tip "Lambda Limitations"
    Lambda functions are limited to a single expression. For complex logic, use regular functions for better readability.

## Function Documentation

Docstrings (documentation strings) are multiline strings used to document Python code. They explain what a function does, its parameters, and its return values.

```python
def calculate_area(length, width):
    """Calculates the area of a rectangle.

    Args:
        length: The length of the rectangle (in meters).
        width: The width of the rectangle (in meters).

    Returns:
        The area of the rectangle (in square meters).

    Raises:
        TypeError: If length or width is not a number.
        ValueError: If length or width is negative.
    """
    if not isinstance(length, (int, float)) or not isinstance(width, (int, float)):
        raise TypeError("Length and width must be numbers.")
    if length < 0 or width < 0:
        raise ValueError("Length and width must be non-negative.")
    return length * width

# Accessing docstrings
print(calculate_area.__doc__)  # Prints the docstring
help(calculate_area)           # Displays the docstring in a more readable format
```

=== "Google Style Docstring"
    ```python
    def my_function(param1, param2):
        """Summary line.

        Extended description of function.

        Args:
            param1 (int): Description of param1.
            param2 (str): Description of param2.

        Returns:
            bool: Description of return value.
        """
        return True
    ```

=== "reStructuredText Style"
    ```python
    def another_function(arg1, arg2):
        """
        Summary line.

        :param arg1: Description of arg1.
        :type arg1: int
        :param arg2: Description of arg2.
        :type arg2: str
        :raises ValueError: If something goes wrong.
        :returns: A tuple containing arg1 and arg2.
        :rtype: tuple
        """
        if arg1 < 0:
            raise ValueError("arg1 must be positive")
        return arg1, arg2
    ```

!!! Success "Docstring Best Practices"
    - Write docstrings for all public functions, classes, and modules
    - Follow a consistent docstring style (e.g., Google or reStructuredText)
    - Keep docstrings up-to-date with your code
    - PEP 257 describes docstring conventions

## Type Hints

Type hints allow you to specify the expected data types of function arguments and return values. While Python is dynamically typed, type hints provide valuable information for code readability and static analysis.

```python
from typing import List, Dict, Tuple, Optional, Union

def greet(name: str) -> None:
    """Greets a person."""
    print(f"Hello, {name}!")

def add(x: int, y: int) -> int:
    """Adds two integers."""
    return x + y

def process_data(data: List[int]) -> Tuple[float, float]:
    """Calculates the mean and standard deviation of a list of numbers."""
    mean = sum(data) / len(data)
    # Simplified std dev calculation for example
    std = 0.0
    return mean, std

def get_value(key: str, data: Dict[str, int]) -> Optional[int]:
    """Retrieves a value from a dictionary.
       Returns None if the key is not found.
    """
    return data.get(key)

# Using Union for multiple possible types
def convert_to_int(value: Union[str, float]) -> int:
    """Converts a string or float to an integer."""
    return int(value)
```

!!! Note "Type Hints"
    Type hints are not enforced at runtime. They are primarily for static analysis tools like MyPy to catch type errors during development.

## Best Practices

Following these best practices when writing functions leads to code that is more readable, maintainable, and robust.

!!! Success "Function Best Practices"
    - **Single Responsibility**: Each function should do one thing well
    - **Descriptive Names**: Function names should clearly indicate what the function does
    - **Small Functions**: Keep functions short and focused
    - **Clear Docstrings**: Document the purpose, arguments, and return values
    - **DRY (Don't Repeat Yourself)**: Avoid duplicating code
    - **Limit Side Effects**: Functions should ideally not modify external state
    - **Parameter Order**: Place most commonly used parameters first
    - **Return Values**: Be consistent in what your functions return

```python
# Bad example: Function does too many things
def process_data(filename):
    with open(filename, 'r') as f:
        data = f.read()
    data = data.lower()
    words = data.split()
    word_count = len(words)
    return word_count

# Better: Breaking it into smaller functions with single responsibilities
def read_file(filename):
    """Read the contents of a file."""
    with open(filename, 'r') as f:
        return f.read()

def count_words(text):
    """Count the number of words in a text."""
    words = text.lower().split()
    return len(words)

# Usage
text = read_file('example.txt')
count = count_words(text)
```

## Check on Learning / Challenges

!!! Warning "Complete challenges on your workstation and dev environment."

- Today's challenges can be found at `~/170D/cybbh/challenges/py_1_2/01_functions/{challenge_number}`
- All challenge prompts are contained in their respective `main.py` file as comments.
- The `solution.py` file exists primarily as a last resort after you've exhausted getting help from the instructor, a classmate, or the web.