# Exception Handling

## Objectives

- Error Handling
    - Traditional vs. Modern Approaches
    - Exception Hierarchy
    - Try/Except/Else/Finally
    - Raising Exceptions
    - Context Managers
    - Debugging Techniques

## Error Handling

Up to this point, you've likely encountered errors in your Python code. These errors, formally called "exceptions," disrupt the normal flow of your program's execution.
Effective error handling is crucial for creating robust and user-friendly applications. It allows your program to gracefully handle unexpected situations, provide informative feedback, and potentially recover from errors without crashing.

## Traditional vs. Modern Approaches

Before diving into Python's exception handling mechanisms, it's helpful to understand how error handling was traditionally done and why modern approaches are preferred.

**Problems with Traditional Error Handling:**

- **Return Codes:** Functions would often use return codes (e.g., integers) to indicate success or failure, and different codes would represent different error conditions.
- **Scattered Checks:** The calling code would need to check these return codes after *every* function call, leading to scattered `if/else` blocks throughout the code.
- **Muddled Logic:** The core logic of the program would be intermingled with numerous error checks, making it harder to read and understand.
- **Decentralized Handling:** There was no central place to manage error handling; it was spread throughout the codebase.

```python
# Example 1 of traditional error handling
def divide_numbers(a, b):
    if b == 0:
        return "Error: Division by zero"  # Return an error string
    return a / b

# Example 2 of traditional error handling
return_code = call_function_that_does_something()
if return_code == val1:
    # Handle error val1
    ...
elif return_code == val2:
    # Handle error val2
    ...

other_return_code = call_another_function()
if other_return_code == val3:
    # Handle error val3
    ...
elif other_return_code == val4:
    # Handle error val4
    ...
```

**Modern Exception Handling Approach:**

- **`try/except` Blocks:** Code that might cause errors is placed within a `try` block. `except` blocks follow, containing code to handle specific types of exceptions.
- **Exception Objects:** When an error occurs, the runtime generates an *exception object* that contains information about the error.
- **Centralized Handling:** Error handling is typically concentrated in the `except` blocks, separating it from the main program logic.
- **Type-Based Handling:** `except` blocks can be specific to different exception *types*, allowing for tailored error handling.

**Benefits of Exception Handling:**

- **Cleaner Code:** Separates error handling from the main logic, making code more readable.
- **Easier Management:** Centralizes error handling, making it easier to maintain and modify.
- **Custom Exceptions:** Allows you to define your own exception types to represent specific error conditions in your application.
- **Programmer Control:** Gives you precise control over how errors are handled.
- **Language Support:** Widely supported by modern programming languages, and enforced in some (like Java).
- **Enforced Handling:** In some languages like Java, exception handling isn't just good practice, it's *required* for certain types of errors (checked exceptions). This forces developers to consider error handling, leading to more robust code.

## Exception Hierarchy

Python exceptions are organized in a hierarchy, with a base `BaseException` class at the top. Most exceptions you'll commonly interact with inherit from the `Exception` class (which itself inherits from `BaseException`).
 Understanding this hierarchy helps you choose the appropriate exception type to catch and create. It is important to know that an `except` clause that catches `exc-type` catches all types of `exc-type` **and any child of `exc-type`**.

For a full list of exceptions, see the [Python Exception Hierarchy](https://docs.python.org/3/library/exceptions.html#exception-hierarchy).

Here's a simplified view of some common exceptions and their relationships:

```
BaseException
 +-- SystemExit
 +-- KeyboardInterrupt
 +-- GeneratorExit
 +-- Exception
      +-- ArithmeticError
      |    +-- FloatingPointError
      |    +-- OverflowError
      |    +-- ZeroDivisionError
      +-- AttributeError
      +-- ImportError
      |    +-- ModuleNotFoundError
      +-- LookupError
      |    +-- IndexError
      |    +-- KeyError
      +-- NameError
      |    +-- UnboundLocalError
      +-- OSError
      |    +-- FileExistsError
      |    +-- FileNotFoundError
      +-- SyntaxError
      |    +-- IndentationError
      |         +-- TabError
      +-- TypeError
      +-- ValueError
      |    +-- UnicodeError
      +-- Warning
           +-- RuntimeWarning
           +-- SyntaxWarning
           +-- ImportWarning
```

!!! Warning "When to catch an Exception"
    - It's tempting to use exception handling for *every* error. However, this is not the best practice.
    - Their intent is to "signal" **Exceptional** circumstances.
    - If an error is expected, for example, a user entering an invalid email address, it's better to handle it with control flow statements like `if/else/while` etc.

**Key Exceptions to Know:**

- `Exception`: The base class for most exceptions. Catching this is very broad. (So don't do it)
- `TypeError`: Raised when an operation or function is applied to an object of inappropriate type.
- `ValueError`: Raised when a function receives an argument of the correct type but an inappropriate value.
- `IndexError`: Raised when a sequence subscript (index) is out of range.
- `KeyError`: Raised when a dictionary key is not found.
- `FileNotFoundError`: Raised when a file or directory is requested but doesn't exist.
- `ZeroDivisionError`: Raised when dividing by zero.
- `NameError`: Raised when a local or global name is not found.
- `AttributeError`: Raised when an attribute reference or assignment fails.
- `ImportError` / `ModuleNotFoundError`: Raised when an imported module cannot be found or loaded.

## Try/Except/Else/Finally

The core of exception handling in Python is the `try/except` block. This structure allows you to:

1. `try`: Attempt to execute a block of code that might raise an exception.
2. `except`: Handle specific exceptions that occur within the `try` block. You can have multiple `except` blocks to catch different types of exceptions.
3. `else`: (Optional) Execute a block of code if *no* exceptions occurred in the `try` block.
4. `finally`: (Optional) Execute a block of code *regardless* of whether an exception occurred or was handled. This is often used for cleanup operations (e.g., closing files).

**Basic Syntax:**

```python
try:
    # Code that might raise an exception
    result = 10 / 0  # This will raise a ZeroDivisionError
    print("This will not run if there is an error.")
except ZeroDivisionError:
    # Handle the ZeroDivisionError
    print("Cannot divide by zero!")
else:
    # Execute if no exception occurred
    print("Division successful. Result:", result)
finally:
    # Always execute this block
    print("This always runs, even if there was an error.")

print("Program continues...")
```

**Multiple `except` Blocks:**

```python
def get_list_element(my_list, index):
    try:
        return my_list[index]
    except IndexError:
        print("Index out of range.")
        return None
    except TypeError:
        print("Invalid index type. Must be an integer.")
        return None
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        return None

my_list = [1, 2, 3]
print(get_list_element(my_list, 5))
# Output: Index out of range. None
print(get_list_element(my_list, "a"))
# Output: Invalid index type. Must be an integer. None
print(get_list_element(my_list, 0))
# Output: 1
print(get_list_element(1, 0))
# Output: An unexpected error occurred:
# 'int' object is not subscriptable None
```

**`else` Clause:**

The `else` clause is useful for code that should only run if the `try` block was successful.

```python
try:
    file = open("my_file.txt", "r")
    data = file.read()
except FileNotFoundError as fnf_error:
    print(f"File not found. {fnf_error.filename}")
    print(f"Error Number: {fnf_error.errno}")
else:
    print("File contents:", data) # Only process data if the file was opened
finally:
    if "file" in locals() and not file.closed: # Important check here!
        file.close() # cleanup the file
```

**`finally` Clause:**

The `finally` clause is crucial for ensuring that resources are released, even if errors occur.

```python
try:
    file = open("my_file.txt", "w")
    file.write("Some data")
    # Simulate an error
    1 / 0
except ZeroDivisionError:
    print("A division by zero error occurred!")
finally:
    file.close()  # File is closed even if the ZeroDivisionError happens
    print("File closed.")
```

!!! Warning "A word on Finally"
    Even if an exception is raised within an `except` or `else` block, the `finally` block will *still* execute.
    However, if an exception is unhandled (no matching `except` block), the program will still terminate *after* the `finally` block runs.

## Aliasing Exceptions and Accessing Attributes

It's highly recommended to alias specific exceptions using `as <variable_name>` in your `except` blocks. This gives you access to the exception object and its attributes, which often provide valuable information about the error.

```python
try:
    # ... code that might raise FileNotFoundError ...
    file = open("nonexistent_file.txt", "r")
except FileNotFoundError as e:
    print(f"File not found: {e.filename}")  # Access the filename attribute
    print(f"Error number: {e.errno}")  # Access the error number

# Example with logging
import logging

try:
    # ... some code that could produce a ValueError
    value = int("abc")
except ValueError as e:
    logging.error(f"ValueError occurred: {e}")
```

**Benefits of Aliasing Exceptions:**

- **Accessing Exception Attributes:** Specific exception types often have attributes that provide additional information about the error.
- **Logging and Reporting:** Aliasing the exception makes your logging or error reporting more informative.
- **Conditional Handling within `except`:** You might want to perform different actions within the `except` block based on the exception's properties.

## Why Continue Running After a `try` Block Fails?

Exception handling in Python isn't just about stopping your program when something goes wrong. It's about *controlling* how your program responds to errors. While some errors are fatal, many are not, and exception handling lets you handle these non-fatal errors gracefully.

Here are some scenarios where continuing execution after a `try` block fails is desired:

- **Non-Critical Errors:** Many errors aren't fatal. The user enters a bad filename, a network request times out, or an optional feature fails. You can handle these gracefully, perhaps by prompting the user for different input, retrying the operation, or logging the error and continuing.

- **Resource Cleanup:** Even if you can't continue the main logic, you *must* ensure resources are released (files closed, network connections terminated, etc.). The `finally` block guarantees this.

- **Error Recovery:** Sometimes, you can actually *recover* from an error. You might retry a network request a few times before giving up, or use default data if a file can't be loaded.

- **Looping Constructs:** Within loops, you might want to handle errors that occur during individual iterations *without* stopping the entire loop.

- **Fault Tolerance:** In larger systems, you want to isolate failures. Exception handling helps prevent a single error from crashing the entire application.

```python
# Example: Non-critical Error (file not found)
filename = input("Enter filename: ")
try:
    with open(filename, "r") as f:
        contents = f.read()
    print(contents)
except FileNotFoundError:
    print(f"File '{filename}' not found. Using default data.")
    contents = "Default data"  # Continue with default data
```

```python
# Example: Resource cleanup
try:
    file = open("my_file.txt", "w")
    # ... code that might raise an exception ...
    file.write("some data")
except Exception as e:
    print(f"An error occurred: {e}")
finally:
    file.close()  # ALWAYS close the file, even if there was an error
```

```python
# Example: Error Recovery (retrying a network request)
import time

def get_data(attempts=3):
    for attempt in range(attempts):
        try:
            # ... code to fetch data from a network ...
            # Simulate success/failure
            if attempt < 2:  # Fail the first two attempts
                raise TimeoutError("Network timeout")
            return "Successfully fetched data"
        except TimeoutError as e:
            print(f"Attempt {attempt + 1} failed: {e}")
            time.sleep(1)  # Wait before retrying
    print("Failed to get data after multiple attempts.")
    return None  # Or raise a custom exception

data = get_data()
if data:
    print(data)
```

```python
# Example: Looping constructs
data_list = ["1", "2", "a", "4", "b", "5"]
results = []
for item in data_list:
    try:
        num = int(item)
        results.append(num * 2)
    except ValueError as e:
        print(f"Skipping invalid item '{item}': {e}")

print(results)  # Output: [2, 4, 8, 10]
```

## Raising Exceptions

So far, we've focused on *handling* exceptions raised by the Python runtime or by libraries. However, you can also *raise* exceptions yourself using the `raise` statement. This is how you signal errors or exceptional conditions within *your* code.

**Basic Syntax:**

```python
raise ExceptionType("Error message")
```

- `ExceptionType`: The type of exception you want to raise (e.g., `ValueError`, `TypeError`, `MyCustomException`).
- `"Error message"`: (Optional) A string describing the error. This message will be part of the exception object.

**Example:**

```python
def validate_age(age):
    if not isinstance(age, int):
        raise TypeError("Age must be an integer.")
    if age < 0:
        raise ValueError("Age cannot be negative.")
    if age > 120:
        raise ValueError("Age cannot be greater than 120.")
    print("Age is valid:", age)


try:
    validate_age(25)  # Valid
    validate_age(-5)  # Raises ValueError
    validate_age("abc")  # Raises TypeError
except ValueError as e:
    print("ValueError:", e)
except TypeError as e:
    print("TypeError:", e)
```

**Re-raising Exceptions:**

You can re-raise the current exception using `raise` without any arguments inside an `except` block. This allows you to perform some handling (like logging) and then pass the exception up the call stack.

```python
def process_data(filename):
    try:
        with open(filename, 'r') as file:
            data = file.read()
            result = parse_data(data)
            return result
    except FileNotFoundError as e:
        # Log the error but re-raise it since we can't recover
        logger.error(f"Missing input file: {filename}")
        raise  # Re-raises the FileNotFoundError
    except ValueError as e:
        # Transform this exception into a more specific one
        # This is different from just re-raising
        logger.warning(f"Data format issue in {filename}: {str(e)}")
        raise DataFormatError(f"Invalid format in {filename}: {str(e)}") from e
    except Exception as e:
        # Log unexpected errors with full details before re-raising
        logger.critical(f"Unexpected error processing {filename}", exc_info=True)
        # We might add context to the logs but still want the original error to propagate
        raise  # Re-raise whatever exception occurred
```

In this example:

1. We're handling a file processing operation which could fail in multiple ways
2. For `FileNotFoundError`, we log the specific issue but re-raise it, letting the caller decide how to handle missing files
3. For `ValueError`, we transform it into a more domain-specific exception while preserving the original cause with `from e`
4. For unexpected exceptions, we log detailed information (with traceback) before re-raising to ensure the error still propagates up the stack

The key benefit of re-raising exceptions is that it lets you:
- Add logging or diagnostic information at different levels of your application
- Perform partial recovery actions or cleanup
- Maintain the original error context and stack trace
- Let higher-level code decide whether and how to recover

Without re-raising, you'd either have to handle the error completely at this level or lose valuable error context when creating new exceptions.

## Understanding Stack Traces

When an exception is not handled, Python prints a stack trace showing the function calls that led to the error. This is extremely useful for debugging.


```python
def func_bad():
    # Divide by zero to raise an exception
    a_num = 100 / 0  # The exception originates here
    print(f'{a_num}')  # This line never executes

def func_call_bad():
    # Call above function to see Python backtracking
    print("About to call func_bad()")
    func_bad()  # When exception occurs, execution here is interrupted
    print("This line never runs")  # This line never executes

def func_call_func_call_bad():
    # Call above function to see Python backtracking
    print("About to call func_call_bad()")
    func_call_bad()  # When exception occurs, execution here is interrupted
    print("This line never runs")  # This line never executes

def main():
    # Python will go through the stack trace
    # looking for an except block
    try:
        print("About to call func_call_func_call_bad()")
        func_call_func_call_bad()  # When exception occurs, execution here is interrupted
        print("This line never runs")  # This line never executes
    except ZeroDivisionError as e:
        # If we uncomment this block, the stack trace wouldn't be printed
        # Instead, this handler would catch the exception
        # print(f"Caught an error: {e}")

        raise  # Re-raise the exception

print("Program starts")
main()
print("This line never runs")  # This line never executes

# Output:
"""
Program starts
About to call func_call_func_call_bad()
About to call func_call_bad()
About to call func_bad()
Traceback (most recent call last):
  File "/home/kmc/repo/delta-students.cybbh/temp.py", line 33, in <module>
    main()
  File "/home/kmc/repo/delta-students.cybbh/temp.py", line 23, in main
    func_call_func_call_bad()  # When exception occurs, execution here is interrupted
  File "/home/kmc/repo/delta-students.cybbh/temp.py", line 15, in func_call_func_call_bad
    func_call_bad()  # When exception occurs, execution here is interrupted
  File "/home/kmc/repo/delta-students.cybbh/temp.py", line 9, in func_call_bad
    func_bad()  # When exception occurs, execution here is interrupted
  File "/home/kmc/repo/delta-students.cybbh/temp.py", line 3, in func_bad
    a_num = 100 / 0  # The exception originates here
ZeroDivisionError: division by zero
"""
```

## How Stack Traces Work in Python

1. **Call Stack Formation**:
    - When you call a function, Python adds it to the "call stack"
    - Each function call creates a new "frame" on the stack with its own local variables
    - When a function completes, its frame is removed from the stack

2. **Exception Propagation**:
    - When an exception occurs (like our division by zero), Python immediately stops execution at that point
    - Python then starts "unwinding" the stack, looking for an exception handler (try/except block)
    - It moves up the stack frame by frame, from the most recently called function back toward the main program

3. **Stack Trace Anatomy**:
    - The traceback shows the path of execution in reverse chronological order (most recent call last)
    - Each line shows a frame in the call stack with:
        - The file name and line number
        - The function name
        - The actual code that was executing when the exception occurred or was passed through

4. **Important Stack Trace Details**:
    - The bottom frame shows where the exception actually occurred (`a_num = 100 / 0`)
    - The frames above show the "path" that led to the error
    - Without a matching `except` block, Python terminates the program and displays the full trace
    - With a matching `except` block, execution continues in the handler and the trace isn't displayed (unless explicitly requested)

5. **Preserving Stack Traces**:
    - When re-raising exceptions with `raise` (no arguments), the original stack trace is preserved
    - When creating new exceptions with `raise NewException from original_exception`, the cause relationship is preserved
    - You can manually capture and print stack traces with the `traceback` module

This understanding helps with debugging because you can follow the execution path backward from the error to see exactly how your program reached the problematic state.

!!! Danger "Production Code Guidelines for 17D/170D course"
    - Under NO circumstances should a user of the application see a stack trace!
    - Always catch exceptions at an appropriate level and provide user-friendly error messages.

## When to Use Exceptions (and When Not To)

It's important to use exceptions appropriately in your code. Here are some guidelines:

**Rule 1: Use Exceptions for Truly Exceptional Circumstances**

- Use exceptions when something *unexpected* has happened, something that prevents your code from continuing normally.
- Examples: file not found, network connection dropped, database error.
- Don't use for expected user errors like invalid form input - use `if/else` for these.

**Rule 2: Don't Use Exceptions for Normal Control Flow**

- Don't use exceptions to handle situations that you *expect* to happen as part of the normal operation of your program.
- Anti-pattern example:

```python
# BAD: Using exceptions for control flow
def get_user_choice(options):
    try:
        choice = int(input("Enter your choice: "))
        if choice < 1 or choice > len(options):
            raise ValueError
        return options[choice - 1]
    except ValueError:
        print("Invalid choice. Please try again.")
        return get_user_choice(options)  # Recursive call on invalid input

# BETTER: Using if/else for control flow
def get_user_choice_better(options):
    while True:
        try:
            choice = int(input("Enter your choice: "))
        except ValueError:
            print("Invalid Input")
            continue

        if 1 <= choice <= len(options):
            return options[choice - 1]
        else:
            print("Invalid choice. Please try again.")
```

**Rule 3: Favor Specific Exceptions Over Broad Exceptions**

- Catch the most specific exception type that you can reasonably handle.
- Avoid catching `Exception` unless you have a very good reason (and you almost always re-raise it or log it).
- Good: `except FileNotFoundError:`, `except ValueError:`
- Bad: `except:` (bare except), `except Exception:` (use only as a last resort)

**Rule 4: Don't Use Exceptions to Validate *Expected* User Input Errors**

- Use `if/else` statements (and possibly loops) to validate user input, not exceptions.
- Invalid user input is a normal occurrence, not an exceptional circumstance.

**Rule 5: Use Exceptions When Failure is Unrecoverable (at that level)**

- If a function cannot recover from an error and continue its intended purpose, it should raise an exception.

## Real-World Examples

Here are practical examples of exception handling at different complexity levels:

### Basic: File Handling

```python
filename = input("Enter the filename: ")

try:
    with open(filename, "r") as file:
        contents = file.read()
        print(contents)
except FileNotFoundError:
    print(f"Error: File '{filename}' not found.")
except Exception as e:
    print(f"An unexpected error occurred: {e}")
```

### Intermediate: Network Request with Retries

```python
import time
import random  # Simulate network issues

class NetworkError(Exception):
    pass

class TimeoutError(NetworkError):
    pass

class ConnectionError(NetworkError):
    pass

def fetch_data(url, max_retries=3):
    for attempt in range(max_retries):
        try:
            # Simulate a network request
            print(f"Attempting to fetch data from {url} (attempt {attempt + 1})...")
            time.sleep(1)  # Simulate network latency
            if random.random() < 0.3:  # 30% chance of TimeoutError
                raise TimeoutError("Request timed out.")
            if random.random() < 0.1:  # 10% chance of ConnectionError
                raise ConnectionError("Connection refused.")

            # Simulate successful data retrieval
            data = f"Data from {url}"
            return data

        except TimeoutError as e:
            print(f"TimeoutError: {e}. Retrying...")
        except ConnectionError as e:
            print(f"ConnectionError: {e}. Retrying...")
        except NetworkError as e: # handles any NetworkError
            print(f"NetworkError {e}. Retrying...")
        except Exception as e:
             print(f"An unexpected error has occurred {e}")
    print(f"Failed to fetch data from {url} after {max_retries} attempts.")
    return None  # Or raise a custom exception

data = fetch_data("https://example.com")
if data:
    print(f"Successfully fetched data: {data}")
```

### Advanced: Database Interaction with Rollback

```python
import logging

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")

class DatabaseError(Exception):
    pass

class Database:
    def __init__(self):
        self.data = {}
        self.in_transaction = False
        self.backup = {}

    def begin_transaction(self):
        self.in_transaction = True
        self.backup = self.data.copy()  # Create a backup
        logging.info("Transaction started.")

    def commit(self):
        if not self.in_transaction:
            raise DatabaseError("Not in a transaction.")
        self.in_transaction = False
        self.backup = {}
        logging.info("Transaction committed.")

    def rollback(self):
        if not self.in_transaction:
            raise DatabaseError("Not in a transaction.")
        self.data = self.backup.copy()  # Restore from backup
        self.in_transaction = False
        self.backup = {}
        logging.info("Transaction rolled back.")

    def set_value(self, key, value):
        if not self.in_transaction:
            raise DatabaseError("Not in a transaction. Must start transaction")
        self.data[key] = value

    def get_value(self, key):
        return self.data.get(key)

db = Database()

try:
    db.begin_transaction()
    db.set_value("name", "Alice")
    db.set_value("age", 30)

    # Simulate an error
    if db.get_value("age") > 25:
        raise ValueError("Age is too high!")

    db.commit()
except ValueError as e:
    logging.error(f"Data validation error: {e}")
    db.rollback()
except DatabaseError as e:
    logging.error(f"Database error: {e}")
    db.rollback() # even if we aren't in a transaction.  it will be handled by the rollback() method.
except Exception as e:
    logging.exception("An unexpected error occurred:")  # Logs traceback
    if db.in_transaction: #must check
        db.rollback()

print(db.data) # Output: {} if rollback, or the new data if commit.
```

## Summary and Key Takeaways

Exception handling is a fundamental aspect of writing robust and maintainable Python code. Here's a summary of the key concepts:

- **Traditional vs. Modern:** Modern exception handling (`try/except`) is superior to traditional error handling (return codes) because it separates error handling from the main program logic, leading to cleaner, more manageable code.
- **Exception Hierarchy:** Python exceptions are organized in a hierarchy, with `BaseException` at the top and most common exceptions inheriting from `Exception`. Understanding this hierarchy is important for choosing the right exceptions to catch.
- **`try/except/else/finally`:**
    - `try`: Encloses code that might raise an exception.
    - `except`: Handles specific exception types. Use `except ExceptionType as e:` to access the exception object. Multiple `except` blocks can handle different exception types.
    - `else`: Executes if *no* exception occurred in the `try` block.
    - `finally`: *Always* executes, regardless of whether an exception occurred, for cleanup operations.
- **Raising Exceptions:** Use the `raise` statement to signal errors within your own code. `raise ExceptionType("Message")`.
- **Re-raising Exceptions:** Use a bare `raise` inside an `except` block to re-raise the current exception after performing some local handling (like logging).
- **When to use Exceptions (and when NOT to):**
    - **Do:** Use exceptions for *truly exceptional* circumstances, unexpected errors that prevent your code from continuing its normal operation (e.g. file not found, network connection lost).
    - **Don't:** Use exceptions for *normal* control flow, or to handle *expected* user input errors. Use `if/else` for validation and expected variations.
By mastering these concepts, you'll be well-equipped to write Python code that handles errors gracefully, provides informative feedback, and is easier to debug and maintain.
