# File System Operations

## Objectives

- Path Management
    - os.path
    - pathlib
- File Operations
    - Context Managers
    - Reading/Writing
    - Binary Files
    - Buffering
    - Seek and Tell
- Directory Operations
- File System Navigation
- Error Handling with File Operations

## Path Management

Python provides two main modules for working with file paths: the older `os.path` (string-based) and the newer `pathlib` (object-oriented). Using the appropriate path management functions is essential for creating cross-platform code that works correctly across different operating systems.

### os.path

The `os.path` module provides functions for manipulating file paths as strings.

```python
import os

# Joining paths
path = os.path.join("folder1", "folder2", "file.txt")
print(path)  # Output: folder1/folder2/file.txt

# Getting the absolute path
abs_path = os.path.abspath("file.txt")
print(abs_path)  # Output: /current/working/directory/file.txt

# Checking if a path exists
print(os.path.exists("file.txt"))  # Output: True or False
print(os.path.exists("folder1"))   # Output: True or False

# Checking if a path is a file or a directory
print(os.path.isfile("file.txt"))  # Output: True or False
print(os.path.isdir("folder1"))    # Output: True or False

# Getting the basename (filename) and dirname (directory)
basename = os.path.basename("/path/to/file.txt")
print(basename)  # Output: file.txt
dirname = os.path.dirname("/path/to/file.txt")
print(dirname)   # Output: /path/to

# Splitting a path into its components
print(os.path.split("/path/to/file.txt"))  # Output: ('/path/to', 'file.txt')
print(os.path.splitext("file.txt"))        # Output: ('file', '.txt')
```

### pathlib

The `pathlib` module provides a more object-oriented approach to file path manipulation. It represents paths as objects instead of strings.

```python
from pathlib import Path

# Creating Path objects
# YES this is legal, but it's confusing and not recommended.
path = Path("folder1") / "folder2" / "file.txt"
path = Path.cwd().joinpath("folder1", "folder2", "file.txt")
print(path)  # Output: folder1/folder2/file.txt

# Getting the absolute path
abs_path = path.absolute()
print(abs_path)  # Output: /current/working/directory/folder1/folder2/file.txt

# Checking if a path exists
print(path.exists())    # Output: True or False

# Checking if a path is a file or a directory
print(path.is_file())   # Output: True or False
print(path.parent.is_dir())  # Output: True or False

# Getting the name, stem, and suffix
print(path.name)    # Output: file.txt
print(path.stem)    # Output: file
print(path.suffix)  # Output: .txt

# Getting the parent directory
print(path.parent)   # Output: folder1/folder2

# Creating a new file
new_file = Path("new_file.txt")
new_file.touch()  # Creates the file if it doesn't exist

# Iterating through files in a directory
for file in Path(".").iterdir():
    print(file)
```

!!! Note "Comparing `os.path` and `pathlib`"
    - `pathlib` is generally preferred for new code due to its more intuitive interface
    - The `/` operator in `pathlib` provides a natural way to join paths (Still not recommended)
    - `pathlib` methods can be chained together for more readable code
    - `os.path` functions work directly with strings, while `pathlib` treats paths as objects

## File Operations

Python provides functions for reading from and writing to files. Using the right file mode and understanding how context managers work is critical for proper file handling.

### Context Managers (with statement)

Context managers automatically handle resources like file handles, ensuring they're properly closed even if errors occur.

```python
# Without context manager (manual cleanup)
try:
    file = open("my_file.txt", "r")
    data = file.read()
    print(data)
except FileNotFoundError:
    print("File not found.")
finally:
    if "file" in locals() and not file.closed:
        file.close()

# With context manager (automatic cleanup)
with open("my_file.txt", "r") as file:
    data = file.read()
    print(data)
# File is automatically closed here
```

!!! Warning "File Resource Management"
    Always use context managers (`with` statements) when working with files to ensure they're properly closed, even if exceptions occur. Failing to close files can lead to resource leaks and data corruption.

??? Tip "Creating Custom Context Managers"
    You can create your own context managers by defining a class with `__enter__` and `__exit__` methods:

    ```python
    class MyContextManager:
        def __enter__(self):
            print("Acquiring resource")
            return "resource"

        def __exit__(self, exc_type, exc_value, traceback):
            print("Releasing resource")
            # Return True to suppress exceptions, False to propagate them
            return False
    ```

### Reading/Writing Files

```python
# Writing to a file
with open("my_file.txt", "w") as f:
    f.write("Hello, world!\n")
    f.write("This is a new line.\n")

# Reading from a file
# read() reads the entire file into a single string
with open("my_file.txt", "r") as f:
    contents = f.read()
    print(contents)

# Reading line by line
with open("my_file.txt", "r") as f:
    for line in f:  # Iterating line by line
        print(line.strip())  # .strip() removes leading/trailing whitespace

# Reading all lines into a list
with open("my_file.txt", 'r') as f:
    lines = f.readlines()
    print(lines)  # Output: ['Hello, world!\n', 'This is a new line.\n']

# Appending to a file
with open("my_file.txt", "a") as f:
    f.write("Another line appended.\n")

# Using 'r+' (read and write)
# r+ opens the file for both reading and writing
with open("my_file.txt", "r+") as f:
    data = f.read()
    f.write("Adding content at the end")
```

!!! Note "File Modes"
    - `'r'`: Read (default)
    - `'w'`: Write (creates a new file or overwrites existing)
    - `'a'`: Append (adds to the end of an existing file)
    - `'x'`: Create (creates a new file, fails if it exists)
    - `'b'`: Binary mode (e.g., `'rb'`, `'wb'`)
    - `'t'`: Text mode (default)
    - `'+'`: Read and write (e.g., `'r+'`, `'w+'`)

### Binary Files

Binary files store data in the same format as it appears in memory, without any encoding/decoding.

```python
# Writing binary data
with open("binary_file.bin", "wb") as f:
    # "Hello" in hex
    f.write(b"\x48\x65\x6c\x6c\x6f")  # Writing raw bytes
    # or
    f.write(b"Hello")  # Writing a string

# Reading binary data
with open("binary_file.bin", "rb") as f:
    data = f.read()
    print(data)  # Output: b'\x48\x65\x6c\x6c\x6f'
```

!!! Warning "Data Serialization"
    This topic will be covered in more detail in Python 2.

### Buffering

Buffering optimizes I/O operations by reading or writing data in chunks (buffers) rather than byte by byte. This significantly improves performance, especially for large files.

```python
# No buffering (binary mode only)
with open("data.bin", "wb", buffering=0) as f:
    f.write(b"data")  # Writes immediately

# Line buffering (text mode)
with open("log.txt", "w", buffering=1) as f:
    f.write("Log entry 1\n")  # Flushes after the newline
    f.write("Log entry 2")    # Doesn't flush yet
    f.write("\n")             # Flushes

# Block buffering with explicit size
with open("large_file.dat", "wb", buffering=8192) as f:  # 8KB buffer
    for i in range(10000):
        f.write(b"chunk")  # Writes in 8KB chunks

# Forcing a buffer flush
with open("output.txt", "w") as f:
    f.write("Some data")
    f.flush()  # Force writing the buffer to disk, even if it's not full
```

!!! Note "Types of Buffering"
    - **No Buffering:** Each read/write operation directly interacts with the underlying device (very slow)
    - **Line Buffering:** Output is flushed when a newline character is encountered (common for interactive terminals)
    - **Block Buffering:** Data is read/written in fixed-size blocks (most common for files)

!!! Tip "Buffering Control"
    The `open()` function's `buffering` parameter lets you control buffering behavior:
    - `buffering=0`: No buffering (only allowed in binary mode)
    - `buffering=1`: Line buffering (only meaningful in text mode)
    - `buffering > 1`: Block buffering with a specified buffer size (in bytes)
    - If buffering isn't specified, the default is usually block buffering

??? Warning "Buffering Gotchas"
    - Excessive flushing (e.g., after every write) negates the benefits of buffering and can severely impact performance
    - If reading/writing in binary, there is no need to turn off buffering with `buffering=0`. This is very slow
    - Don't rely on implicit flushing. Always use `with` or explicitly call `f.close()` to guarantee data is written

### Seek and Tell

The `seek()` and `tell()` methods give you fine-grained control over the file pointer's position, allowing you to read or write at specific locations within a file.

```python
import os

with open("data.txt", "r+") as f:
    print(f"Initial position: {f.tell()}")  # Output: 0

    data = f.read(5)  # Read the first 5 bytes
    print(f"Read: {data}")
    print(f"Position after read: {f.tell()}")  # Output: 5

    f.seek(10)  # Move to byte 10
    print(f"Position after seek: {f.tell()}")  # Output: 10

    data = f.read(5)
    print(f"Read: {data}")
    print(f"Position after read: {f.tell()}")

    f.seek(0)  # Go back to the beginning
    print(f"Position after seek to beginning: {f.tell()}")

    f.seek(-5, os.SEEK_END)  # Go 5 bytes before the end
    print(f"Position after seek from end: {f.tell()}")
    data = f.read()
    print(f"Read to end: {data}")
```

```python
# Example in binary mode
import os

with open('my_binary_file.bin', 'wb+') as file:
    file.write(b'0123456789abcdef')  # Write some initial data
    file.seek(5)                     # Move to the 6th byte (index 5)
    print(file.read(1))              # Read one byte: b'5'
    file.seek(-3, os.SEEK_CUR)       # Move 3 bytes back from current position
    print(file.read(1))              # Read one byte: b'3'
    file.seek(-2, os.SEEK_END)       # Move to 2 bytes before the end
    print(file.read())               # Read the last two bytes: b'ef'
```

!!! Note "Seek Reference Points"
    The `seek(offset, whence)` method takes two parameters:
    - `offset`: The number of bytes to move
    - `whence`: The reference point for the offset:
        - `0` (or `os.SEEK_SET`): Beginning of the file (default)
        - `1` (or `os.SEEK_CUR`): Current position
        - `2` (or `os.SEEK_END`): End of the file

!!! Warning "Text Mode Limitations"
    - In text mode, `seek()` with a `whence` value of `1` (relative) or `2` (from the end) is generally not allowed, unless the offset is `0` or a value returned by a previous `tell()` call
    - This limitation exists because character encodings can have variable byte lengths
    - In binary mode, `seek()` is much more flexible, allowing arbitrary movement within the file

??? Tip "Seek and Tell Tricks"
    - Remember that `f.tell()` returns a byte offset, not a character offset, even in text mode
    - If you `seek()` past the end of a file and write, you fill the file with null bytes
    - In text mode, `seek()` is primarily used to return to a previous position obtained from `tell()`

## Directory Operations

Python provides functions for creating, listing, and removing directories through both the `os` module and `pathlib`.

```python
import os
import shutil
from pathlib import Path

# Creating directories with os
os.mkdir("new_directory")  # Creates a single directory
os.makedirs("parent_dir/child_dir")  # Creates parent directories if needed

# Creating directories with pathlib
new_dir = Path("another_directory")
new_dir.mkdir()

nested_dir = Path("parent/child")
nested_dir.mkdir(parents=True, exist_ok=True)  # Creates parent dirs, no error if exists

# Listing directory contents with os
contents = os.listdir(".")  # Lists contents of current directory
print(contents)

# Listing directory contents with pathlib
for item in Path(".").iterdir():
    print(item)

# Removing directories
os.rmdir("new_directory")  # Only works if directory is empty
Path("another_directory").rmdir()  # Same as above, only for empty directories

# Removing directories with contents
shutil.rmtree("parent_dir")  # Removes directory and all contents
```

!!! Danger "Using `shutil.rmtree()`"
    The `shutil.rmtree()` function removes a directory and all its contents permanently. Use it with extreme caution, as it can lead to data loss.

## File System Navigation

Python allows you to get and change the current working directory using the `os` module. By "change directory" in the context of the running program, we mean change the directory that the program is "looking at".  It always starts with the directory that the program was run from.

```python
import os

# Get the current working directory
current_dir = os.getcwd()
print(f"Current directory: {current_dir}")

# Change the current working directory
os.chdir("/path/to/another/directory")  # Replace with a valid path

# Get the new current working directory
new_current_dir = os.getcwd()
print(f"New current directory: {new_current_dir}")
```

!!! Warning "No pathlib Equivalent for `chdir()`"
    The `pathlib` module doesn't have a direct equivalent to `os.chdir()`. If you need to change the current directory, you must use `os.chdir()`.

### Recursively Walking Directories

The `os.walk()` function allows you to traverse a directory tree, accessing all files and subdirectories.

```python
import os

directory = "my_folder"
for root, dirs, files in os.walk(directory):
    print(f"Current directory: {root}")
    print(f"Subdirectories: {dirs}")
    print(f"Files: {files}")
    print("---")

    # Process specific files
    for file in files:
        if file.endswith(".txt"):
            full_path = os.path.join(root, file)
            print(f"Text file found: {full_path}")
```

## Error Handling with File Operations

File operations can encounter various errors. It's important to handle these gracefully.

```python
try:
    with open("nonexistent_file.txt", "r") as f:
        contents = f.read()
        print(contents)
except FileNotFoundError:
    print("Error: The file was not found.")
except PermissionError:
    print("Error: You do not have permission to access the file.")
except Exception as e:  # Catch any other exception
    print(f"An unexpected error occurred: {e}")
else:
    print("File read successfully.")
```

!!! Tip "Common File-Related Exceptions"
    - `FileNotFoundError`: The specified file doesn't exist
    - `PermissionError`: You don't have permissions to access the file
    - `IsADirectoryError`: You tried to open a directory as a file
    - `FileExistsError`: You tried to create a file that already exists with mode 'x'
    - `IOError`: General input/output error