# The Python Logging API

## Objectives

- Understand the Python's logging module
- Configure basic logging
- Learn logging levels and their meanings
- Format log output
- Understand the logging class hierarchy
- Use log handlers for different output destinations
- Configure logging with files or dictionaries

## Introduction to Logging

Logging is the activity of recording the inputs, outputs, and behavior of a program as it executes. The developer adds logging calls to indicate that certain events have occurred. Logs help with troubleshooting, debugging, and monitoring applications.

!!! Note
    Logging is not the same as debugging. Debugging is a real-time activity done by a programmer to find bugs, while logging creates a persistent record that can be examined after execution. BUT -- logging can be a great addition to your debugging toolkit.

The Python logging module provides several key components:

- **Loggers**: The interface that application code directly uses
- **Handlers**: Send log records to appropriate destinations
- **Formatters**: Specify the layout of log records in the final output
- **Filters**: Provide finer control over which log records to output

## Basic Configuration

Setting up basic logging in Python is straightforward using the `logging` module:

```python
import logging

# Configure the logger with a file and minimum logging level
logging.basicConfig(
    filename='mylogfile.log',
    level=logging.WARNING,
)

# Log messages at different levels
logging.warning('This message will be logged')
logging.debug('This message will NOT be logged')
logging.error('This message will be logged')
```

The resulting log file would contain:

```
WARNING:root:This message will be logged
ERROR:root:This message will be logged
```

!!! Note
    If you don't specify a filename, log messages will be printed to the console (stdout).


??? Note "Rich Equivalent"
    ```python
    import logging
    from rich.logging import RichHandler
    from rich.console import Console

    # Configure rich console
    console = Console()

    # Set up logging with RichHandler
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(message)s",
        handlers=[RichHandler(rich_tracebacks=True)]
    )

    logger = logging.getLogger("rich_example")

    # Log messages at different levels
    logger.debug("Debug message with [bold blue]rich[/bold blue] formatting")
    logger.info("Info message with [green]colors[/green]")
    logger.warning("Warning message with [yellow]highlights[/yellow]")
    logger.error("Error message with [bold red]emphasis[/bold red]")
    ```


## Logging Levels

Logging levels categorize entries by urgency. The lower the log level, the more output is generated. When you set a logging level, messages at that level or higher will be logged.

| Level          | Value | Meaning                                                                           |
|----------------|-------|-----------------------------------------------------------------------------------|
| CRITICAL/FATAL | 50    | A serious error indicating the program may be unable to continue running          |
| ERROR          | 40    | Software is not able to perform some function due to a problem                    |
| WARNING        | 30    | Something unexpected happened; software is still working                          |
| INFO           | 20    | Confirmation that software is working as expected                                 |
| DEBUG          | 10    | Detailed information for diagnosing problems                                      |
| UNSET          | 0     | Logs everything                                                                   |

### Getting and Configuring the Logger Object

Many logging operations require using the logger object:

```python
import logging

# Get the root logger
logger = logging.getLogger()

# Get a named logger
module_logger = logging.getLogger(__name__)

# Change logging level at runtime
logger.setLevel(logging.DEBUG)
```

### Capturing Stack Traces in Logs

When handling exceptions, you can include the full stack trace in your logs:

```python
import logging

logger = logging.getLogger(__name__)

try:
    open('/path/to/does/not/exist', 'rb')
except (SystemExit, KeyboardInterrupt):
    raise
except Exception:
    logger.error('Failed to open file', exc_info=True)
```

This would log something like:

```
ERROR:__main__:Failed to open file
Traceback (most recent call last):
  File "example.py", line 6, in <module>
    open('/path/to/does/not/exist', 'rb')
FileNotFoundError: [Errno 2] No such file or directory: '/path/to/does/not/exist'
```

## Formatting Log Output

You can customize the format of log messages using the `format` parameter:

```python
import logging

logging.basicConfig(
    format='%(asctime)s - %(levelname)s - %(module)s - %(message)s',
    level=logging.INFO
)

logging.info('Application started')
```

This produces output like:
```
2025-02-27 12:34:56,789 - INFO - main - Application started
```

### Common Format Directives

| Directive | Meaning |
|-----------|---------|
| `%(name)s` | Name of the logger |
| `%(levelno)s` | Numeric logging level |
| `%(levelname)s` | Text logging level ('DEBUG', 'INFO', etc.) |
| `%(pathname)s` | Full pathname of the source file |
| `%(filename)s` | Filename portion of the pathname |
| `%(module)s` | Module name |
| `%(lineno)d` | Source line number |
| `%(funcName)s` | Function name |
| `%(asctime)s` | Human-readable time |
| `%(message)s` | The logged message |
| `%(thread)d` | Thread ID |
| `%(process)d` | Process ID |

!!! Tip
    Format specifiers ending with `s` output a string, `d` is an integer, and `f` is a float.

## Logging Class Hierarchy

The logging module supports a hierarchy of loggers with different names. This helps identify where log messages come from:

```python
import logging

logging.basicConfig(level=logging.WARNING)

# Create named loggers
logger1 = logging.getLogger('package1.module1')
logger2 = logging.getLogger('package2.module2')

logger1.warning('This message comes from one module')
logger2.warning('And this message comes from another module')
```

Output:
```
WARNING:package1.module1:This message comes from one module
WARNING:package2.module2:And this message comes from another module
```

Best practice is to use the current module's name as the logger name:

```python
import logging

# Create a logger with the current module name
logger = logging.getLogger(__name__)

def process_data():
    logger.debug("Processing data...")
```

## Using Log Handlers

Handlers route log messages to different output destinations. Python provides several built-in handlers:

- **StreamHandler**: Sends logs to streams (stdout, stderr, files)
- **FileHandler**: Sends logs to a file
- **RotatingFileHandler**: Sends logs to files with rotation (size limits)
- **TimedRotatingFileHandler**: Rotates logs based on time
- **SMTPHandler**: Sends logs via email
- **SysLogHandler**: Sends logs to the system logger (Unix/Linux)
- **NTEventLogHandler**: Sends logs to the Windows event log
- **SocketHandler**: Sends logs over a network socket

### Configuring a SysLogHandler (Linux/Unix)

```python
import logging
import logging.handlers

# Create a logger
logger = logging.getLogger('MyLogger')
logger.setLevel(logging.DEBUG)

# Create a handler that writes to the system log
handler = logging.handlers.SysLogHandler(address='/dev/log')
logger.addHandler(handler)

# Log messages
logger.debug('This is a debug message')
logger.critical('This is a critical message')
# Both of these will be sent to the system log
```

### Configuring a RotatingFileHandler

This handler creates multiple log files when they reach a certain size:

```python
import logging
from logging.handlers import RotatingFileHandler

# Create a logger
logger = logging.getLogger('my_logger')
logger.setLevel(logging.DEBUG)

# Create a handler that rotates files
handler = RotatingFileHandler(
    'app.log',
    maxBytes=2000,  # Maximum file size in bytes
    backupCount=5   # Keep up to 5 backup files
)
logger.addHandler(handler)

# Add a formatter
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

# Log messages
for _ in range(1000):
    logger.debug('This is log message number %d', _)
```

This will create files like `app.log`, `app.log.1`, `app.log.2`, etc.

## Configuring Logging with Files or Dictionaries

Instead of configuring logging with code, you can use external configuration.

### Using a Configuration File

```ini
# logging_config.ini
[loggers]
keys=root

[handlers]
keys=consoleHandler,fileHandler

[formatters]
keys=simpleFormatter

[logger_root]
level=DEBUG
handlers=consoleHandler,fileHandler

[handler_consoleHandler]
class=StreamHandler
level=INFO
formatter=simpleFormatter
args=(sys.stdout,)

[handler_fileHandler]
class=handlers.RotatingFileHandler
level=DEBUG
formatter=simpleFormatter
args=('app.log', 'a', 10000, 3)

[formatter_simpleFormatter]
format=%(asctime)s - %(name)s - %(levelname)s - %(message)s
datefmt=%Y-%m-%d %H:%M:%S
```

Then in your Python code:

```python
import logging
import logging.config

# Configure from file
logging.config.fileConfig('logging_config.ini')

# Get the root logger
logger = logging.getLogger()

# Log messages
logger.debug('Debug message')
logger.info('Info message')
logger.warning('Warning message')
```

!!! Tip "Best Practices for Logging"
    - Use named loggers with module names (`__name__`)
    - Configure appropriate log levels based on the environment
    - Include context in log messages (user IDs, request IDs, etc.)
    - Apply different handlers for different severity levels

??? "Rich Logging Example 1"

    ```python
    ## Example: Custom Rich Handler with Log Levels
    import logging
    from rich.console import Console
    from rich.theme import Theme
    from rich.logging import RichHandler

    # Create a custom theme for different log levels
    custom_theme = Theme({
        "debug": "dim cyan",
        "info": "green",
        "warning": "yellow",
        "error": "bold red",
        "critical": "bold white on red"
    })

    console = Console(theme=custom_theme)

    # Create custom rich handler
    handler = RichHandler(
        console=console,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        show_time=True
    )

    # Configure logger
    logger = logging.getLogger("rich_levels")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)

    # Log with rich formatting at different levels
    logger.debug("System starting up...")
    logger.info("User [bold]admin[/bold] logged in")
    logger.warning("Database connection pool at 80% capacity")
    logger.error("Failed to process transaction #12345")
    logger.critical("System shutting down due to fatal error")
    ```

??? "Rich Logging Example 2"

    ```python
    ## Example: Rich Progress Bars in Logs
    import time
    import logging
    from rich.console import Console
    from rich.logging import RichHandler
    from rich.progress import Progress, TextColumn, BarColumn, TaskProgressColumn

    # Set up basic logging with rich
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        handlers=[RichHandler()]
    )

    logger = logging.getLogger("rich_progress")
    console = Console()

    # Log with a progress bar
    logger.info("Starting long-running operation")

    with Progress(
        TextColumn("[bold blue]{task.description}"),
        BarColumn(complete_style="green", finished_style="green"),
        TaskProgressColumn(),
        console=console
    ) as progress:
        task = progress.add_task("[cyan]Processing...", total=100)

        for i in range(100):
            # Do some work
            time.sleep(0.05)
            # Update progress
            progress.update(task, advance=1)
            # Occasionally log important info
            if i % 25 == 0:
                logger.info(f"Milestone reached: {i}% complete")

    logger.info("[bold green]Operation completed successfully![/bold green]")
    ```

??? "Rich Logging Example 3"

    ```python
    ## Example: Rich Tables in Logs
    import logging
    from rich.console import Console
    from rich.logging import RichHandler
    from rich.table import Table

    # Configure logging with rich
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        handlers=[RichHandler()]
    )

    logger = logging.getLogger("rich_tables")
    console = Console()

    # Create a rich table
    def log_user_statistics():
        table = Table(title="User Statistics")

        table.add_column("Username", style="cyan")
        table.add_column("Login Count", justify="right", style="green")
        table.add_column("Status", style="magenta")

        table.add_row("alice", "42", "Active")
        table.add_row("bob", "13", "Inactive")
        table.add_row("charlie", "7", "Locked")
        table.add_row("david", "89", "Active")

        # Log the table by rendering it to string and logging
        logger.info("User statistics report:")
        console.print(table)

    log_user_statistics()
    ```

??? "Rich Logging Example 4"

    ```python
    ## Example: Rich File and Console Logging
    import logging
    import sys
    from rich.console import Console
    from rich.logging import RichHandler
    from rich.theme import Theme

    # Set up custom theme
    custom_theme = Theme({
        "info": "green",
        "warning": "yellow",
        "error": "red",
        "critical": "bold red reverse"
    })

    # Create console with theme
    console = Console(theme=custom_theme)

    # Create handlers
    rich_handler = RichHandler(
        console=console,
        rich_tracebacks=True,
        show_time=True,
        show_path=False
    )

    # Also send logs to a file
    file_handler = logging.FileHandler("application.log")
    file_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setFormatter(file_formatter)

    # Configure logger
    logger = logging.getLogger("dual_output")
    logger.setLevel(logging.DEBUG)
    logger.addHandler(rich_handler)
    logger.addHandler(file_handler)

    # Log messages
    logger.debug("This message goes to both console (with rich formatting) and file")
    logger.info("System status: [bold green]OK[/bold green]")
    logger.warning("Warning: [yellow]High CPU usage detected[/yellow]")
    logger.error("Error in module X: [bold red]Connection refused[/bold red]")
    ```

??? "Rich Logging Example 5"

    ```python
    ## Example: Rich Tracebacks and Exception Handling
    import logging
    from rich.console import Console
    from rich.logging import RichHandler
    from rich.traceback import install

    # Install rich tracebacks
    install(show_locals=True)

    # Configure rich logging
    console = Console()
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(message)s",
        handlers=[RichHandler(console=console, rich_tracebacks=True)]
    )

    logger = logging.getLogger("exception_handler")

    # Example function that will raise an exception
    def process_data(data):
        try:
            # Simulate an error
            result = data['key'] / 0
            return result
        except Exception as e:
            logger.exception(f"Failed to process data: {e}")
            raise

    # Try to call the function
    try:
        sample_data = {"key": 42}
        logger.info(f"Processing data: {sample_data}")
        process_data(sample_data)
    except Exception:
        logger.critical("Processing failed, please check logs for details")
    ```

??? "Rich Logging Example 6"

    ```python
    ## Example: Rich Context Variables and Structured Logging
    import logging
    import time
    from rich.console import Console
    from rich.logging import RichHandler
    from rich.panel import Panel
    from rich import box
    import uuid

    # Create console
    console = Console()

    # Configure logging
    logging.basicConfig(
        level=logging.INFO,
        format="%(message)s",
        handlers=[RichHandler(console=console, rich_tracebacks=True, markup=True)]
    )

    logger = logging.getLogger("structured_logging")

    # Create a context manager for logging with request context
    class RequestContext:
        def __init__(self, user_id=None, request_path=None):
            self.request_id = str(uuid.uuid4())[:8]
            self.user_id = user_id
            self.request_path = request_path
            self.start_time = time.time()

        def __enter__(self):
            console.print(Panel(
                f"[bold blue]Request {self.request_id} started[/bold blue]\n"
                f"User: {self.user_id}\n"
                f"Path: {self.request_path}",
                box=box.ROUNDED
            ))
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            duration = time.time() - self.start_time
            status = "[red]Error[/red]" if exc_type else "[green]Success[/green]"
            console.print(Panel(
                f"[bold blue]Request {self.request_id} completed[/bold blue]\n"
                f"Status: {status}\n"
                f"Duration: {duration:.2f}s",
                box=box.ROUNDED
            ))
            return False

    # Simulate a request
    def process_request(user_id, path):
        with RequestContext(user_id=user_id, request_path=path) as ctx:
            logger.info(f"Processing request [dim]({ctx.request_id})[/dim]")

            # Simulate some work
            logger.info("Authenticating user")
            time.sleep(0.5)

            logger.info("Retrieving data")
            time.sleep(0.7)

            if path == "/admin" and user_id != "admin":
                logger.error("[bold red]Permission denied![/bold red]")
                raise PermissionError("Unauthorized access to admin area")

            logger.info("[green]Request processed successfully[/green]")
            return {"status": "success", "data": {"value": 42}}

    # Example usage
    try:
        process_request("alice", "/dashboard")
        process_request("bob", "/admin")  # This will fail
    except Exception as e:
        pass  # Context manager will log the error
    ```


??? "Rich Logging Example 7"

    ```python
    ## Example: Rich Logging Configuration with dictConfig
    import logging
    import logging.config
    from rich.console import Console
    from rich.logging import RichHandler

    # Create console for rich output
    console = Console()

    # Custom formatter that handles rich markup
    class RichStyleFormatter(logging.Formatter):
        def format(self, record):
            # Convert log level to rich style
            level_styles = {
                "DEBUG": "[cyan]DEBUG[/cyan]",
                "INFO": "[green]INFO[/green]",
                "WARNING": "[yellow]WARNING[/yellow]",
                "ERROR": "[red]ERROR[/red]",
                "CRITICAL": "[bold red on white]CRITICAL[/bold red on white]"
            }

            # Replace level name with styled version
            record.levelname = level_styles.get(record.levelname, record.levelname)

            # Use super class to format the message
            return super().format(record)

    # Define logging configuration dictionary
    LOGGING_CONFIG = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'rich': {
                '()': RichStyleFormatter,
                'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s',
                'datefmt': '[dim]%Y-%m-%d %H:%M:%S[/dim]',
            },
            'file': {
                'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
            },
        },
        'handlers': {
            'console': {
                '()': RichHandler,
                'level': 'DEBUG',
                'rich_tracebacks': True,
                'show_time': False,
                'console': console,
                'show_path': False,
            },
            'file': {
                'class': 'logging.handlers.RotatingFileHandler',
                'level': 'INFO',
                'formatter': 'file',
                'filename': 'app.log',
                'maxBytes': 10485760,  # 10MB
                'backupCount': 5,
            },
        },
        'loggers': {
            '': {  # Root logger
                'handlers': ['console', 'file'],
                'level': 'DEBUG',
            },
            'app': {
                'handlers': ['console', 'file'],
                'level': 'DEBUG',
                'propagate': False,
            },
        }
    }

    # Configure logging
    logging.config.dictConfig(LOGGING_CONFIG)

    # Get loggers
    root_logger = logging.getLogger()
    app_logger = logging.getLogger('app')

    # Log some messages
    root_logger.debug("System initialization")
    app_logger.info("Application started")
    app_logger.warning("Database connection pool running low")
    app_logger.error("Failed to process payment for order #12345")
    ```