# Modules and Packages

## Day 3: Modules and Packages - Organizing Your Code

### Module Basics

**Topic Introduction:** Modules are the foundation of code organization in Python.  They allow you to group related code (functions, classes, variables) into separate files, making your projects more structured and manageable.  Modules promote reusability and prevent naming conflicts.  This is essential for any project larger than a single script.

**Core Concepts and Syntax:**

*   A module is simply a Python file (`.py` extension) containing Python code.
*   `import module_name`:  Imports the entire module.
*   `from module_name import something`:  Imports specific attributes (functions, classes, variables) from the module.
*   `import module_name as alias`: Imports the module with a different name (alias).

**Detailed Code Examples:**

Create a file named `my_module.py`:

```python
# my_module.py
def greet(name):
    """Greets a person."""
    print(f"Hello, {name}!")

def add(x, y):
    """Adds two numbers."""
    return x + y

PI = 3.14159
```

Now, create another file (in the same directory) to use the module:

```python
# main.py
import my_module

my_module.greet("Alice")  # Output: Hello, Alice!
result = my_module.add(5, 3)
print(result)  # Output: 8
print(my_module.PI) # Output: 3.14159

from my_module import greet, add

greet("Bob")  # Output: Hello, Bob!
print(add(10, 2)) # Output: 12

from my_module import PI as my_pi
print(my_pi) # Output: 3.14159

import my_module as mm # Alias import
mm.greet("Carol") # Output: Hello, Carol!
```

**Tips, Tricks, and Advanced Techniques:**

*   `if __name__ == "__main__":` block:  This code runs only when the module is executed directly (not when imported).  Useful for testing.

```python
# my_module.py (modified)
def greet(name):
    """Greets a person."""
    print(f"Hello, {name}!")

if __name__ == "__main__":
    greet("Test User")  # This will only run when my_module.py is executed directly
```

**Best Practices and Industry Standards:**

*   Use descriptive module names (lowercase with words separated by underscores).
*   Organize related code into modules.

**Warnings and Potential Issues:**

*   Circular imports (where module A imports module B, and module B imports module A) can cause problems.

---

### Package Structure

**Topic Introduction:** Packages take modularity a step further by allowing you to organize modules into directories.  This creates a hierarchical structure for your code, making it even more organized and manageable, especially for large projects.  Packages help avoid naming conflicts between modules in different parts of your project.

**Core Concepts and Syntax:**

*   A package is a directory containing Python modules and a special file named `__init__.py` (which can be empty).
*   The `__init__.py` file tells Python that the directory should be treated as a package.

**Detailed Code Examples:**

Create the following directory structure:

```
my_package/
    __init__.py
    module1.py
    module2.py
```

`__init__.py`: (Can be empty)

```python
# __init__.py (optional content)
# This file can contain initialization code for the package
# It can also define __all__ to control what gets imported with 'from my_package import *'
# __all__ = ["module1", "module2"]
```

`module1.py`:

```python
# module1.py
def function1():
    print("Function 1 from module 1")
```

`module2.py`:

```python
# module2.py
def function2():
    print("Function 2 from module 2")
```

Now, create a file outside the `my_package` directory:

```python
# main.py
import my_package.module1
import my_package.module2

my_package.module1.function1()  # Output: Function 1 from module 1
my_package.module2.function2()  # Output: Function 2 from module 2

from my_package import module1

module1.function1() # Output: Function 1 from module 1

from my_package.module2 import function2

function2() # Output: Function 2 from module 2
```

**Tips, Tricks, and Advanced Techniques:**

*   You can have nested packages (packages within packages).

**Best Practices and Industry Standards:**

*   Use packages to organize large projects into logical units.
*   Keep the `__init__.py` file simple (or empty) unless you have specific initialization needs.

**Warnings and Potential Issues:**

*   Make sure the `__init__.py` file exists in each package directory.

---

### Import System

**Topic Introduction:** Understanding Python's import system clarifies how modules and packages are located and loaded.  This knowledge is crucial for resolving import errors, managing dependencies, and understanding the scope of your code.

**Core Concepts and Syntax:**

*   **Absolute Imports:**  Specify the full path to the module or package, starting from the top-level package (if any).
*   **Relative Imports:** Specify the path to the module or package *relative* to the current module.  Use dots (`.`) to indicate the current directory or parent directories.

**Detailed Code Examples:**

Consider this slightly more complex package structure:

```
my_project/
    main.py
    package1/
        __init__.py
        moduleA.py
        moduleB.py
    package2/
        __init__.py
        moduleC.py
```

`package1/moduleA.py`:

```python
# package1/moduleA.py
def functionA():
    print("Function A from module A")
```

`package1/moduleB.py`:

```python
# package1/moduleB.py
from . import moduleA  # Relative import (same package)
# from my_project.package1 import moduleA # would also work (absolute import)

def functionB():
    moduleA.functionA()
    print("Function B from module B")
```

`package2/moduleC.py`:

```python
# package2/moduleC.py
from ..package1 import moduleB  # Relative import (parent package)

def functionC():
    moduleB.functionB()
    print("Function C from module C")
```

`main.py`:

```python
# main.py
from package1 import moduleB
from package2 import moduleC

moduleB.functionB()
moduleC.functionC()
```

**Tips, Tricks, and Advanced Techniques:**

*   Relative imports are useful within packages to avoid hardcoding the package name.
*   Absolute imports are generally preferred for clarity and to avoid ambiguity.

**Best Practices and Industry Standards:**

*   Use absolute imports for top-level imports in your main scripts.
*   Use relative imports within packages to refer to other modules in the same package.
*   PEP 328 describes the import system in detail.

**Warnings and Potential Issues:**

*   Incorrect relative imports can lead to `ImportError`.
*   Circular relative imports can cause problems.

---

### Creating Modules

(Covered in Module Basics - this section just reinforces it)

---

### Module Search Path

**Topic Introduction:**  Python uses a search path to find modules and packages when you import them.  Understanding this search path is essential for troubleshooting import errors and ensuring that Python can find your custom modules and packages.

**Core Concepts and Syntax:**

*   `sys.path`: A list of strings that specifies the directories Python searches for modules.
*   The search path includes:
    *   The directory containing the input script (or the current directory).
    *   The `PYTHONPATH` environment variable (if set).
    *   Installation-dependent default paths (where standard library modules are located).

**Detailed Code Examples:**

```python
import sys

print(sys.path)  # Prints the module search path

# Modifying the search path (temporarily)
sys.path.append("/path/to/my/modules")  # Adds a directory to the search path

# Modifying the search path (permanently - usually not recommended)
# Set the PYTHONPATH environment variable in your shell configuration file.
```

**Tips, Tricks, and Advanced Techniques:**

*   You can use `sys.path.insert(0, "/path/to/my/modules")` to add a directory to the *beginning* of the search path, giving it higher priority.

**Best Practices and Industry Standards:**

*   Avoid modifying `sys.path` directly unless necessary.
*   Use packages and proper project structure to organize your code.

**Warnings and Potential Issues:**

*   Modifying `sys.path` incorrectly can lead to unexpected behavior.
*   Having multiple modules with the same name in different directories on the search path can lead to conflicts.

---

### Package Distribution (Brief Overview)

**Topic Introduction:**  Packaging and distributing your Python code allows others to easily install and use your modules and packages.  This is essential for sharing your work with the community or deploying your applications.

**Core Concepts and Syntax:** (This is a high-level overview; detailed steps are beyond the scope of this lesson)

*   `setuptools` and `poetry` are popular tools for packaging Python projects.
*   A `setup.py` file (or `pyproject.toml` for `poetry`) is used to describe your package and its dependencies.
*   You can upload your package to the Python Package Index (PyPI) to make it publicly available.

**Detailed Code Examples:** (Conceptual)

```
# Example setup.py (simplified)
from setuptools import setup, find_packages

setup(
    name="my_package",
    version="0.1.0",
    packages=find_packages(),
    install_requires=[
        "requests",
        "numpy",
    ],
)
```

**Best Practices and Industry Standards:**

*   Follow the official Python packaging guide for detailed instructions.

**Warnings and Potential Issues:**

*   Packaging can be complex, especially for projects with many dependencies.

---

### Common Standard Library Modules

**Topic Introduction:**  Python's standard library is a vast collection of modules that provide a wide range of functionality, from file I/O to networking to data processing.  Familiarizing yourself with these modules will significantly enhance your productivity.

**Core Concepts and Syntax:** (Brief descriptions and key use cases)

*   **`os`:**  Operating system interfaces (e.g., interacting with the file system, environment variables).  *Key Use Cases:* File/directory manipulation, process management.
*   **`sys`:** System-specific parameters and functions (e.g., accessing command-line arguments, modifying the module search path). *Key Use Cases:* Script configuration, system interaction.
*   **`shutil`:** High-level file operations (e.g., copying, moving, deleting files and directories). *Key Use Cases:* File/directory management.
*   **`datetime`:**  Classes for manipulating dates and times. *Key Use Cases:* Working with dates and times, scheduling tasks.
*   **`math`:** Mathematical functions (e.g., trigonometric functions, logarithms). *Key Use Cases:* Numerical computations.
*   **`random`:**  Generate pseudo-random numbers. *Key Use Cases:* Simulations, games, cryptography (for non-security-critical purposes).
*   **`json`:**  Encode and decode JSON data.  *Key Use Cases:* Working with APIs, data serialization.
*   **`re`:** Regular expressions (pattern matching in strings). *Key Use Cases:* Text processing, data validation.
*   **`collections`:**  Specialized container datatypes (e.g., `namedtuple`, `deque`, `Counter`). *Key Use Cases:* Data structures beyond the built-in types.
*   **`itertools`:** Functions creating iterators for efficient looping. *Key Use Cases:* Working with iterators, combinatorial tasks.
