# Unit Testing Python Code

## Objectives

- Learn unit testing terminology
- Learn the properties of good unit tests
- Learn how to get started with unit testing with the unittest module
- Understand unittest mock objects
- Learn how to use pytest as an alternative testing framework
- Compare unittest and pytest approaches

## Introduction

Testing is a vital stage in the software development process. No real software is developed without extensive testing. Software developers need to be aware of testing types, team responsibilities, and the toolsets available for testing.

While manual testing (running code, supplying inputs, and analyzing outputs) is important, this lesson focuses on **automated unit testing** using Python libraries.

Among the many test types (unit, stress, integration, load, security, etc.), **unit testing** is universally recognized as the programmer's responsibility. No programmer submits code for review until it has passed all unit tests.

This lesson covers:
- The `unittest` framework for writing and running tests
- Creating mock objects to isolate code dependencies
- The `pytest` framework as a modern alternative
- Choosing the right testing approach for your project

## Unit Testing Defined

Unit testing is a software testing method for validating individual units of source code in isolation.

```python
# A unit is typically a function, method, or class that can be tested independently
def add(a, b):
    return a + b

# A unit test for the add function
def test_add():
    assert add(2, 3) == 5
```

In procedural programming, a unit is often a function or procedure. In object-oriented programming, a unit is usually a class or method.

By writing tests for the smallest testable units, then testing interactions between those units, developers can build comprehensive tests for complex applications.

!!! Note "Key Concept"
    Unit testing is the testing of a single code entity when isolated completely from its dependencies.

## Unit Testing Terminology

- **Unit**: Any discrete module of code that can be tested in isolation (function, method, or class)
- **Test Case**: A single function validating a specific aspect of the software
- **Setup Method**: Establishes needed data/objects for the unit test
- **Teardown Method**: Cleans up data/objects created during test execution
- **Test Fixture**: Preparation needed to perform tests and associated cleanup
- **Test Suite**: A series of test cases that test multiple steps in a scenario
- **Test Runner**: A library or tool that locates test suites and executes them
- **Class Under Test (CUT)**: The class being tested
- **Fake Class**: Any class that simulates dependencies needed by a CUT

There are two kinds of fakes:
- **Stub**: A fake that exists purely to allow the test to run
- **Mock**: A fake that tracks the behavior of the CUT and determines test success or failure

## Properties of Good Unit Tests

### Focus on a Single Functionality

A test case should focus on one tiny bit of functionality and prove it correct. Test cases should be small functions, usually less than 10 lines of code.

```python
# Good: Single-focus test case
def test_user_login_succeeds_with_valid_credentials():
    user = User("testuser", "password123")
    result = user.login("testuser", "password123")
    assert result is True

# Good: Another single-focus test case
def test_user_login_fails_with_invalid_password():
    user = User("testuser", "password123")
    result = user.login("testuser", "wrong_password")
    assert result is False
```

!!! Tip "Test One Thing"
    A test case should ask a single question, so you can focus on each design decision and more easily debug when tests fail.

### Independence and Idempotence

Each test unit must be fully independent and able to run alone or within the test suite, regardless of execution order.

```python
# These methods must be part of a unittest.TestCase class
class TestUserFunctionality(unittest.TestCase):
    def setUp(self):
        self.temp_db = create_test_database()
        self.user = User("test_user", "password", self.temp_db)

    def tearDown(self):
        self.temp_db.close()
        delete_test_database(self.temp_db)

    def test_user_login(self):
        # Test code here
        result = self.user.login("test_user", "password")
        self.assertTrue(result)
```

!!! Warning "Test Independence"
    Tests that depend on other tests or on a specific execution order are brittle and difficult to maintain.

### Self-Explanatory Test Cases

Use long, descriptive names for testing functions to clearly communicate their purpose.

```python
# Good: Descriptive test name
def test_square_of_negative_number_returns_positive_result():
    assert square(-4) == 16

# Not as good: Vague test name
def test_square():
    assert square(-4) == 16
```

!!! Note "Readability"
    Test function names are displayed when a test fails and should be as descriptive as possible.

### Other Best Practices

- Write fast tests that run in milliseconds
- Learn how to run a single test or test case
- Run the full test suite before and after coding sessions
- Implement hooks to run tests before pushing code
- Write a failing test to mark where work was interrupted
- Write tests for bugs when they're discovered
- Consider test code as documentation for new developers

## Getting Started with Unit Testing with unittest

Test code usually contains three components:
- **Setup**: Arranges objects and test data
- **Action**: Invokes the method or function being tested
- **Assertion**: Verifies the action worked as expected

This is often called the **Arrange-Act-Assert** pattern.

### The `assert` Statement

The heart of unit testing is using assertions to verify code behavior.

```python
# Basic assert statement
assert sum([1, 2, 3]) == 6, "Result should be 6"

# If the assertion is False, an AssertionError is raised
assert sum([1, 2, 3, 4]) == 6, "Result should be 6"
# AssertionError: Result should be 6
```

### The `unittest` Module

The `unittest` module has been part of the Python standard library since version 2.1. It contains a testing framework, test runner, and mocking tools.

Requirements for writing and executing tests with unittest:

1. Test cases are methods within a class
2. The test class must inherit from `unittest.TestCase`
3. Test case methods begin with the prefix `test_`
4. Tests use special assertion methods from unittest.TestCase
5. The test class must call `unittest.main()`

```python
#!/usr/bin/env python3

import unittest

class TestSum(unittest.TestCase):
    def test_sum(self):
        self.assertEqual(sum([1, 2, 3]), 6, "Result should be 6")

    def test_sum_tuple(self):
        self.assertEqual(sum((1, 2, 2)), 5, "Result should be 5")

if __name__ == '__main__':
    unittest.main()
```

### Assertion Methods in unittest

Below are the most commonly used assertion methods in unittest:

| Method | Equivalent Python expression |
|--------|------------------------------|
| `assertEqual(a, b)` | a == b |
| `assertTrue(x)` | x is True (bool(x) == True) |
| `assertFalse(x)` | x is False (bool(x) == False) |
| `assertIs(a, b)` | a is b |
| `assertIsNone(x)` | x is None |
| `assertIn(a, b)` | a in b |
| `assertIsInstance(a, b)` | isinstance(a, b) |

!!! Note "Type-Specific Methods"
    Python calls type-specific methods when executing `assertEqual()`. For example, `assertListEqual(a, b)` is used for comparing lists.

Additional assertion methods check for exceptions, warnings, or log messages:

| Method | Checks that: |
|--------|-------------|
| `assertRaises(exc, func, *args, **kwds)` | func(*args, **kwds) raises exc |
| `assertRaisesRegex(exc, r, func, *args, **kwds)` | func(*args, **kwds) raises exc and message matches regex r |
| `assertWarns(warn, func, *args, **kwds)` | func(*args, **kwds) raises warn |
| `assertWarnsRegex(warn, r, func, *args, **kwds)` | func(*args, **kwds) raises warn and message matches regex r |
| `assertLogs(logger, level)` | The with block logs on logger with minimum level |

Methods for specific checks:

| Method | Equivalent Python expression |
|--------|------------------------------|
| `assertAlmostEqual(a, b)` | round(a-b, 7) == 0 |
| `assertGreater(a, b)` | a > b |
| `assertLess(a, b)` | a < b |
| `assertRegex(s, r)` | r.search(s) is not None |
| `assertCountEqual(a, b)` | a and b have the same elements in the same number |

### The `setUp()` and `tearDown()` Methods

Use `setUp()` and `tearDown()` methods to establish a test environment and clean up afterward.

```python
import unittest

class TestDatabase(unittest.TestCase):
    def setUp(self):
        # Create a fresh test database
        self.connection = create_test_database()

    def test_record_insertion(self):
        # Test inserting a record
        result = self.connection.insert("test", {"name": "John"})
        self.assertTrue(result.success)

    def tearDown(self):
        # Clean up by removing the test database
        self.connection.close()
        delete_test_database()
```

!!! Note "Method Execution Order"
    `setUp()` runs before each test method.
    `tearDown()` runs after each test method.

There are also class-level setup and teardown methods:
- `setUpClass()` runs once before all test methods
- `tearDownClass()` runs once after all test methods

### Test Discovery

Test discovery helps find tests in the codebase automatically.

```bash
python -m unittest discover
```

This command searches the current directory for files named `test*.py` and runs all tests found.

## unittest Mock Objects

Mock objects simulate the behavior of complex, real objects when it's impractical to use the real object in a test.

### When to Use Mocks

Mocks are useful when the real object:

1. Supplies non-deterministic results (e.g., current time)
2. Has states difficult to create or reproduce (e.g., network error)
3. Is slow (e.g., a database)
4. Does not yet exist or may change behavior
5. Would need code exclusively for testing purposes

### Using Mock Objects with Python

Python provides mocking tools through the `unittest.mock` package, which includes:
- The `Mock` class for creating stub objects
- The `patch()` decorator for substituting mock objects temporarily

```python
from unittest.mock import Mock, patch

# Creating a mock directly
database = Mock()
database.connect.return_value = True
database.get_record.return_value = {"id": 1, "name": "John"}

# Using patch to substitute a mock
@patch('mymodule.Database')
def test_user_retrieval(mock_database):
    mock_database.return_value.get_user.return_value = {"name": "John"}
    result = get_username(1)  # Function that uses Database
    assert result == "John"
```

### Using the `@patch` Decorator

The `@patch` decorator allows temporary substitution of classes or functions with mock objects.

```python
import unittest
from unittest.mock import patch

class TestLongRunningOperation(unittest.TestCase):
    @patch('mymodule.LongOperation.process')
    def test_processing(self, mock_process):
        # Configure the mock to return a specific value
        mock_process.return_value = "processed"

        # Call the function that uses the patched method
        result = call_processing_function()

        # Verify the result
        self.assertEqual(result, "processed")

        # Verify the mock was called with expected arguments
        mock_process.assert_called_with(expected_args)
```

## Introduction to pytest

Pytest is a popular alternative to unittest, offering a simpler syntax and powerful features. It's not part of the standard library but can be installed via pip.

```bash
pip install pytest
```

### Basic pytest Usage

Unlike unittest, pytest allows you to write tests as simple functions rather than class methods:

```python
# test_sample.py
def test_addition():
    assert 1 + 1 == 2

def test_string_concatenation():
    assert "Hello" + " " + "World" == "Hello World"
```

To run tests:

```bash
pytest test_sample.py
```

### Advantages of pytest

1. **Simpler syntax**: Uses Python's built-in `assert` statement instead of special methods
2. **Comprehensive output**: Provides detailed information when tests fail
3. **Powerful fixtures**: More flexible than unittest's setup/teardown methods
4. **Auto-discovery**: Automatically finds test files and functions
5. **Parametrization**: Run a test with multiple inputs
6. **Rich plugin ecosystem**: Extends functionality with hundreds of plugins

### pytest Fixtures

Fixtures are a powerful feature of pytest, allowing you to define reusable test data or setup operations.

```python
import pytest

@pytest.fixture
def database_connection():
    # Setup
    connection = create_test_database()
    # Provide the fixture value
    yield connection
    # Teardown (runs after the test function completes)
    connection.close()
    delete_test_database()

def test_database_insertion(database_connection):
    # The fixture value is passed as an argument
    result = database_connection.insert("test", {"name": "John"})
    assert result.success
```

!!! Note "Fixture Scope"
    Fixtures can have different scopes: `function` (default), `class`, `module`, or `session`. This gives you fine-grained control over how often setup and teardown occurs.

### Advanced pytest Features

#### Parametrized Tests

Run the same test with multiple inputs:

```python
import pytest

@pytest.mark.parametrize("input,expected", [
    (1, 1),
    (2, 4),
    (3, 9),
    (4, 16)
])
def test_square(input, expected):
    assert input * input == expected
```

#### Test Marking

Mark tests for selective execution:

```python
@pytest.mark.slow
def test_slow_operation():
    # Time-consuming test
    ...

# Run only tests marked 'slow'
# pytest -m slow
```

#### Built-in Fixtures

pytest includes many built-in fixtures:
- `tmp_path`: Creates a temporary directory for tests
- `capsys`: Captures stdout/stderr
- `monkeypatch`: Temporarily modify objects or environment variables

```python
def test_print_output(capsys):
    print("Hello World")
    captured = capsys.readouterr()
    assert captured.out == "Hello World\n"
```

## unittest vs pytest: Which to Choose?

Both unittest and pytest have strengths and weaknesses. Here's a comparison to help choose between them:

### Philosophy and Approach

- **unittest**: Follows the xUnit testing pattern with a class-based approach
- **pytest**: Emphasizes simplicity with function-based tests and powerful fixtures

### Syntax Comparison

```python
# unittest approach
import unittest

class TestMathFunctions(unittest.TestCase):
    def test_addition(self):
        self.assertEqual(1 + 1, 2)

# pytest approach
def test_addition():
    assert 1 + 1 == 2
```

### Fixtures/Setup

```python
# unittest setup/teardown
def setUp(self):
    self.resource = setup_resource()

def tearDown(self):
    self.resource.cleanup()

# pytest fixture
@pytest.fixture
def resource():
    resource = setup_resource()
    yield resource
    resource.cleanup()
```

### When to Choose unittest

- Your project needs to be compatible with Python's standard library only
- You prefer a class-based, object-oriented approach to testing
- You're working with a legacy codebase already using unittest
- You need to maintain compatibility with Python 2

### When to Choose pytest

- You want simpler test syntax with less boilerplate
- You need flexible fixture management
- You want rich failure reports with detailed context
- You plan to use plugins to extend testing capabilities
- You value a modern testing approach with active community development

!!! Tip "Best of Both Worlds"
    pytest can run unittest tests, so you can gradually migrate from unittest to pytest or use both together.

## Check on Learning / Challenges

!!! Warning "Complete challenges on your VTA workstation and dev environment."

- Today's challenges can be found at `~/170D/cybbh/challenges/testing/{challenge_number}`
- All challenge prompts are contained in their respective `main.py` file as comments.
- The `solution.py` file exists primarily as a last resort after you've exhausted getting help from the instructor, a classmate, or the web.