# Introduction to Python Classes / Object-Oriented Programming (OOP)

## Topics

* Defining a Class
* Class Constructor
* Properties
* Methods
* Special Methods

## Introduction to Python Classes

* Python Classes represents real-world(mostly) entities as objects in code
* Objects primarily consist of:
    * **Attributes**: Data describing object characteristics/state
    * **Behaviors**: Methods defining object actions
* OOP benefits:
    * Intuitive modeling of complex systems
    * More organized, modular, and maintainable code

* In OOP, objects interact to accomplish tasks
* Python provides constructs to:
    * Create objects
    * Assign attributes
    * Code behaviors
    * Enable object interactions

* This chapter covers:
    * OOP tools, features, and keywords in Python
    * Encapsulation: Merging attributes and methods
* Data hiding: Concealing internal object structure
* **NOT** covered in this chapter:
    * **Inheritance**: Parent/child relationships between classes
    * **Polymorphism**: Context-specific behavior

## Simple Class Example

* In Python, classes and types are essentially the same thing
* Every class defines a new type
* The `type()` function returns the class of an object


```python
class BankAccount:
    def __init__(self, balance, acct_holder):
        self._balance = balance
        self._account_holder = acct_holder

account_1 = BankAccount(100, "John")
account_2 = BankAccount(200, "Jane")

print(f'BankAccount: {BankAccount}')
print(f'type(account_1): {type(account_1)}')
print()
print(f'account_1: {account_1}')
print(f'account_2: {account_2}')
print()
print(f'account_1.balance: {account_1._balance}')
print(f'account_2.acct_holder: {account_2._account_holder}')

# Output:
# BankAccount: <class '__main__.BankAccount'>
# type(account_1): <class '__main__.BankAccount'>

# account_1: <__main__.BankAccount object at 0x7feb867a9c60>
# account_2: <__main__.BankAccount object at 0x7feb867a9c30>

# account_1.balance: 100
# account_2.acct_holder: Jane
```

# Class Constructor

Special method `__init__()` initializes objects (dunder method)
First parameter is always `self`
Assign attributes in the constructor:

### NOTE:

* The `_` prefix is a convention to indicate that the attribute is private
* True private attributes are not enforced in Python and the user can still access the attribute

```python
    def __init__(self, balance, acct_holder):
        self._balance = balance
        self._account_holder = acct_holder
```

# Class Functions aka Methods

- Define object behaviors
- First parameter is always `self`
- Example:

```python
class BankAccount:
    def __init__(self, balance, acct_holder):
        self._balance = balance
        self._account_holder = acct_holder

    def close_account(self):
        if self._acct_status == "Open":
            self._acct_status = "Closed"
        else:
            print(f'Account {self._account_id} already closed!')
```

# Properties

- Use `@property` *decorator* for getter methods
- Use `@<property_name>.setter` for setter methods
- Allows controlled access to attributes:

```python
class BankAccount:
    def __init__(self, balance, acct_holder):
        self._balance = balance
        self._account_holder = acct_holder

    @property
    def account_holder(self):
        return self._account_holder

    @account_holder.setter
    def account_holder(self, value):
        self._account_holder = value

account_1 = BankAccount(100, "John")
account_1.account_holder = "Jane"
print(f'Account holder: {account_1.account_holder}')

```

# Special Methods

- Special methods, also known as "dunder" (double underscore) methods, customize object behavior
- They are automatically called in specific situations
- Here's a table of commonly used dunder methods:

| Method | Description | When it's called |
|--------|-------------|------------------|
| `__init__(self, ...)` | Initializes a new object | When creating a new instance |
| `__str__(self)` | Returns a string representation | When using `str()` or `print()` |
| `__repr__(self)` | Returns a detailed string representation | In interactive mode or when using `repr()` |
| `__eq__(self, other)` | Defines equality comparison | When using `==` |
| `__add__(self, other)` | Defines addition | When using `+` |
| `__sub__(self, other)` | Defines subtraction | When using `-` |
| `__mul__(self, other)` | Defines multiplication | When using `*` |
| `__truediv__(self, other)` | Defines true division | When using `/` |

## `__str__()` vs `__repr__()`

The main differences between `__str__` and `__repr__` in Python are:
1. Purpose:
    * `__str__`: Provides a human-readable, concise string representation of the object.
    * `__repr__`: Provides a more detailed, unambiguous representation, often used for debugging and development.
2. Usage:
    * `__str__`: Called by `str()` function and `print()` statement.
    * `__repr__`: Called by `repr()` function and in interactive mode when evaluating an expression.
3. Default behavior:
    * If `__str__` is not defined, Python falls back to `__repr__`.
    * If `__repr__` is not defined, a default implementation is used.
4. Convention:
    * `__str__` should be readable and concise.
    * `__repr__` should ideally contain enough information to recreate the object.

Here's a simple example to illustrate:

```python
class BankAccount:
    def __init__(self, balance, acct_holder):
        self._balance = balance
        self._account_holder = acct_holder

    def __str__(self):
        return f"{self._account_holder} has {self._balance} in their account."

    def __repr__(self):
        return f"BankAccount(balance={self._balance}, acct_holder={self._account_holder})"

b = BankAccount(100, "John")
print(b)
print(repr(b))
# Output:
# John has 100 in their account.
# BankAccount(balance=100, acct_holder=John)
```

## `__eq__()` Example

```python
class BankAccount:
    def __init__(self, balance, acct_holder):
        self._balance = balance
        self._account_holder = acct_holder

    def __eq__(self, other):
        if isinstance(other, BankAccount):
            if self._balance == other._balance
                return True
            else:
                return False
        else:
            return False

account_1 = BankAccount(100, "John")
account_2 = BankAccount(100, "John")
account_3 = BankAccount(200, "Jane")

print(account_1 == account_2)
print(account_1 == account_3)

# Output:
# True
# False
```

## `__add__()` Example

```python
class BankAccount:
    def __init__(self, balance, acct_holder):
        if not isinstance(balance, (int, float)) or balance < 0:
            raise ValueError("Balance must be a non-negative number")
        if not isinstance(acct_holder, str) or not acct_holder:
            raise ValueError("Account holder must be a non-empty string")
        self._balance = balance
        self._account_holder = acct_holder

    def __add__(self, other):
        if not isinstance(other, BankAccount):
            raise TypeError("Can only add BankAccount objects")
        if self._account_holder != other._account_holder:
            raise ValueError("Can only add accounts with the same account holder")
        return BankAccount(self._balance + other._balance, self._account_holder)

account_1 = BankAccount(100, "John")
account_2 = BankAccount(200, "John")
account_3 = BankAccount(300, "Jane")
same_acct = account_1 + account_2
print(same_acct)
diff_acct = account_1 + account_3
print(diff_acct)

# Output:
# John has 300 in their account.
# ValueError: Can only add accounts with the same account holder
```

## Review


* Classes/OOP allows us to model real-world entities as objects in code
* An object is a single instance of a class
* The data contained within the object are called the object’s attributes; the procedures used/coded for the object are called the object’s methods. Together, the object’s attributes and methods are called the object’s members.
* All instances of a class have the same attribute names and same methods (behaviors).
* The `class` keyword is used to start a class definition.
* `__init__()` is a special method that is called when a new instance of a class is created. It is used to initialize the object's attributes.
* New instances of an object are created by calling the class name as if it were a function and passing the arguments needed to initialize the object's attributes.
* Python classes do not have truely private attributes. A user can access the attributes of an object if they know the name of the attribute, but it's a best practice to code properties for the class to control access to the attributes.
* It's common practice to override python default dunder methods to customize the behavior of the class, thus allowing new and powerful interactions.






