# Abstract Base Classes

## Client and Interface Definition

* Interfaces in programming are abstract types that define a contract or a set of methods that a class must implement. They serve as a blueprint for classes, specifying what methods the class should have without dictating how these methods should be implemented.
* Client: Code that calls the interface
    * Does not care about implementation and doesn't have to even be the same language (Think web API)

## Duck Typing Review

* When you hear "duck typing", think "attributes are more important than type"
* Python is dynamically typed, and utilizes duck typing all over the place
    * `len()` function on strings, lists, etc.
* Regular methods and attributes can be used to enable duck typing.
* Special methods are more commonly used to enable duck typing by adding standardized behavior to user-defined types(objects)
    * Sets of some of these special methods are make up python's predefined protocols(covered in detail later)
        * Collection Protocol
            * `__iter__`
            * `__len__`
            * `__contains__`
        * Iterator Protocol
            * `__iter__`
            * `__next__`

```python
class Duck:
    def make_sound(self):
        print("The duck quacks")

    def swim(self):
        print("The duck is swimming")

    def fly(self):
        print("The duck is flying")

class Penguin:
    def make_sound(self):
        print("The penguin makes whatever sound penguins make")

    def swim(self):
        print("The penguin is swimming")

class Ostritch:
    def make_sound(self):
        print("The ostrich honks")

for bird in [Duck(), Penguin(), Ostritch()]:
    bird.make_sound() # Duck typed for make_sound()

# The duck quacks
# The penguin makes whatever sound penguins make
# The ostrich honks

for bird in [Duck(), Penguin(),]:
    bird.swim() # Duck typed for swim()

# The duck is swimming
# The penguin is swimming
```

Static type checking using the above concepts is not possible. That's where Abstract Base Classes and Protocols come in.


## Protocols vs Abstract Base Classes

* Nominal subtyping is when a class explicitly inherits from another class.
* Structural subtyping is when a class is considered a subtype of another class if it has the same attributes and methods as the other class, regardless of whether it inherits from it. (Think "Static duck typing")
* IDE's and programs like Mypy can perform static type checking for both nominal and structural subtyping. (This is a huge benefit to developers)


### Abstract Base Classes

* Abstract Base Classes(ABCs) are classes that define a set of methods that must be implemented by any class that inherits from them.
* ABCs use the inheritance mechanism (nominal subtyping) to ensure that subclasses implement the required methods.
* ABCs are suitable when you have control over the class hierarchy and want to define a consistent interface across subclasses.

The two common ways to create ABCs are:

* Using the `abc` module `from abc import ABC, abstractmethod`
* Using the `collections.abc` module `from collections.abc import Iterable, Collection`

```python
from abc import ABC, abstractmethod

class Bird(ABC):
    @abstractmethod
    def make_sound(self):
        pass

class Duck(Bird):
    def make_sound(self):
        print("The duck quacks")

    def fly(self):
        print("The duck is flying")

class Penguin(Bird):
    def make_sound(self):
        print("The penguin makes whatever sound penguins make")

    def swim(self):
        print("The penguin is swimming")
```

This is a generic ABC example.  You don't "get" anything from inheriting from `ABC`.  It's just a way to define a set of methods that must be implemented by any class that inherits from it. `@abstractmethod` is a decorator that is used to mark a method as abstract. You also can't create an instance of `Bird` because it is an abstract class.

If either of the subclasses of `Bird` do not implement the `make_sound` method, the following error will be raised:

`TypeError: Can't instantiate abstract class Duck with abstract methods make_sound`


```python
from collections.abc import Iterable

def process_items(items: Iterable[int]) -> None:
    for item in items:
        print(item)

# This function can now accept lists, tuples, sets, or any custom iterable

```

The collections.abc module is used to define a set of protocols that can be used to check if an object conforms to a protocol.
Since lists, tuples, sets, and other built-in types already implement the `__iter__` and `__len__` methods, they can be used as arguments to the `process_items()` function.
This also enables static type checking with IDEs and other tools.


### Protocols

* Protocols use structural subtyping(static duck typing) to check if an object conforms to a protocol.
* Protocols are useful in scenarios where modifying class hierarchies is impractical or when there’s no clear inheritance relationship between classes.
* If two object support a protocol (Python defined or user defined), then they are considered to be duck-typed and compatible.
* To use "capital P" Protocols, you need to import the `typing` module and use the `Protocol` class.

```python
from typing import Protocol, List

class Drawable(Protocol):
    def draw(self) -> None:
        ...

class Circle:
    def draw(self) -> None:
        print("Drawing a circle")

class Square:
    def draw(self) -> None:
        print("Drawing a square")

class Triangle:
    def draw(self) -> None:
        print("Drawing a triangle")

def draw_shapes(shapes: List[Drawable]) -> None:
    for shape in shapes:
        shape.draw()

# Usage
shapes: List[Drawable] = [Circle(), Square(), Triangle()]
draw_shapes(shapes)

# Drawing a circle
# Drawing a square
# Drawing a triangle
```

**Explanation:**

* `Drawable` Protocol:
    * We define a `Drawable` protocol with a single method `draw()`.
    * The `...` in the method body is a placeholder, indicating that the method doesn't need an implementation in the protocol.
* Concrete Classes:
    * `Circle`, `Square`, and `Triangle` classes are defined, each with a `draw()` method.
    * These classes don't explicitly inherit from `Drawable`, but they conform to its interface.
* `draw_shapes()` Function:
    * This function takes a list of `Drawable` objects.
    * It demonstrates how we can use the `Drawable` protocol as a type hint.
* Usage:
    * We create a list of shapes and pass it to `draw_shapes()`.
    * The type checker recognizes `Circle`, `Square`, and `Triangle` as valid `Drawable` types because they have a `draw()` method.
* Key Points:
    * Protocols define a set of methods that a class should have, without requiring explicit inheritance.
    * Classes that implement the methods defined in a protocol are considered to be compatible with that protocol, even if they don't explicitly inherit from it.
    * This allows for more flexible and decoupled designs, as classes can be compatible with multiple protocols without complex inheritance hierarchies.
    * Protocols are particularly useful for type hinting and static type checking, helping catch potential errors before runtime.
    * In this example, `Drawable` serves as a contract that any class with a `draw()` method can fulfill, allowing for a flexible way to define "drawable" objects without forcing a specific inheritance structure.


External Resources:
- [Real Python - Duck Typing in Python](https://realpython.com/duck-typing-python/)
- [Real Python - Python Protocol](https://realpython.com/python-protocol/)
- [Python ABCs](https://docs.python.org/3/library/collections.abc.html#collections-abstract-base-classes)




Credits:

- [ArjanCodes YT](https://youtu.be/xvb5hGLoK0A?si=miPVoF7m2g63ZyZl)
- [ArjanCodes GitHub](https://github.com/ArjanCodes/2021-protocol-vs-abc/tree/main/protocol/iot)
