"""
abc vs collections.abc

- abc is a standard library module that provides the infrastructure for defining abstract base classes (ABCs) in Python.
- collections.abc is a module that provides a set of abstract base classes for collections in Python.


"""

# Example 1: Using standard abc
from abc import ABC, abstractmethod

class MyListABC(ABC):
    @abstractmethod
    def __getitem__(self, index):
        pass

    @abstractmethod
    def __len__(self):
        pass

class MyCustomList(MyListABC):
    def __init__(self, data):
        self._data = list(data)

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)

# Usage
custom_list_abc = MyCustomList([1, 2, 3, 4, 5])
print(len(custom_list_abc))  # 5
print(custom_list_abc[2])    # 3

# But we're missing a lot of expected list-like behavior:
# This will raise an AttributeError
# print(list(custom_list_abc))

# Example 2: Using collections.abc
from collections.abc import Sequence

class MyCustomSequence(Sequence):
    def __init__(self, data):
        self._data = list(data)

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)

# Usage
custom_sequence = MyCustomSequence([1, 2, 3, 4, 5])
print(len(custom_sequence))  # 5
print(custom_sequence[2])    # 3

# Now we get a lot more functionality for free:
print(list(custom_sequence))  # [1, 2, 3, 4, 5]
print(3 in custom_sequence)   # True
print(custom_sequence.index(4))  # 3
print(custom_sequence.count(5))  # 1

# We can even use slicing!
print(custom_sequence[1:4])  # [2, 3, 4]

# And reverse
print(list(reversed(custom_sequence)))  # [5, 4, 3, 2, 1]