import argparse
from abc import ABC, abstractmethod

class Account(ABC):
    def __init__(self, account_number, initial_balance):
        self._account_number = account_number
        self._balance = initial_balance

    @abstractmethod
    def deposit(self, amount):
        pass

    @abstractmethod
    def withdraw(self, amount):
        pass

    @abstractmethod
    def check_balance(self):
        pass

class SavingsAccount(Account):
    def deposit(self, amount):
        self._balance += amount
        return f"Deposited ${amount}. New balance: ${self._balance}"

    def withdraw(self, amount):
        if self._balance >= amount:
            self._balance -= amount
            return f"Withdrew ${amount}. New balance: ${self._balance}"
        return "Insufficient funds"

    def check_balance(self):
        return f"Current balance: ${self._balance}"

class CheckingAccount(Account):
    def deposit(self, amount):
        self._balance += amount
        return f"Deposited ${amount}. New balance: ${self._balance}"

    def withdraw(self, amount):
        if self._balance >= amount:
            self._balance -= amount
            return f"Withdrew ${amount}. New balance: ${self._balance}"
        return "Insufficient funds"

    def check_balance(self):
        return f"Current balance: ${self._balance}"

class Customer:
    def __init__(self, name):
        self.name = name
        self.accounts = {}

    def add_account(self, account_number, account_type, initial_balance):
        if account_type.lower() == "savings":
            self.accounts[account_number] = SavingsAccount(account_number, initial_balance)
        elif account_type.lower() == "checking":
            self.accounts[account_number] = CheckingAccount(account_number, initial_balance)
        else:
            return "Invalid account type"
        return f"Account {account_number} added successfully"

    def deposit(self, account_number, amount):
        if account_number in self.accounts:
            return self.accounts[account_number].deposit(amount)
        return "Account not found"

    def withdraw(self, account_number, amount):
        if account_number in self.accounts:
            return self.accounts[account_number].withdraw(amount)
        return "Account not found"

    def check_balance(self, account_number):
        if account_number in self.accounts:
            return self.accounts[account_number].check_balance()
        return "Account not found"

class Bank:
    def __init__(self):
        self.customers = {}

    def add_customer(self, customer_name):
        if not customer_name.strip():
            return "Customer name cannot be blank"
        if customer_name not in self.customers:
            self.customers[customer_name] = Customer(customer_name)
            return f"Customer {customer_name} added successfully"
        return "Customer already exists"

    def get_customer(self, customer_name):
        return self.customers.get(customer_name, None)


def main():
    parser = argparse.ArgumentParser(description="Bank Application")
    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # Add customer
    add_customer_parser = subparsers.add_parser("add_customer", help="Add a new customer")
    add_customer_parser.add_argument("customer_name", help="Name of the customer")

    # Add account
    add_account_parser = subparsers.add_parser("add_account", help="Add a new account")
    add_account_parser.add_argument("customer_name", help="Name of the customer")
    add_account_parser.add_argument("account_number", help="Account number")
    add_account_parser.add_argument("account_type", choices=["savings", "checking"], help="Type of account")
    add_account_parser.add_argument("initial_balance", type=float, help="Initial balance")

    # Deposit
    deposit_parser = subparsers.add_parser("deposit", help="Deposit money")
    deposit_parser.add_argument("customer_name", help="Name of the customer")
    deposit_parser.add_argument("account_number", help="Account number")
    deposit_parser.add_argument("amount", type=float, help="Amount to deposit")

    # Withdraw
    withdraw_parser = subparsers.add_parser("withdraw", help="Withdraw money")
    withdraw_parser.add_argument("customer_name", help="Name of the customer")
    withdraw_parser.add_argument("account_number", help="Account number")
    withdraw_parser.add_argument("amount", type=float, help="Amount to withdraw")

    # Check balance
    check_balance_parser = subparsers.add_parser("check_balance", help="Check account balance")
    check_balance_parser.add_argument("customer_name", help="Name of the customer")
    check_balance_parser.add_argument("account_number", help="Account number")

    args = parser.parse_args()

    bank = Bank()

    if args.command == "add_customer":
        print(bank.add_customer(args.customer_name))

    elif args.command == "add_account":
        customer = bank.get_customer(args.customer_name)
        if customer:
            print(customer.add_account(args.account_number, args.account_type, args.initial_balance))
        else:
            print("Customer not found")

    elif args.command == "deposit":
        customer = bank.get_customer(args.customer_name)
        if customer:
            print(customer.deposit(args.account_number, args.amount))
        else:
            print("Customer not found")

    elif args.command == "withdraw":
        customer = bank.get_customer(args.customer_name)
        if customer:
            print(customer.withdraw(args.account_number, args.amount))
        else:
            print("Customer not found")

    elif args.command == "check_balance":
        customer = bank.get_customer(args.customer_name)
        if customer:
            print(customer.check_balance(args.account_number))
        else:
            print("Customer not found")

    else:
        parser.print_help()

if __name__ == "__main__":
    main()


#python bank_app.py add_customer "John Doe"
# python bank_app.py add_account "John Doe" "12345" savings 1000
# python bank_app.py deposit "John Doe" "12345" 500
# python bank_app.py withdraw "John Doe" "12345" 200
# python bank_app.py check_balance "John Doe" "12345"