import cmd
from abc import ABC, abstractmethod

class Account(ABC):
    def __init__(self, account_number, initial_balance):
        self._account_number = account_number
        self._balance = initial_balance

    @abstractmethod
    def deposit(self, amount):
        pass

    @abstractmethod
    def withdraw(self, amount):
        pass

    @abstractmethod
    def check_balance(self):
        pass

class SavingsAccount(Account):
    def deposit(self, amount):
        self._balance += amount
        return f"Deposited ${amount}. New balance: ${self._balance}"

    def withdraw(self, amount):
        if self._balance >= amount:
            self._balance -= amount
            return f"Withdrew ${amount}. New balance: ${self._balance}"
        return "Insufficient funds"

    def check_balance(self):
        return f"Current balance: ${self._balance}"

class CheckingAccount(Account):
    def deposit(self, amount):
        self._balance += amount
        return f"Deposited ${amount}. New balance: ${self._balance}"

    def withdraw(self, amount):
        if self._balance >= amount:
            self._balance -= amount
            return f"Withdrew ${amount}. New balance: ${self._balance}"
        return "Insufficient funds"

    def check_balance(self):
        return f"Current balance: ${self._balance}"

class Customer:
    def __init__(self, name):
        self.name = name
        self.accounts = {}

    def add_account(self, account_number, account_type, initial_balance):
        if account_type.lower() == "savings":
            self.accounts[account_number] = SavingsAccount(account_number, initial_balance)
        elif account_type.lower() == "checking":
            self.accounts[account_number] = CheckingAccount(account_number, initial_balance)
        else:
            return "Invalid account type"
        return f"Account {account_number} added successfully"

    def deposit(self, account_number, amount):
        if account_number in self.accounts:
            return self.accounts[account_number].deposit(amount)
        return "Account not found"

    def withdraw(self, account_number, amount):
        if account_number in self.accounts:
            return self.accounts[account_number].withdraw(amount)
        return "Account not found"

    def check_balance(self, account_number):
        if account_number in self.accounts:
            return self.accounts[account_number].check_balance()
        return "Account not found"

class Bank:
    def __init__(self):
        self.customers = {}

    def add_customer(self, customer_name):
        if not customer_name.strip():
            return "Customer name cannot be blank"
        if customer_name not in self.customers:
            self.customers[customer_name] = Customer(customer_name)
            return f"Customer {customer_name} added successfully"
        return "Customer already exists"

    def get_customer(self, customer_name):
        return self.customers.get(customer_name, None)

class BankApp(cmd.Cmd):
    intro = "Welcome to the Bank App. Type 'help' for a list of commands."
    prompt = "(bank) "

    def __init__(self):
        super().__init__()
        self.bank = Bank()

    def do_add_customer(self, arg):
        """Add a new customer: add_customer <customer_name>"""
        customer_name = arg.strip()
        print(self.bank.add_customer(customer_name))

    def do_add_account(self, arg):
        """Add a new account: add_account <customer_name> <account_number> <account_type> <initial_balance>"""
        args = arg.split()
        if len(args) != 4:
            print("Invalid number of arguments")
            return
        customer_name, account_number, account_type, initial_balance = args
        customer = self.bank.get_customer(customer_name)
        if customer:
            print(customer.add_account(account_number, account_type, float(initial_balance)))
        else:
            print("Customer not found")

    def do_deposit(self, arg):
        """Deposit money: deposit <customer_name> <account_number> <amount>"""
        args = arg.split()
        if len(args) != 3:
            print("Invalid number of arguments")
            return
        customer_name, account_number, amount = args
        customer = self.bank.get_customer(customer_name)
        if customer:
            print(customer.deposit(account_number, float(amount)))
        else:
            print("Customer not found")

    def do_withdraw(self, arg):
        """Withdraw money: withdraw <customer_name> <account_number> <amount>"""
        args = arg.split()
        if len(args) != 3:
            print("Invalid number of arguments")
            return
        customer_name, account_number, amount = args
        customer = self.bank.get_customer(customer_name)
        if customer:
            print(customer.withdraw(account_number, float(amount)))
        else:
            print("Customer not found")

    def do_check_balance(self, arg):
        """Check account balance: check_balance <customer_name> <account_number>"""
        args = arg.split()
        if len(args) != 2:
            print("Invalid number of arguments")
            return
        customer_name, account_number = args
        customer = self.bank.get_customer(customer_name)
        if customer:
            print(customer.check_balance(account_number))
        else:
            print("Customer not found")

    def do_quit(self, arg):
        """Quit the application"""
        print("Thank you for using the Bank App. Goodbye!")
        return True

if __name__ == "__main__":
    BankApp().cmdloop()