"""
This module demonstrates dynamic class creation using metaclasses in Python.

Dynamic class creation allows for the programmatic definition of classes at runtime,
which is useful for creating flexible and adaptable code structures. This technique
is often used in ORMs (Object-Relational Mappers) to create model classes based on
database schemas.

The module defines a function that creates a model class with specified fields,
simulating a simple ORM-like behavior.
"""

def create_model_class(table_name, **fields):
    def create_property(field):
        # Create a property that accesses a private attribute
        return property(lambda self: getattr(self, f'_{field}', None))

    class_attrs = {
        '__tablename__': table_name,
        **{field_name: create_property(field_name) for field_name in fields}
    }

    # Dynamically create and return a new class
    return type(table_name, (object,), class_attrs)

# Usage
User = create_model_class('users', id=int, name=str, email=str)
user = User()
print(User.__tablename__)  # Output: users
# This demonstrates that a User class was dynamically created with a 'users' table name