import math

# Part 1: Building a Calculator

class InputValidator:
    @staticmethod
    def validate_number(value):
        if not isinstance(value, (int, float)):
            raise ValueError("Input must be a number")

class Calculator:
    def __init__(self):
        self._metadata = {
            'add': 0, 'subtract': 0, 'multiply': 0, 'divide': 0,
            'square_root': 0, 'power': 0
        }

    def _update_metadata(func):
        def wrapper(self, *args, **kwargs):
            self._metadata[func.__name__] += 1
            return func(self, *args, **kwargs)
        return wrapper

    @_update_metadata
    def add(self, a, b):
        InputValidator.validate_number(a)
        InputValidator.validate_number(b)
        return a + b

    @_update_metadata
    def subtract(self, a, b):
        InputValidator.validate_number(a)
        InputValidator.validate_number(b)
        return a - b

    @_update_metadata
    def multiply(self, a, b):
        InputValidator.validate_number(a)
        InputValidator.validate_number(b)
        return a * b

    @_update_metadata
    def divide(self, a, b):
        InputValidator.validate_number(a)
        InputValidator.validate_number(b)
        if b == 0:
            raise ValueError("Cannot divide by zero")
        return a / b

    @_update_metadata
    def square_root(self, a):
        InputValidator.validate_number(a)
        if a < 0:
            raise ValueError("Cannot calculate square root of a negative number")
        return math.sqrt(a)

    @_update_metadata
    def power(self, base, exponent):
        InputValidator.validate_number(base)
        InputValidator.validate_number(exponent)
        return math.pow(base, exponent)

    def clear_metadata(self):
        for key in self._metadata:
            self._metadata[key] = 0

    def get_metadata(self):
        return self._metadata.copy()

# Part 2: Building a Person Object

class Person:
    def __init__(self, first_name, last_name, age, email):
        self._first_name = first_name
        self._last_name = last_name
        self._age = age
        self._email = email
        self._job_title = None
        self._employer = None
        self._is_employed = False
        self._friends = []

    @property
    def full_name(self):
        return f"{self._first_name} {self._last_name}"

    def introduce(self):
        return f"Hello, I'm {self.full_name}. I'm {self._age} years old."

    def celebrate_birthday(self):
        self._age += 1
        return f"Happy birthday to me! I'm now {self._age} years old."

    def hire(self, job_title, employer):
        self._job_title = job_title
        self._employer = employer
        self._is_employed = True

    def fire(self):
        self._job_title = None
        self._employer = None
        self._is_employed = False

    def add_friend(self, friend):
        if friend not in self._friends:
            self._friends.append(friend)

    def remove_friend(self, friend):
        if friend in self._friends:
            self._friends.remove(friend)

    def list_friends(self):
        return [friend.full_name for friend in self._friends]

class Student(Person):
    def __init__(self, first_name, last_name, age, email, student_id):
        super().__init__(first_name, last_name, age, email)
        self._student_id = student_id
        self._grades = {}

    def add_grade(self, subject, grade):
        self._grades[subject] = grade

    def get_gpa(self):
        if not self._grades:
            return 0
        return sum(self._grades.values()) / len(self._grades)

class Employee(Person):
    def __init__(self, first_name, last_name, age, email, employee_id, salary):
        super().__init__(first_name, last_name, age, email)
        self._employee_id = employee_id
        self._salary = salary

    def give_raise(self, amount):
        self._salary += amount

    def get_salary(self):
        return self._salary

# Main execution
if __name__ == "__main__":
    # Test Calculator
    calc = Calculator()
    print(calc.add(5, 3))  # 8
    print(calc.subtract(10, 4))  # 6
    print(calc.multiply(2, 6))  # 12
    print(calc.divide(15, 3))  # 5.0
    print(calc.square_root(16))  # 4.0
    print(calc.power(2, 3))  # 8.0
    print(calc.get_metadata())  # Shows usage count for each method
    calc.clear_metadata()
    print(calc.get_metadata())  # All counts reset to 0

    # Test Person, Student, and Employee
    person = Person("John", "Doe", 30, "john@example.com")
    print(person.introduce())
    print(person.celebrate_birthday())

    person.hire("Developer", "Tech Co.")
    print(f"{person.full_name} is employed: {person._is_employed}")
    person.fire()
    print(f"{person.full_name} is employed: {person._is_employed}")

    friend1 = Person("Jane", "Smith", 28, "jane@example.com")
    friend2 = Person("Mike", "Johnson", 32, "mike@example.com")
    person.add_friend(friend1)
    person.add_friend(friend2)
    print(f"{person.full_name}'s friends: {person.list_friends()}")

    student = Student("Alice", "Brown", 20, "alice@example.com", "S12345")
    student.add_grade("Math", 85)
    student.add_grade("Science", 92)
    print(f"{student.full_name}'s GPA: {student.get_gpa()}")

    employee = Employee("Bob", "Wilson", 35, "bob@example.com", "E67890", 50000)
    print(f"{employee.full_name}'s salary: ${employee.get_salary()}")
    employee.give_raise(5000)
    print(f"{employee.full_name}'s new salary: ${employee.get_salary()}")


# This solution covers all the exercises:

# 1. The `Calculator` class is implemented with basic arithmetic operations.
# 2. It's extended with `square_root` and `power` methods.
# 3. Input validation is implemented using a separate `InputValidator` class.
# 4. A metadata feature is added to log method usage frequency.
# 5. The `Person` class is created with basic attributes.
# 6. Methods for self-introduction and birthday celebration are added, with proper encapsulation.
# 7. Employment-related attributes and methods are added to the `Person` class.
# 8. Friend-tracking functionality is implemented in the `Person` class.
# 9. `Student` and `Employee` subclasses are created with additional attributes and methods.

# The main execution block at the end demonstrates how to use these classes and methods. This solution provides a comprehensive implementation of all the requested features while following object-oriented design principles.