
### Custom Exceptions

You can define your own custom exception classes to represent specific error conditions in your code.  This makes your error handling more precise and informative.  Custom exceptions should inherit from `Exception` or a more specific built-in exception class.

```python
class ValidationError(Exception):
    """Custom exception for validation errors."""
    def __init__(self, message, field=None):
        super().__init__(message)
        self.field = field

def validate_age(age):
    if not isinstance(age, int):
        raise ValidationError("Age must be an integer.", field="age")
    if age < 0:
        raise ValidationError("Age cannot be negative.", field="age")
    if age > 150:
        raise ValidationError("Age seems unreasonably high.", field="age")

try:
    validate_age("thirty")
except ValidationError as e:
    print(f"Validation Error: {e}, Field: {e.field}")

try:
    validate_age(-5)
except ValidationError as e:
    print(f"Validation Error: {e}, Field: {e.field}")
```

**Benefits of Custom Exceptions:**

*   **Clarity:**  Clearly indicate specific error types in your application.
*   **Organization:**  Group related errors under a common base class.
*   **Flexibility:**  Add custom attributes (like `field` in the example) to provide more context about the error.
*   **Maintainability:** Easier to catch and handle specific error conditions.

**Trick:** You can create a hierarchy of custom exceptions to represent different levels of error granularity.

```python
class MyBaseError(Exception):
    pass

class NetworkError(MyBaseError):
    pass

class TimeoutError(NetworkError):
    pass

class ConnectionError(NetworkError):
    pass
```
