# Python Decorators:

## Learning Objectives

By the end of this lesson, students will be able to:

1. Understand the concept of first-class functions in Python
2. Explain and implement inner functions
3. Define and use decorators in Python
4. Handle arguments in decorated functions
5. Preserve function metadata when using decorators
6. Recognize common use cases for decorators

## First-Class Functions

Python treats functions as first-class objects, which means they can be:

- Assigned to variables
- Passed as arguments to other functions
- Returned from functions
- Used in expressions
- Modified and stored in data structures

```python
# First-Class Functions: Function as objects, arguments, return values

def hello(name):
    return f"hello {name}"


def hello_friend(name):
    return f"hello {name}, how's it going friend?"


def greet(func):
    return func("Jimmy")


print(greet(hello))  # hello Jimmy
print(greet(hello_friend))  # hello Jimmy, how's it going friend?
```

```python
# Inner Functions


def parent():
    print("Printing from parent()")

    def first_child():
        print("Printing from first_child()")

    def second_child():
        print("Printing from second_child()")

    second_child()  # Printing from second_child()
    first_child()  # Printing from first_child()


parent()
# Printing from parent()
# Printing from second_child()
# Printing from first_child()
second_child()  # NameError: name 'second_child' is not defined
```

```python
# Simple Decorator (Legacy)

def decorator(func):
    def wrapper():
        print("Something is happening before the function is called.")
        func()
        print("Something is happening after the function is called.")

    return wrapper


def do_stuff():
    print("I'm doing stuff")


do_stuff = decorator(do_stuff)
do_stuff()
# "Something is happening before the function is called."
# "I'm doing stuff"
# "Something is happening after the function is called."
```

```python
# Simple Decorator (Modern pythonic approach)
def decorator(func):
    def wrapper():
        print("Something is happening before the function is called.")
        func()
        print("Something is happening after the function is called.")

    return wrapper


@decorator
def do_stuff():
    print("I'm doing stuff")


do_stuff() # Same output
```

```python
# Decorator with arguments
def do_twice(func):
    def wrapper_do_twice(*args, **kwargs):
        func(*args, **kwargs)
        func(*args, **kwargs)

    return wrapper_do_twice


@do_twice
def hello(name):
    print(f"I'm saying hello {name}")


hello("Jimmy")

```


## Examples

### @property

* Allows us to define a method that can be accessed like an attribute
* Can be used to define a getter, setter, and deleter for an attribute

[property.py](./code/decorators/property.py)

### functools.wraps() (Function Signature Preservation)

* Allows us to preserve metadata when defining decorators
* Example also illustrates how to pass return values from decorated functions

[know_thyself.py](./code/decorators/know-thyself.py)

### @timer

* Allows us to time how long a function takes to run

[timer.py](./code/decorators/timer.py)

### Enable Logging

* Allows us to use conditional logging

[my_logging.py](./code/decorators/my_logging.py)

### Fibonacci Cache

* Allows us to cache the results of a function so we don't have to recompute them

* [fib-cache_1.py](./code/decorators/fib-cache_1.py)
* [fib-cache_2.py](./code/decorators/fib-cache_2.py)

### Dataclass

* Allows us to define a class with automatically generated special methods

* [dataclass_1.py](./code/decorators/dataclass_1.py)
* [dataclass_2.py](./code/decorators/dataclass_2.py)

## Key Takeaways

* Decorator is a function that takes another function as input and returns a new function.
* Can be used to add functionality to a function without modifying the function itself
* Can simplify code and leverage powerful features easily
* **Caveat:** If overdone or poorly documented, can make code harder to read and understand

??? Tip "Additional Resources"

    - [Real Python - Decorators](https://realpython.com/primer-on-python-decorators/)
    - [Real Python - Timer](https://realpython.com/python-timer/)