# Duck Typing

## Learning Objectives

By the end of this lesson, students will be able to:

1. Understand the concept of duck typing and evaluate its advantages and disadvantages
2. Explore how Python's built-in classes and tools leverage duck typing for flexibility
3. Examine the role of special methods and protocols in facilitating duck typing
4. Identify and compare alternative approaches to duck typing available in Python

## Review of Type Systems

* **Dynamic Typing:**
    * Python is dynamically typed
    * The interpreter checks only when the code is run
    * type of variable is allowed to change

* **Static Typing:**
  * type checks are performed at compile time (Java, C, C++)
  * changing variable type is more strict if not restricted completely ()

## Duck Typing

!!! Tip "From the Python Docs (on duck typing)"
    A programming style which does not look at an object’s type to determine if it has the right interface; instead, the method or attribute is simply called or used (“If it looks like a duck and quacks like a duck, it must be a duck.”) By emphasizing interfaces rather than specific types, well-designed code improves its flexibility by allowing polymorphic substitution.

    Duck-typing avoids tests using `type()` or `isinstance()`. (Note, however, that duck-typing can be complemented with abstract base classes.) Instead, it typically employs `hasattr()` tests or EAFP programming.

Duck typing is a type system where an object is considered compatible with a given type if it has all the methods and attributes that the type requires. This type system supports the ability to use objects of independent and decoupled classes in a specific context as long as they adhere to some common interface.



```python title="duck 1"
class Dog:
    def run(self):
        print("The dog is running")

    def make_sound(self):
        print("The dog is barking")

class Cat:
    def run(self):
        print("The cat is running")

    def make_sound(self):
        print("The cat is meowing")

class Horse:
    def run(self):
        print("The horse is running")

    def make_sound(self):
        print("The horse is neighing")

# Because they share the same interface, we can use them in a familiar and flexible way

mammals = [Dog(), Cat(), Horse()]

for mammal in mammals:
    mammal.make_sound()
    mammal.run()

# The dog is barking
# The dog is running
# The cat is meowing
# The cat is running
# The horse is neighing
# The horse is running
```

Python doesn't care what object mammal is, it just calls the expected methods. If the object provides the expected methods, it will work.

## Pros and Cons

### Pros

* **Flexibility:** Duck typing allows for more flexible code as it does not require explicit type declarations. This can lead to more reusable and adaptable code.
* **Simplicity:** You can simplify code by focusing on the required behavior rather than thinking of specific types, classes, and the relationships between them. This allows for more concise and expressive code.
* **Prototyping:** You can quickly create objects that exhibit the necessary behavior without complex type definitions. This is useful during the initial stages of development, where you may not have fully fleshed out class hierarchies or interfaces.

### Cons

* **Runtime Errors:** You might face errors related to missing methods or attributes, which may only appear at runtime. This can lead to unexpected behavior or crashes if an object doesn’t conform to the expected behavior.
* **Lack of explicitness:** You might make your code less explicit and more challenging to understand. The lack of explicit interface definitions might make it more difficult to grasp what behavior an object must exhibit.

## Duck Typing in Python Built-ins

* *lists*, *dicts*, *strings*, *tuples* all support *iteration*, *sorting*, *reversing*, etc..

```python title="duck 2"
numbers = [1, 2, 3]
person = ("Jane", 25, "Python Dev")
letters = "abc"
ordinals = {"one": "first", "two": "second", "three": "third"}
even_digits = {2, 4, 6, 8}
collections = [numbers, person, letters, ordinals, even_digits]

for collection in collections:
    print()
    for value in collection:
        print(value, end=" ")
print()
# 1 2 3
# Jane 25 Python Dev
# a b c
# one two three
# 8 2 4 6
```

## Special Methods and Protocols

!!! Note "Reminder"
    * **Special methods:** The double underscore methods provided by Python (e.g. `__init__`, `__str__`, `__len__`, etc.)
    * **Protocols:** Sets of special methods that support specific features.
        * Iteration Protocol: `__iter__` [ `__next__` ]
        * Descriptor Protocol: `__get__`, `__set__`, `__delete__`
        * Sequence Protocol: `__getitem__`, `__len__`, `__contains__`, `__iter__`
        * Context Manager Protocol: `__enter__`, `__exit__`

**What makes the below custom queue class "duck typed"?**

```python title="Custom Queue Class"
from collections import deque

class Queue:
    def __init__(self):
        self._elements = deque()

    def enqueue(self, element):
        self._elements.append(element)

    def dequeue(self):
        return self._elements.popleft()

    def __iter__(self):
        return iter(self._elements)

    def __len__(self):
        return len(self._elements)

    def __reversed__(self):
        return reversed(self._elements)

    def __contains__(self, element):
        return element in self._elements

queue = Queue()
queue.enqueue(1)
queue.enqueue(2)
queue.enqueue(3)

print(*queue)
print(len(queue))
print(3 in queue)
# 1 2 3
# 3
# True
```

Now your queue class can be used in any context that expects an iterable, a len, and a contains check.


## Practical Exercises

### 1. Identify and fix the issue in the following code:
```python
class Duck:
    def quack(self):
        print("Quack!")

class Goose:
    def honk(self):
        print("Honk!")

def make_sound(bird):
    bird.quack()

duck = Duck()
goose = Goose()
make_sound(duck)
make_sound(goose) # This line will raise an AttributeError
```

??? Warning "1. Solution"

    ```python
    class Duck:
        def make_sound(self):
            print("Quack!")

    class Goose:
        def make_sound(self):
            print("Honk!")

    def make_sound(bird):
        bird.make_sound()

    duck = Duck()
    goose = Goose()
    make_sound(duck)
    make_sound(goose)
    ```

### 2. Identify and fix the issue in the following code:
```python
class Airplane:
    def fly(self):
        print("Flying!")
class Car:
    def drive(self):
        print("Driving!")

def travel(vehicle):
    if isinstance(vehicle, Airplane):
        vehicle.fly()
    elif isinstance(vehicle, Car):
        vehicle.drive()

airplane = Airplane()
car = Car()
travel(airplane)
travel(car)
```

??? Warning "2. Solution"

    ```python
    class Airplane:
        def move(self):
            print("Flying!")

    class Car:
        def move(self):
            print("Driving!")

    def travel(vehicle):
        vehicle.move()

    airplane = Airplane()
    car = Car()
    travel(airplane)
    travel(car)
    ```

### 3. Create a `Shape` class with a `draw()` method, and two subclasses `Circle` and `Square`. Implement the `draw()` method for each shape.

```python
# Your code here



# Don't edit below this line
shapes = [Circle(), Square()]
for shape in shapes:
    shape.draw()

```
??? Warning "3. Solution"

    ```python
    class Shape:
        def draw(self):
            pass

    class Circle(Shape):
        def draw(self):
            print("Drawing a circle")

    class Square(Shape):
        def draw(self):
            print("Drawing a square")

    shapes = [Circle(), Square()]
    for shape in shapes:
        shape.draw()
    ```

### 4. Create a custom `Temperature` class that supports comparison operations (`<`, `>`, `==`, etc.) based on the temperature value.

```python
class Temperature:
    # Your code here



# Don't edit below this line
t1 = Temperature(20)
t2 = Temperature(25)
print(t1 < t2) # True
print(t1 > t2) # False
print(t1 == Temperature(20)) # True
print(t1, t2) # 20°C 25°C
```

??? Warning "4. Solution"

    ```python
    class Temperature:
        def __init__(self, celsius):
            self.celsius = celsius

        def __lt__(self, other):
            return self.celsius < other.celsius

        def __gt__(self, other):
            return self.celsius > other.celsius

        def __eq__(self, other):
            return self.celsius == other.celsius


    def str(self):
        return f"{self.celsius}°C"


    t1 = Temperature(20)
    t2 = Temperature(25)
    print(t1 < t2)  # True
    print(t1 > t2)  # False
    print(t1 == Temperature(20))  # True
    print(t1, t2)  # 20°C 25°C
    ```

### 5. Implement a `TextProcessor` class that can work with different types of text sources (file, string, list of strings) using duck typing.

```python
class TextProcessor:
    # Your code here


# Don't edit below this line
with open('sample.txt', 'w') as f:
    f.write("This is a sample file.")
    file_source = open('sample.txt', 'r')
    string_source = "This is a sample string."
    list_source = ["This", "is", "a", "sample", "list."]
    processor = TextProcessor()
    print(f'file: {processor.process(file_source)}') # 5
    print(f'string: {processor.process(string_source)}') # 5
    print(f'list: {processor.process(list_source)}') # 5
    file_source.close()
```

??? Warning "5. Solution"

    ```python
    class TextProcessor:
        def process(self, source):
            try:
                # Try to read from the source (file-like object)
                content = source.read()
                return len(content.split())
            except AttributeError:
                try:
                    # Try to split the source (string-like object)
                    return len(source.split())
                except AttributeError:
                    # Assume it's an iterable of words
                    return len(list(source))

        # OR YOU COULD DO IT THIS WAY

        def process(self, source):
            if hasattr(source, 'read'):
                # File-like object
                content = source.read()
                return len(content.split())
            elif hasattr(source, 'split'):
                # String-like object
                return len(source.split())
            elif hasattr(source, '__iter__'):
                # Iterable object
                return len(list(source))
            else:
                raise TypeError("Unsupported source type")

    # Don't edit below this line
    with open('sample.txt', 'w') as f:
        f.write("This is a sample file.")
    file_source = open('sample.txt', 'r')
    string_source = "This is a sample string."
    list_source = ["This", "is", "a", "sample", "list."]
    processor = TextProcessor()
    print(f'file: {processor.process(file_source)}') # 5
    print(f'string: {processor.process(string_source)}') # 5
    print(f'list: {processor.process(list_source)}') # 5
    file_source.close()
    ```