# Object-Oriented Design

Before introducing the details of building classes in Python, it is important to realize that there are many design principles that have been defined over the years.

## DRY

One such principle is the **DRY** principle.

The **DRY** acronym stands for **D**on’t **R**epeat **Y**ourself.

* The principle was defined by Andy Hunt and Dave Thomas in the book: _The Pragmatic Programmer, Addison-Wesley Professional, 1999._

Simply put: *_if code is being copied and pasted from one part of the code to another, then the DRY principle is not being adhered to._*

The example that follows demonstrates another violation of the DRY principle:

### Poor Design
```python
#!/usr/bin/env python3

catalog = ["The Art of War", "Lord of the Rings", "Pride and Prejudice"]

def check_out_book(book_name):
    if book_name in catalog:
        print(f"Checking out {book_name}...")
    else:
        print(f"{book_name} is not in our catalog.")

def return_book(book_name):
    if book_name in catalog:
        print(f"Returning {book_name}...")
    else:
        print(f"{book_name} is not in our catalog.")

```

The above code breaks the DRY principle since there is repetition in checking if a book is in the catalog for both the check_out_book and return_book functions.

### Dry Design
```python
#!/usr/bin/env python3

catalog = ["The Art of War", "Lord of the Rings", "Pride and Prejudice"]

def is_in_catalog(book_name):
    if book_name not in catalog:
        print(f"{book_name} is not in our catalog.")
        return False
    return True

def check_out_book(book_name):
    if is_in_catalog(book_name):
        print(f"Checking out {book_name}...")

def return_book(book_name):
    if is_in_catalog(book_name):
        print(f"Returning {book_name}...")

```

In the DRY design, we've moved the logic of checking if a book is in the catalog to a separate function is_in_catalog. This function is then used by both the check_out_book and return_book functions. This makes the code more modular and avoids repetition.

## SOLID Design Principles

**SOLID** is a set of object-oriented design principles aimed at making software more understandable, flexible, and maintainable.
These principles were introduced by Robert C. Martin, often known as Uncle Bob. The mnemonic **SOLID** itself was coined by Michael Feathers.

Here's a breakdown of the **SOLID** principles:

### Single Responsibility Principle (SRP)

- Definition: A class should have only one reason to change.
- Explanation: This principle emphasizes that a class should focus on a single task or responsibility. By ensuring that a class
  addresses only one concern, we can make the system more modular and easier to understand. Changes in one section of an application
  will be less likely to affect other sections.

**Violation of SRP:**

```python
class Calculator:
    def __init__(self):
        self.users = []

    def register_user(self, user_name):
        if user_name not in self.users:
            self.users.append(user_name)
            print(f"{user_name} has been registered.")
        else:
            print(f"{user_name} is already registered.")

    def add(self, a, b):
        return a + b

```

In the example above, the `Calculator` class is responsible for both registering users of the calculator, and for preforming math operations, illustrated by the `register_user` and `add` methods.

The use of the class to manage users and do math functions is a violation of SRP, here is a better implementation:

```python
class UserManager:
    def __init__(self):
        self.users = set()

    def register_user(self, user_name):
        if user_name not in self.users:
            self.users.add(user_name)
            print(f"{user_name} has been registered.")
        else:
            print(f"{user_name} is already registered.")

    def is_registered(self, user_name):
        return user_name in self.users

class Calculator:
    def __init__(self, user_manager):
        self.user_manager = user_manager

    def add(self, a, b, user_name):
        if self.user_manager.is_registered(user_name):
            return a + b
        else:
            print(f"{user_name} is not registered.")
            return None

```

In the above example, the `Calculator` class handles the `add` method, and a separate class `UserManager` registers the users.  In our implementation users need to be registered to use the Calculator, but the `UserManager` class operates on its own can can be reused in other tasks.

And most importantly for our SRP adherence.  The classes need to be modified **ONLY** when there is a need to change how they preform their single purpose.

### Open Closed Principle (OCP)
- Definition: Software entities (classes, modules, functions, etc.) should be open for extension but closed for modification.
- Explanation: This principle means that the behavior of a module can be extended without modifying its source code. This is achieved
by using interfaces or abstract classes, allowing new functionality to be added as new classes that implement these interfaces,
rather than altering existing code.

```python
class Rectangle:
    def __init__(self, width, height):
        self.width = width
        self.height = height

class AreaCalculator:
    def compute_area(self, rectangle: Rectangle):
        return rectangle.width * rectangle.height

```

In the above example we would need to modify the `AreaCalculator` classes code to support calculating the area of other shape objects.  We cannot simply extend it.  Lets look at an abstract class example where we would be able to extend the `Shape` abstract class to support another shape and have an area method:

```python
from math import pi
from abc import ABC, abstractmethod

class Shape(ABC):
    @abstractmethod
    def area(self):
        pass

class Rectangle(Shape):
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def area(self):
        return self.width * self.height

# Add a Circle Class
class Circle(Shape):
    def __init__(self, radius):
        self.radius = radius

    def area(self):
        return pi * self.radius * self.radius

```

### Liskov Substitution Principle (LSP)

- Definition: Objects of a superclass should be able to be replaced with objects of a subclass without affecting the correctness
  of the program.
- Explanation: This principle is essential for maintaining the inherent hierarchical relationships in an inheritance structure. If a
  derived class cannot be substituted for its base class, then the inheritance hierarchy is violated, leading to unexpected behavior.

```python
class Bird:
    def fly(self):
        return "I can fly"

class Duck(Bird):
    def swim(self):
        return "I can swim"

class Ostrich(Bird):
    def fly(self):
        raise NotImplementedError("Ostriches can't fly")

def make_bird_fly(bird: Bird):
    return bird.fly()

# This works fine
duck = Duck()
print(make_bird_fly(duck))  # Output: I can fly

# This violates LSP
ostrich = Ostrich()
print(make_bird_fly(ostrich))  # Raises NotImplementedError
```
In the above example, The `Penguin` class violates LSP because children of the  `Bird` class should fly since there is a `fly` method in the parent class.

**Here is a better example which follows LSP:**

```python
class Animal:
    def move(self):
        pass

class FlyingBird(Animal):
    def move(self):
        return "I can fly"

class SwimmingBird(Animal):
    def move(self):
        return "I can swim"

class Duck(FlyingBird, SwimmingBird):
    def move(self):
        return "I can fly and swim"

class Ostrich(Animal):
    def move(self):
        return "I can run"

def make_animal_move(animal: Animal):
    return animal.move()

# Now this works for all animals
duck = Duck()
ostrich = Ostrich()
print(make_animal_move(duck))     # Output: I can fly and swim
print(make_animal_move(ostrich))  # Output: I can run
```


In this refactored version, all subclasses of `Animal` can be used interchangeably in the `make_animal_move` function without altering the correctness of the program, thus adhering to the Liskov Substitution Principle.


By following LSP, we create more robust and flexible code that's easier to extend and maintain. It helps in creating abstractions that are correct at all levels, leading to better object-oriented design.


### Interface Segregation Principle (ISP)

- Definition: Clients should not be forced to depend on interfaces they do not use.
- Explanation: This principle suggests that a class should not be forced to implement interfaces it doesn't use. Instead of having
  one large interface, it's beneficial to split it into smaller and more specific interfaces so that clients will only need to know
  about the methods that are of interest to them.

```python
from abc import ABC, abstractmethod
class MultiFunctionalDevice(ABC):
    @abstractmethod
    def print(self):
        pass

    @abstractmethod
    def scan(self):
        pass

    @abstractmethod
    def fax(self):
        pass

class BasicPrinter(MultiFunctionalDevice):
    def print(self):
        print("Printing document...")

    def scan(self):
        raise NotImplementedError("This device does not support scanning.")

    def fax(self):
        raise NotImplementedError("This device does not support faxing.")

```

In the above example, we have a `BasicPrinter` child class of the `MultiFunctionalDevice` class. But a `BasicPrinter` has no need for features defined in the parent which it doesn't have itself.  So `Exceptions` have to be raised if these methods are called.

**Here is better example which follows ISP:**
```python
from abc import ABC, abstractmethod
class Printable(ABC):
    @abstractmethod
    def print(self):
        pass

class Scannable(ABC):
    @abstractmethod
    def scan(self):
        pass

class Faxable(ABC):
    @abstractmethod
    def fax(self):
        pass

class BasicPrinter(Printable):
    def print(self):
        print("Printing document...")

class FaxScanner(Faxable, Scannable):
    def fax(self):
        print("Sending fax...")

    def scan(self):
        print("Scanning document...")

```

### Dependency Inversion Principle (DIP)

- Definition: High-level modules should not depend on low-level modules. Both should depend on abstractions.
- Explanation: This principle advocates for the decoupling of software modules. By ensuring that both high-level and low-level
  objects depend on the same abstractions, we can reduce the system's rigidity, making it more flexible and scalable.

```python
class LightBulb:
    def __init__(self):
        self.on = False

    def turn_on(self):
        self.on = True

    def turn_off(self):
        self.on = False

class Switch:
    def operate(self, bulb: LightBulb):
        if bulb.on:
            bulb.on = False
        else:
            bulb.on = True

```

In the above example, the class `Switch` is dependent on the `LightBulb` class.  `LightBulb` is an object which can have complex logic, and the `Switch` class depends on it because it specifically manipulates a `LightBulb` object.  But "switching" something is a high level and abstract operation, which we may want to extend to more objects other than `LightBulb`.

View [switchable.py](../code/ood/switchable.py) for a better implementation of DIP.

## Object Oriented Programming

Object oriented programming is a style that lends itself to the principle that a software solution should closely model the problem domain.

* This is accomplished by modeling real world entities or processes
    and representing them in software.

    * Choosing which model(s) to create can have a large influence on
        how the problem is addressed and the solution is formed.

    * Each model can be developed to a different level of detail.

Object orientation is characterized by the following.

### Encapsulation

* The coupling of data and methods, where the methods often provide a public interface to protected or hidden data.

* Related is this notion of hiding. We **hide** the data and the implementation of the method. Only the contract – the method name, the parameters an return type, and documented (albeit abstract) semantics – are significant.

```python
class BankAccount:
    def __init__(self):
        self.balance = 0

account = BankAccount()
account.balance = -100  # Directly modifying the balance - Not good!
print(account.balance)  # Prints -100

```

In the above example a `BankAccount` has a `balance` attribute which is accessed and modified outside of the class, violating Encapsulation

**Correct Encapsulation Implementation:**

```python
class BankAccount:
    def __init__(self):
        self._balance = 0  # private attribute

    def deposit(self, amount):
        if amount > 0:
            self._balance += amount
            return True
        return False

    def withdraw(self, amount):
        if 0 < amount <= self._balance:
            self._balance -= amount
            return True
        return False

    def get_balance(self):
        return self._balance


account = BankAccount()
account.deposit(100)
account.withdraw(50)
print(account.get_balance())  # Prints 50
```

### Inheritance

The derivation of specific data types from more general types, giving rise to a hierarchy of data types that benefits heavily from code reuse.

```python
class Shape:
    def __init__(self, sides):
        self.sides = sides

class Rectangle(Shape):
    def __init__(self, width):
        super().__init__(4)
        self.width = width

    def perimeter(self):
        return 4 * self.width

class Quadrilateral(Rectangle):
    def __init__(self, side_lengths):
        super().__init__(side_lengths[0])
        self.side_lengths = side_lengths

class Square(Shape):
    def __init__(self, side_length):
        super().__init__(4)
        self.side_length = side_length

    def area(self):
        return self.side_length * self.side_length

```

In the above example there are a lot of issues with Inheritance.

* `Quadrilateral` inherits from `Rectangle`, should be the other way around.
* `Rectangle` defines a perimeter method which assumes all sides are the same side.
  * `Quadrilateral` inherits this method leading to both shapes having an incorrect method
* `Rectangle` does not even except multiple side lengths in the `__init__`, leading to incorrect construction
* `Square` does not inherit from `Rectangle` and loses access to the `area` and `perimeter` methods

**Here is a cleaner version of the above:**
```python
class Shape():
    def __init__(self, sides):
        self.sides = sides

# class Triangle(Shape):
#     def __init__(self, side_lengths):
#         super().__init__(3)
#         self.side_lengths = side_lengths

class Quadrilateral(Shape):
    def __init__(self):
        super().__init__(4)

class Rectangle(Quadrilateral):
    def __init__(self, width, height):
        super().__init__()
        self.width = width
        self.height = height

    def area(self):
        return self.width * self.height

    def perimeter(self):
        return 2 * (self.width + self.height)

class Square(Rectangle):
    def __init__(self, side_length):
        super().__init__(side_length, side_length)

```

### Abstraction

* The focusing on the necessary details, ignoring the less important concrete details for specific data types.

```python
class Car:
    def __init__(self):
        self._engine_temperature = 70  # in degrees
        self._spark_plug_status = "Good"
        self._oil_level = "Full"

    def check_spark_plug(self):
        return self._spark_plug_status

    def get_engine_temperature(self):
        return self._engine_temperature

    def get_oil_level(self):
        return self._oil_level

    def drive(self):
        pass

```

In the above example, it is clear from the `drive` method being in the `Car` class that access to this class is intended for base car users who just want to drive the car.  But the `Car` class is also allowing them access to more specific and complicated details about the car, the spark plug status, oil level, etc.

**Here is a better implementation of Abstraction:**

```python
class Car:
    def __init__(self):
        self._engine = Engine()
        self._transmission = Transmission()

    def start(self):
        self._engine.start()

    def drive(self):
        pass

    def stop(self):
        self._engine.stop()

```

This this new example, the driver can stop, start, and drive the Car, but the finer technical details are likely contained in our `Engine` and `Transmission` classes.

### Polymorphism
#TODO: Add eafp vs lbyl example

* The substitution of variables or objects of one type with variables or objects of another type without loss of functionality.

* If more than one type implements a contract, each type that does is said to be polymorphic with respect to that contract.

* Consider the idea of a `speak` method. Every `animal` that we could put into a list, and have a loop iterate over the list and call `speak`, without knowing anything else about that `animal` object, can be said to be polymorphic for `speak`.

* Unlike statically checked languages like Java and C++, because Python is a *duck-typed* language, we cannot expect the Python interpreter to know whether or not a given object has the desired method. Even if we were to apply Python’s  Type Hints, they are not enforced. Accordingly, we have three choices:
* 1) assume that the method is available, and crash if it doesn’t
* 2) wrap the code in a try-except block
* 3) check to see if the method is present.

```python
class Dog():
    def bark(self):
        print("Woof!")

class Cat:
    def speak(self):
        print("Meow!")

class Snake:
    pass

animals = [Dog(), Cat(), Snake()]

for animal in animals:
    animal.speak()

```

The above raises an `Attribute` error for both `Dog` and `Snake` since neither has a `speak` method defined.

**Here is a better implementation of Polymorphism:**

```python
class Animal(ABC):
    @abstractmethod
    def make_sound(self):
        pass

class Dog(Animal):
    def make_sound(self):
        print("Woof!")

class Cat(Animal):
    def make_sound(self):
        print("Meow!")

class Bird(Animal):
    def make_sound(self):
        print("Chirp!")

animals: list[Animal] = [Dog(), Cat(), Bird()]

for animal in animals:
    animal.make_sound()

```
## Exercises

Download the **Challenge** Script and follow the instructions in the comments.  If you get stuck, you can always view the **Solution** script.

[Challenges](../code/ood/challenges/ood_challenge_1.py)

[Solutions](../code/ood/solutions/ood_solution_1.py)

