# Python Data Model

## Review of Object Oriented Concepts

In Python's object-oriented system, everything is an object, including code. This differs from languages like Java, which distinguish between primitives and objects.

Every Python object has three fundamental characteristics:

1. **Identity**: A unique, unchangeable identifier for the object (accessible via the `id()` function).

2. **Type**: Defines the operations and methods the object supports (retrieved using the `type()` function).

3. **Value**: Can be mutable (changeable) or immutable (unchangeable), depending on the object's type.

Python uses automatic memory management. Objects are never explicitly destroyed; instead, the interpreter uses reference counting and garbage collection to clean up unreachable objects.

Some objects, called containers, can hold references to other objects. These references are considered part of the container's value.


### Abstraction

- A process of hiding details of implementation in programs and data

Example:
```python
class Car:
    def start_engine(self):
        # Complex engine starting process hidden
        print("Engine started")

    def drive(self):
        # Complex driving mechanics hidden
        print("Car is moving")

my_car = Car()
my_car.start_engine()  # User doesn't need to know how the engine starts
my_car.drive()         # User doesn't need to know how the car moves
```

### Encapsulation

- A language mechanism for restricting direct access to some of the
  object's components

  **OR**

- A language construct that facilitates the bundling of data with the
  methods (or other functions) operating on that data

Example:
```python
class BankAccount:
    def __init__(self):
        self.__balance = 0  # Private attribute

    def deposit(self, amount):
        if amount > 0:
            self.__balance += amount

    def get_balance(self):
        return self.__balance

account = BankAccount()
account.deposit(100)
print(account.get_balance())  # Accessing balance through a method
# print(account.__balance)  # This would raise an AttributeError
```

### Polymorphism

- Providing a single interface to entities of different types.
  Polymorphism is the concept that multiple types of objects might be
  able to work in a given situation.

Example:
```python
class Dog:
    def speak(self):
        return "Woof!"

class Cat:
    def speak(self):
        return "Meow!"

def animal_sound(animal):
    return animal.speak()

dog = Dog()
cat = Cat()

print(animal_sound(dog))  # Output: Woof!
print(animal_sound(cat))  # Output: Meow!
```

### Inheritance

Enables new classes to receive—or inherit—the properties and methods of
existing classes

Example:
```python
class Animal:
    def __init__(self, name):
        self.name = name

    def speak(self):
        pass

class Dog(Animal):
    def speak(self):
        return f"{self.name} says Woof!"

class Cat(Animal):
    def speak(self):
        return f"{self.name} says Meow!"

dog = Dog("Buddy")
cat = Cat("Whiskers")

print(dog.speak())  # Output: Buddy says Woof!
print(cat.speak())  # Output: Whiskers says Meow!
```

### Composition

- Composition expresses relationships between objects. Think about a
  chair. A chair has a seat. A chair has a back. And a chair has a set
  of legs. The phrase "has a" implies a relationship where the chair
  owns, or at minimum, uses, another object. It is this "has a"
  relationship which is the basis for composition.

Example:
```python
class Engine:
    def start(self):
        return "Engine started"

class Wheels:
    def rotate(self):
        return "Wheels rotating"

class Car:
    def __init__(self):
        self.engine = Engine()
        self.wheels = Wheels()

    def drive(self):
        return f"{self.engine.start()}, {self.wheels.rotate()}"

my_car = Car()
print(my_car.drive())  # Output: Engine started, Wheels rotating
```

## Python Data Model

Python's data model is a [formal specification](https://docs.python.org/3/reference/datamodel.html) that defines how objects behave in the language. It emphasizes simplicity and extensibility through classes.

Key elements of the Python Data Model include:

### Special ("magic"/"dunder") methods:

```python
class Book:
    def __init__(self, title, author):
        self.title = title
        self.author = author

    def __str__(self):
        return f"{self.title} by {self.author}"

    def __len__(self):
        return len(self.title)

book = Book("1984", "George Orwell")
print(book)  # Output: 1984 by George Orwell
print(len(book))  # Output: 4
```

<!-- ### Descriptors:

The descriptor protocol in Python's data model is defined by implementing one or more of these methods: `__get__()`, `__set__()`, and `__delete__()`.

1. `__get__(self, obj, type=None)`: Called to get the attribute value. `obj` is the instance the attribute was accessed through, or None when accessed through the class.

2. `__set__(self, obj, value)`: Called to set the attribute value. `obj` is the instance the attribute was accessed through.

3. `__delete__(self, obj)`: Called to delete the attribute. `obj` is the instance the attribute was accessed through.

Descriptors allow you to customize how attribute access, assignment, and deletion are handled for class attributes. They are the mechanism behind properties, methods, and class methods in Python. When a class attribute is a descriptor, its `__get__()`, `__set__()`, or `__delete__()` method overrides Python's default behavior for attribute access.


=== "Descriptor"

    ```python
    class Celsius:
        def __init__(self, temperature=0):
            self._temperature = temperature

        def get(self):
            return self._temperature

        def set(self, value):
            self._temperature = value

    class Fahrenheit:
        def __init__(self, celsius):
            self._celsius = celsius

        def get(self):
            return (self._celsius.get() * 9/5) + 32

        def set(self, value):
            self._celsius.set((value - 32) * 5/9)

    class Temperature:
        def __init__(self):
            self._celsius = Celsius()
            self._fahrenheit = Fahrenheit(self._celsius)

        def get_celsius(self):
            return self._celsius.get()

        def set_celsius(self, value):
            self._celsius.set(value)

        def get_fahrenheit(self):
            return self._fahrenheit.get()

        def set_fahrenheit(self, value):
            self._fahrenheit.set(value)

    # Usage
    temp = Temperature()
    temp.set_celsius(25)
    print(f"Celsius: {temp.get_celsius()}°C")  # Output: Celsius: 25°C
    print(f"Fahrenheit: {temp.get_fahrenheit()}°F")  # Output: Fahrenheit: 77.0°F

    temp.set_fahrenheit(68)
    print(f"Celsius: {temp.get_celsius()}°C")  # Output: Celsius: 20.0°C
    print(f"Fahrenheit: {temp.get_fahrenheit()}°F")  # Output: Fahrenheit: 68.0°F
    ```

=== "Property"

    ```python
    class Temperature:
        def __init__(self):
            self.celsius = 0

        @property
        def celsius(self):
            return self._celsius

        @celsius.setter
        def celsius(self, value):
            if isinstance(value, (int, float)):
                self._celsius = value
            else:
                raise ValueError("Temperature must be a number")

        @property
        def fahrenheit(self):
            return (self._celsius * 9/5) + 32

        @fahrenheit.setter
        def fahrenheit(self, value):
            if isinstance(value, (int, float)):
                self._celsius = (value - 32) * 5/9
            else:
                raise ValueError("Temperature must be a number")

    # Usage
    temp = Temperature()
    temp.celsius = 25
    print(f"Celsius: {temp.celsius}°C")  # Output: Celsius: 25°C
    print(f"Fahrenheit: {temp.fahrenheit}°F")  # Output: Fahrenheit: 77.0°F

    temp.fahrenheit = 68
    print(f"Celsius: {temp.celsius}°C")  # Output: Celsius: 20.0°C
    print(f"Fahrenheit: {temp.fahrenheit}°F")  # Output: Fahrenheit: 68.0°F
    ``` -->

### Iteration:

In Python's iteration model, there are two key concepts: iterables and iterators.

1. **Iterable**: An object capable of returning its members one at a time. Examples include lists, strings, and files. Iterables implement the `__iter__()` method.

2. **Iterator**: An object representing a stream of data. Iterators implement both `__iter__()` and `__next__()` methods.

Let's look at examples of both:

* [iter_1.py](./code/python_data_model/iter_1.py)
* [fib.py](./code/python_data_model/fib.py)

### Context Managers:

The context manager protocol in Python's data model is defined by implementing two special methods: `__enter__()` and `__exit__()`. These methods allow an object to define actions to be taken when entering and exiting a context (as used with the `with` statement).

1. `__enter__(self)`: Called when entering the context. It should return the resource to be used in the context.

2. `__exit__(self, exc_type, exc_value, traceback)`: Called when exiting the context, even if an exception occurred. It can handle exceptions, perform cleanup, and determine if an exception should be re-raised.

Here's how these methods work in the FileManager example:

=== "Context Manager"

    ```python
    class FileManager:
        def __init__(self, filename, mode):
            self.filename = filename
            self.mode = mode
            self.file = None

        def __enter__(self):
            self.file = open(self.filename, self.mode) # python built-in function
            return self.file

        def __exit__(self, exc_type, exc_val, exc_tb):
            if self.file:
                self.file.close()

    with FileManager('example.txt', 'w') as f:
        f.write('Hello, Context Manager!')
    ```

    When using a `with` statement with a FileManager object:

    1. Python calls `FileManager('example.txt', 'w').__enter__()`.
    2. The `__enter__()` method opens the file and returns the file object.
    3. The body of the `with` statement is executed.
    4. After the `with` block, Python calls `__exit__()`, which closes the file.
    5. If an exception occurs in the `with` block, it's passed to `__exit__()`, which can handle it or let it propagate.

This protocol ensures proper resource management, automatically handling setup and cleanup operations. It's commonly used for file I/O, database connections, and other scenarios requiring resource acquisition and release.

The context manager protocol provides a clean and pythonic way to manage resources, ensuring proper setup and teardown regardless of how a block of code exits (normally or via an exception). It's a key feature of Python's "with" statement and is widely used for resource management tasks.

## Object Oriented Best Practices

1. **Identify and separate varying aspects**: Isolate the parts of your program that are likely to change from those that remain constant. This makes your code more flexible and easier to maintain.

2. **Encapsulate what varies**: By encapsulating (hiding) the parts that change, you protect the rest of your code from being affected by these changes.

3. **Favor composition over inheritance**: Use a "HAS-A" relationship instead of an "IS-A" relationship when possible. This allows for more flexibility and avoids the potential pitfalls of a rigid inheritance hierarchy.

4. **Program to interfaces, not implementations**: In Python, due to duck typing, this means focusing on the behavior (methods) an object should have, rather than its specific class. Abstract Base Classes (ABCs) can be used to define these interfaces more formally when needed.

Here's an example that demonstrates these principles:

```python
from abc import ABC, abstractmethod

# Interface (Abstract Base Class)
class PaymentProcessor(ABC):
    @abstractmethod
    def process_payment(self, amount):
        pass

# Concrete implementations
class CreditCardProcessor(PaymentProcessor):
    def process_payment(self, amount):
        print(f"Processing credit card payment of ${amount}")

class PayPalProcessor(PaymentProcessor):
    def process_payment(self, amount):
        print(f"Processing PayPal payment of ${amount}")

# Using composition to allow for flexible payment processing
class Order:
    def __init__(self, processor: PaymentProcessor):
        self.processor = processor

    def checkout(self, amount):
        self.processor.process_payment(amount)

# Usage
credit_order = Order(CreditCardProcessor())
credit_order.checkout(100)  # Output: Processing credit card payment of $100

paypal_order = Order(PayPalProcessor())
paypal_order.checkout(50)  # Output: Processing PayPal payment of $50
```

In this example:

1. We've identified that the payment processing method is likely to vary and separated it into its own class hierarchy.
2. The varying payment processing logic is encapsulated in separate classes (`CreditCardProcessor`, `PayPalProcessor`).
3. We use composition in the `Order` class, which has-a `PaymentProcessor`, rather than inheriting from it.
4. We program to the `PaymentProcessor` interface (ABC), allowing for easy substitution of different payment methods.

This design allows us to easily add new payment methods without changing the `Order` class, demonstrating the flexibility and maintainability gained from following these principles.

## Creating an Instance of a Class

The process of creating a class instance in Python involves two special methods: `__new__()` and `__init__()`.

### `object.__new__(cls[, ...])`

- A static method that creates and returns a new instance of the class.
- Takes the class (`cls`) as its first argument.
- Additional arguments are passed to the object constructor.

### `object.__init__(self[, ...])`

- Called after `__new__()` creates the instance, but before it's returned to the caller.
- Initializes the newly created object.
- Takes the instance (`self`) as its first argument, followed by any constructor arguments.
- In derived classes, must explicitly call the base class's `__init__()` method to ensure proper initialization:
  - `BaseClass.__init__(self, [args...])`
  - `super().__init__([args...])`

### Key Points

1. `__new__()` creates the object, while `__init__()` initializes it.
2. `__new__()` is called before `__init__()`.
3. `__new__()` returns the new instance, which is then passed to `__init__()` as `self`.
4. Overriding `__new__()` is less common than overriding `__init__()`.
5. Both methods are crucial in Python's object creation process, but serve different purposes.

Understanding the roles of `__new__()` and `__init__()` is important for advanced Python programming, especially when working with metaclasses or immutable types.

### Singleton Demo

Here's a practical example of a Singleton class that might be used for a database connection:

```python
class DatabaseConnection:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            print("Creating new instance")
            cls._instance = super().__new__(cls)
            # Initialize any expensive resources here

            # Dynamic attribute assignment
            cls._instance.connection = some_database_library.connect(...)
        return cls._instance

    def __init__(self):
        print("Initializing instance")
        # This will be called every time DatabaseConnection() is called,
        # even if a new instance isn't created

        # Reset State
        # or
        # Implement Logging Logic

    def connect(self):
        self.connection = some_database_library.connect(...)
        print("Connected to database")

    def is_connected(self):
        # Check if connection is alive
        return hasattr(self, 'connection') and self.connection.is_alive()

    def query(self, sql):
        if not self.is_connected():
            self.connect()
        print(f"Executing query: {sql}")
        return self.connection.execute(sql)
# Usage
db1 = DatabaseConnection()
db2 = DatabaseConnection()

# Check if db1 and db2 are the same instance
print(db1 is db2)  # True

db1.query("SELECT * FROM users")
db2.query("UPDATE users SET active = True")
# Output:
# Creating new instance
# Initializing instance
# Initializing instance
# True
# Executing query: SELECT * FROM users
# Executing query: UPDATE users SET active = True
```
In this example:

1. `__new__` is used to ensure only one instance of `DatabaseConnection` is ever created.
2. The first time `DatabaseConnection()` is called, a new instance is created and stored in `_instance`.
3. Subsequent calls return the existing instance without creating a new one.
4. `__init__` is called every time `DatabaseConnection()` is called, even when returning an existing instance.
5. Expensive setup (like establishing a database connection) is done in `__new__` to ensure it happens only once.

This demonstrates how `__new__` controls instance creation while `__init__` is still called for each `DatabaseConnection()` call. It's a practical use of the Singleton pattern for managing a shared resource like a database connection.

## Method Resolution Order (MRO)

In Python 3, all classes implicitly inherit from the root `object` class. When using multiple inheritance, Python creates a specific path through the inheritance tree to resolve method calls. This path is called the Method Resolution Order (MRO).

### Example

Consider this class hierarchy:

```python
class A:
    def say_hello(self):
        print("Hi, I am an A!")

class B(A):
    def say_hello(self):
        print("Hi, I am a B!")

class C(A):
    def say_hello(self):
        print("Hi, I am a C!")

class D(B, C):
    pass

d = D()
# Normal method call
d.say_hello()  # Output: Hi, I am a B!

# Calling A's say_hello directly
A.say_hello(d)  # Output: Hi, I am an A!

# Alternative syntax
d.__class__.__mro__[3].say_hello(d)  # Output: Hi, I am an A!
```


The `say_hello()` method from class B is called because B comes before C in D's inheritance list.

You can view a class's MRO using the `__mro__` attribute:

```python
print(D.__mro__)
# (<class '__main__.D'>, <class '__main__.B'>, <class '__main__.C'>, <class '__main__.A'>, <class 'object'>)

```
## Python Data Model Exercises and Solutions

## 1. Special Methods Exercise: Implement a Custom Sequence

Create a class called `Fibonacci` that represents the Fibonacci sequence. Implement the following special methods:

- `__init__(self, n)`: Initialize the sequence with `n` numbers.
- `__len__(self)`: Return the length of the sequence.
- `__getitem__(self, index)`: Allow indexing to retrieve Fibonacci numbers.
- `__str__(self)`: Return a string representation of the sequence.

Example usage:
```python
fib = Fibonacci(8)
print(len(fib))  # Should print 8
print(fib[6])    # Should print 8 (the 7th Fibonacci number)
print(fib)       # Should print something like "Fibonacci(0, 1, 1, 2, 3, 5, 8, 13)"
```

??? warning "Solution"

    ```python
    class Fibonacci:
        def __init__(self, n):
            self.n = n
            self.sequence = [0, 1]
            for i in range(2, n):
                self.sequence.append(self.sequence[i-1] + self.sequence[i-2])

        def __len__(self):
            return self.n

        def __getitem__(self, index):
            if 0 <= index < self.n:
                return self.sequence[index]
            raise IndexError("Fibonacci index out of range")

        def __str__(self):
            return f"Fibonacci({', '.join(map(str, self.sequence))})"

    # Test the implementation
    fib = Fibonacci(8)
    print(len(fib))  # Output: 8
    print(fib[6])    # Output: 8
    print(fib)       # Output: Fibonacci(0, 1, 1, 2, 3, 5, 8, 13)
    ```

## 2. Descriptor Exercise: Temperature Converter Descriptor

Create a descriptor class called `Celsius` that represents a temperature in Celsius. Implement the following:

- When the temperature is set, store it in Celsius.
- When the temperature is retrieved, return it in Celsius.
- Implement a property called `fahrenheit` that converts the temperature to Fahrenheit when accessed.

Use this descriptor in a `Temperature` class.

Example usage:
```python
class Temperature:
    celsius = Celsius()

temp = Temperature()
temp.celsius = 25
print(temp.celsius)     # Should print 25
print(temp.fahrenheit)  # Should print 77.0
```

??? warning "Solution"

    ```python
    class Celsius:
        def __init__(self):
            self._temperature = None

        def __get__(self, obj, objtype):
            return self._temperature

        def __set__(self, obj, value):
            if isinstance(value, (int, float)):
                self._temperature = value
            else:
                raise ValueError("Temperature must be a number")

    class Temperature:
        celsius = Celsius()

        @property
        def fahrenheit(self):
            return (self.celsius * 9/5) + 32

    # Test the implementation
    temp = Temperature()
    temp.celsius = 25
    print(temp.celsius)     # Output: 25
    print(temp.fahrenheit)  # Output: 77.0
    ```

## 3. Iteration Exercise: Custom Iterator for Prime Numbers

Create a class called `PrimeIterator` that generates prime numbers up to a specified limit. Implement the following:

- `__init__(self, limit)`: Initialize with an upper limit.
- `__iter__(self)`: Return the iterator object.
- `__next__(self)`: Generate the next prime number.

Example usage:
```python
primes = PrimeIterator(20)
for prime in primes:
    print(prime, end=' ')  # Should print: 2 3 5 7 11 13 17 19
```

??? warning "Solution"

    ```python
    class PrimeIterator:
        def __init__(self, limit):
            self.limit = limit
            self.current = 2

        def __iter__(self):
            return self

        def __next__(self):
            while self.current < self.limit:
                if self.is_prime(self.current):
                    prime = self.current
                    self.current += 1
                    return prime
                self.current += 1
            raise StopIteration

        def is_prime(self, n):
            if n < 2:
                return False
            for i in range(2, int(n**0.5) + 1):
                if n % i == 0:
                    return False
            return True

    # Test the implementation
    primes = PrimeIterator(20)
    for prime in primes:
        print(prime, end=' ')  # Output: 2 3 5 7 11 13 17 19
    ```

## 4. Context Manager Exercise: Database Connection Context Manager

Create a context manager class called `DatabaseConnection` that simulates opening and closing a database connection. Implement the following:

- `__init__(self, db_name)`: Initialize with a database name.
- `__enter__(self)`: Simulate opening a connection and return a cursor.
- `__exit__(self, exc_type, exc_value, traceback)`: Simulate closing the connection.

Example usage:
```python
with DatabaseConnection('mydb') as cursor:
    cursor.execute("SELECT * FROM users")  # Simulated SQL execution
```

??? warning "Solution"

    ```python
    class DatabaseConnection:
        def __init__(self, db_name):
            self.db_name = db_name
            self.connection = None

        def __enter__(self):
            print(f"Opening connection to {self.db_name}")
            self.connection = f"Connection to {self.db_name}"
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            if self.connection:
                print(f"Closing connection to {self.db_name}")
                self.connection = None

        def execute(self, query):
            if self.connection:
                print(f"Executing query: {query}")
            else:
                raise Exception("No active connection")

    # Test the implementation
    with DatabaseConnection('mydb') as cursor:
        cursor.execute("SELECT * FROM users")

    # Output:
    # Opening connection to mydb
    # Executing query: SELECT * FROM users
    # Closing connection to mydb
    ```

## 5. Method Resolution Order (MRO) Exercise: Analyze MRO in Multiple Inheritance

Given the following class hierarchy:

```python
class A:
    def method(self):
        return "A"

class B(A):
    def method(self):
        return "B"

class C(A):
    def method(self):
        return "C"

class D(B, C):
    pass
```

1. What is the Method Resolution Order (MRO) for class D?
2. What will be the output of `D().method()`?
3. Modify class D to call all parent methods using `super()`.

??? warning "Solution"

    * `1. The Method Resolution Order (MRO) for class D is:`

    ```python
    print(D.__mro__)
    # Output: (<class '__main__.D'>, <class '__main__.B'>, <class '__main__.C'>, <class '__main__.A'>, <class 'object'>)
    ```

    * `2. The output of `D().method()` will be:`

    ```python
    print(D().method())
    # Output: B
    ```

    * `3. Modified class D to call all parent methods using `super()`:`

    ```python
    class A:
        def method(self):
            return "A"

    class B(A):
        def method(self):
            return f"B > {super().method()}"

    class C(A):
        def method(self):
            return f"C > {super().method()}"

    class D(B, C):
        def method(self):
            return f"D > {super().method()}"

    # Test the implementation
    print(D().method())
    # Output: D > B > C > A
    ```

    This modified version demonstrates how `super()` follows the MRO to call each parent's method in the correct order.