# Unified Modeling Language

Unified Modeling Language (UML) is a general-purpose language to standardize visualizing system design. It began as the merged work of Grady Booch, James Rumbaugh, and Ivar Jacobson.

UML, itself, is as vast subject, easily on par in size and complexity to this course, if not larger. This unit merely provides an executive overview of one core concept.

UML provides a way to visualize:

- Any activities
- Components of the system
- How the system runs
- How entities interact (components and interfaces)
- External user interfaces

UML diagrams are a partial graphic representation of the system’s model. UML diagrams provide two views:

1. Static (structural)

    - Emphasizes structure using objects, attributes, operations, and relationships
    - Includes class diagrams and composite structure diagrams

2. Dynamic (behavioral)

    - Shows collaborations between objects and changes to internal state of objects
    - Includes sequence diagrams, activity diagrams, and state machine diagrams

There are a number of Open Source tools for authoring UML, including:

- [StarUML](http://staruml.io/)
- [Umbrello](https://apps.kde.org/umbrello/)
- [draw.io](https://app.diagrams.net/)
- PyLint3 (includes [Pyreverse](https://pylint.readthedocs.io/en/latest/pyreverse.html), which does the work)

```python
class Person:
    def __init__(self, name, age):
        self.name = name
        self.age = age

    def greet(self):
        return f"Hello, my name is {self.name}"

class Student(Person):
    def __init__(self, name, age, student_id):
        super().__init__(name, age)
        self.student_id = student_id

    def study(self):
        return f"{self.name} is studying"

class Course:
    def __init__(self, name, code):
        self.name = name
        self.code = code
        self.students = []

    def add_student(self, student):
        self.students.append(student)
```

```bash
pip install pylint==3.2.5
pyreverse -o png example.py
pyreverse -ASmy -o png example.py
# -A: Include association relationships
# -S: Include names of all attributes and methods
# m: Include module names in the diagram
# y: Show only classes in the diagram (no modules)
```

[Great UML Video](https://www.youtube.com/watch?v=HK_gBzLHdQQ)

## UML Class Diagrams

A class diagram is a static structure diagram that shows

- The system’s classes
- Their attributes
- Their operations (methods)
- Their relationships to each other

The class diagram is the main building block for object-oriented modeling. Class diagrams with detail can often be easily translated to code. Class diagrams can also be used for data modeling.


The diagram above illustrates the student class.

- The class diagram is broken into three parts.

    1. The first part indicates the name of the class.
    2. The second section lists attributes of the class. In general, they should be lowercase. Notice ID does not follow this guideline.
    3. The third section describes the operations (methods) of the class. Again, these are usually lowercase. Parameters for the methods could be included within the parentheses.

- The + (plus sign) indicates public visibility. In Python, this may not be too helpful, but in languages such as Java or C++, you can indicate whether a method or attribute is:

    - \+ public
    - \- private
    - \# protected
    - / derived (often combined with another symbol)
    - ~ package

- Scope, such as class or instance, can be indicated as well. Class members (class variables) are indicated by an underlined name.
- A challenge with Python and UML is how to describe the use of decorators (and metaclasses). One way to do so is by using a UML notation known as a stereotype, represented as «stereotypeName». The fancy symbols are known as guillemets Another way would be to use UML classifiers. There is no standard approach.

### Instance Relationships

In class diagrams, you can express that instances and classes are **associated** by drawing a line between them. A simple line indicates an association.

**Aggregation** can be thought of as "catalog" containment. A course would have a catalog (or list) of students. Aggregation is a more specific variation of the "has a" association relationship, and is indicated with an open diamond on the end connected to the containing class. For example, a course has students in the diagram below.


**Composition** is used to express when one class is part of the whole, as in an organ is part of the human body. Composition is often described as "physical" containment, and is indicated with a filled diamond on the end connected to the containing class. For example, a car has an engine which has a piston in the diagram below.


The final aspect of the diagram that we’ll mention is that each has-a relationship can have a **multiplicity**. That refers to the number of targets in the relationship. The default *multiplicity* is 1, which means that there must be 1 and only 1 target of the relationship. In the engine diagram, above, that means that a car has 1 engine, an engine has 1 car, an engine has 1 piston, and a piston has 1 engine. The relationship is mandatory.

To change the *multiplicity* from 1 to something else, we could use the following:

- 0..1: an optional relationship, with at most 1 target. This can be modeled with a simple object reference.
- 0..n: an optional relationship, with at most n targets. This might be modeled with a collection of object references.
- 0..\*: an optional relationship with an unspecified upper limit on the number of targets.
- Similarly, the relationship could be mandatory, as in the case of 1..N.
- The relationship could specify a number, N, where exactly N targets are necessary. For example, a V8 engine needs exactly 8 pistons.
- The general form is expressed as n..m, where n is the lower bound, and m is the upper bound. The default, if you don’t list a cardinality, would be 1..1, simplified as 1.
- Finally, we can express an enumeration of values, such as 4..6, 8, 12, perhaps representing that an engine can have 4, 5, 6, 8 or 12 pistons.

To revisit the car-engine-piston example, keeping them all as mandatory, we might (purely for illustrative purposes) describe a V8 engine as:


### Class Relationships

Generalization is used to show that one of the two related classes is considered a specialized form of the other.

The diagram below shows Instructor **is a** kind of person (usually). Also, a student **is a** kind of person.


Use an association with an open arrow head for the class that is the more general of the two. This relationship is described as **inheritance**. Person can be referred to as the super, base, or parent class. Student can be referred to as a subclass, derived, or child class.

## Design by Contract

Software designers should define formal, precise, and verifiable interface specifications for software components. The metaphor comes from business where a client and a supplier agree on a contract that defines:

- The supplier must provide a certain product.
- The client must pay the fee and is entitled to the product.

When we engage in Test-Driven Development, we are writing tests according to the contracts we assert in the tests, and validating that those contracts are met by the code under test. In object-oriented programming, if a routine provides a certain functionality, it may:

- Expect a certain condition to be guaranteed by any module that calls it.
- Guarantee a certain property on exit.
- Maintain a certain class property.

A developer must answer these three questions about their classes:

- What does the contract expect?
- What does the contract guarantee?
- What does the contract maintain?

The answers are known as the:

- Preconditions
- Postconditions
- Invariants

The idea of a contract extends to the method or procedure level. A method could include these additional pieces of information and their meanings:

- Acceptable and unacceptable input
- Return values and types
- Error and exception condition values and types
- Side effects
- Performance guarantees for time or space (rarely)
