import src.instructor_tools as it
from typing import Dict, Tuple, Type, ClassVar
from dataclasses import dataclass

class ExampleError(Exception):
    """Base exception class for example-related errors."""
    pass

class InvalidChoiceError(ExampleError):
    """Exception raised for invalid menu choices."""
    pass

@dataclass
class MutabilityExample:
    """Example demonstrating mutable vs immutable objects."""

    @staticmethod
    def demo_code() -> str:
        """Returns the demonstration code as a string."""
        return """
x = 10
y = x
x = x + 1
# x is 11, y is 10
if x == y:
    print("x and y are the same")
else:
    print("x and y are not the same")

l1 = [1, 2, 3]
l2 = l1
l1.append(4)
# l1 is [1, 2, 3, 4], l2 is [1, 2, 3, 4]
if l1 == l2:
    print("l1 and l2 are the same")
else:
    print("l1 and l2 are not the same")
"""

    @staticmethod
    def run() -> None:
        """Runs the mutability demonstration."""
        try:
            # Immutable example
            x: int = 10
            y: int = x
            x = x + 1

            it.cprint(
                "[blue]x and y are {}the same[/blue]".format(
                    "" if x == y else "not "
                )
            )

            # Mutable example
            l1: list[int] = [1, 2, 3]
            l2: list[int] = l1
            l1.append(4)

            it.cprint(
                "[blue]l1 and l2 are {}the same[/blue]".format(
                    "" if l1 == l2 else "not "
                )
            )

        except Exception as e:
            raise ExampleError(f"Error in mutability example: {e}")

# Define examples dictionary for the menu
examples: Dict[str, Tuple[str, Type[MutabilityExample]]] = {
    '1': ('Immutable Objects Example', MutabilityExample),
}

def get_user_choice() -> str:
    """
    Gets and validates user input for menu choice.

    Returns:
        The validated user choice

    Raises:
        InvalidChoiceError: If the input is invalid
    """
    try:
        choice = it.console.input("[green]Enter your choice (0-5): [/green]")
        if choice not in examples and choice != '0':
            raise InvalidChoiceError(f"Invalid choice: {choice}")
        return choice
    except Exception as e:
        raise InvalidChoiceError(f"Error getting user choice: {e}")

def run_example(choice: str) -> None:
    """
    Runs the selected example.

    Args:
        choice: The menu choice selected

    Raises:
        ExampleError: If there's an error running the example
    """
    try:
        name, example_class = examples[choice]
        it.cprint(f"\n[bold blue]Running: {name}[/bold blue]")

        # Show the demo code
        it.print_code(example_class.demo_code().strip())
        it.cprint("\n[bold magenta]Output:[/bold magenta]")

        # Run the example
        example_class.run()

    except Exception as e:
        raise ExampleError(f"Error running example {choice}: {e}")

def main() -> None:
    """Main function to run the interactive menu."""
    while True:
        try:
            it.clear_screen()
            it.display_menu("Mutable vs Immutable Examples", examples)

            choice = get_user_choice()
            if choice == '0':
                it.cprint("[red]Exiting the program. Goodbye![/red]")
                break

            run_example(choice)
            input("\nPress Enter to continue...")

        except InvalidChoiceError as e:
            it.cprint(f"[bold red]Invalid choice: {e}[/bold red]")
            input("\nPress Enter to continue...")
        except ExampleError as e:
            it.cprint(f"[bold red]Error running example: {e}[/bold red]")
            input("\nPress Enter to continue...")
        except it.ConsoleOutputError as e:
            it.cprint(f"[bold red]Console output error: {e}[/bold red]")
            input("\nPress Enter to continue...")
        except Exception as e:
            it.cprint(f"[bold red]Unexpected error: {e}[/bold red]")
            input("\nPress Enter to continue...")

if __name__ == "__main__":
    main()