def byte_to_variation_selector(byte: int) -> str:
    """Converts a byte (0-255) to its corresponding variation selector character."""
    if not 0 <= byte <= 255:
        raise ValueError("Byte must be between 0 and 255")

    if byte < 16:
        return chr(0xFE00 + byte)
    else:
        return chr(0xE0100 + (byte - 16))

def encode(base_char: str, bytes_to_encode: bytes) -> str:
    """Encodes a byte string into a base character followed by variation selectors."""
    if len(base_char) != 1:
        raise ValueError("Base character must be a single Unicode character")
    result = base_char
    for byte in bytes_to_encode:
        result += byte_to_variation_selector(byte)
    return result

def variation_selector_to_byte(char: str) -> int | None:
    """Converts a variation selector character to its corresponding byte (or None)."""
    if len(char) != 1:
        return None #Not a single character.

    codepoint = ord(char)
    if 0xFE00 <= codepoint <= 0xFE0F:
        return codepoint - 0xFE00
    elif 0xE0100 <= codepoint <= 0xE01EF:
        return codepoint - 0xE0100 + 16
    else:
        return None

def decode(encoded_string: str) -> bytes:
    """Decodes a string containing variation selectors back into the original bytes."""
    result = bytearray()
    found_vs = False  # Flag to track if we've encountered a variation selector
    for char in encoded_string:
        byte = variation_selector_to_byte(char)
        if byte is not None:
            found_vs = True
            result.append(byte)
        elif found_vs:
            # Stop decoding once we see a non-VS character *after* the first VS
            break  # Important to stop after the encoded data!
    return bytes(result)

if __name__ == "__main__":
    # --- Example Usage ---
    data_to_encode = "hello".encode('utf-8')  # Encode the string "hello" to bytes
    base_character = "😊"
    encoded_string = encode(base_character, data_to_encode)
    print(f"Encoded string: {encoded_string}")  # Output: 😊󠅘󠅕󠅜󠅜󠅟 (may look like just 😊)
    print(f"Encoded string (repr): {repr(encoded_string)}") #Shows the actual characters.

    decoded_bytes = decode(encoded_string)
    decoded_string = decoded_bytes.decode('utf-8', errors='ignore') # Decode back to string, handling potential errors
    print(f"Decoded bytes: {decoded_bytes}")  # Output: b'hello'
    print(f"Decoded string: {decoded_string}")  # Output: hello

    # --- Test Cases ---
    assert decode(encode("A", b"test")) == b"test"
    assert decode(encode("🙂", b"\x00\xff\x7f")) == b"\x00\xff\x7f"
    assert decode("A") == b""  # Empty input
    assert decode("😊") == b"" #Just the base character
    assert decode("😊A") == b""
    assert decode("😊󠅘A") == b"h"

    #Test using a string.
    encoded = encode("B", "This is a test.".encode('utf-8'))
    decoded = decode(encoded).decode('utf-8')
    print(f'Original: "This is a test." Encoded: {encoded}, Decoded: {decoded}')

    #Edge case: empty byte string
    encoded_empty = encode("X", b"")
    decoded_empty = decode(encoded_empty)
    assert decoded_empty == b""

    print("All tests passed!")

    decoded_bytes2 = decode("😊󠅘󠅕󠅜󠅜󠅟")
    print(decoded_bytes2)
    decoded_string2 = decoded_bytes2.decode('utf-8', errors='ignore')
    print(f"Decoded string2: {decoded_string2}")