"""Mutable vs Immutable Objects in Python - Deep vs Shallow Copy"""
# pylint: disable=W0105

from rich.console import Console
from rich.syntax import Syntax
from icecream import ic
import copy
import sys
import os

try:
    # add the directory above the current directory to the Python path
    sys.path.append(
        os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    )
    import src.instructor_tools as it
except ImportError:
    print("Error: instructor_tools not found")
    exit(1)

it.clear_screen()
c = Console()


# -----------------------------------------------------------------------
# Direct Reference Example
# -----------------------------------------------------------------------
def ex1():
    """Reference Code:
    import copy
    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = first__list
    """

    c.rule(
        title="[red]Direct Reference Example[/red]",
        style="bold green",
        align="center",
    )

    code_block = """import copy
temp = ["a", "b", "c"]
first__list = [temp, 2, 3]
second_list = first__list
    """

    it.print_comment("""
# Create a temporary mutable object
# and a list that contains it
# and a list that is a reference to the
# original list
                     """)
    code = Syntax(
        code_block,
        "python",
        line_numbers=True,
    )
    c.print(code)

    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = first__list

    ic(first__list)
    ic(second_list)
    ic(id(first__list))
    ic(id(second_list))
    ic(first__list is second_list)
    c.input(">>>")


# -----------------------------------------------------------------------
# Shallow Copy Example
# -----------------------------------------------------------------------
def ex2():
    """Reference Code:
    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)
    """

    c.rule(
        title="[red]Shallow Copy Example[/red]",
        style="bold green",
        align="center",
    )

    code_block = """temp = ["a", "b", "c"]
first__list = [temp, 2, 3]
second_list = copy.copy(first__list)  # shallow copy
# Could also use: second_list = list(first__list)
# Or: second_list = first__list[:]
    """

    it.print_comment("""
# Create a shallow copy of the list
# This creates a new list object, but the objects
# inside the list are still the same
                     """)
    code = Syntax(
        code_block,
        "python",
        line_numbers=True,
    )
    c.print(code)

    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)

    # Show the lists are different objects
    ic(first__list is second_list)
    c.input(">>>")

    it.print_comment("""
# But the contents of the lists refer to the same objects
# Let's check the IDs of each element
                     """)

    code = Syntax(
        """ic(id(first__list[0]), id(second_list[0]))
ic(id(first__list[1]), id(second_list[1]))
ic(id(first__list[2]), id(second_list[2]))""",
        "python",
        line_numbers=True,
    )
    c.print(code)

    ic(id(first__list[0]), id(second_list[0]))
    ic(id(first__list[1]), id(second_list[1]))
    ic(id(first__list[2]), id(second_list[2]))
    c.input(">>>")


# -----------------------------------------------------------------------
# Deep Copy Example
# -----------------------------------------------------------------------
def ex3():
    """Reference Code:
    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)
    third__list = copy.deepcopy(first__list)
    """

    c.rule(
        title="[red]Deep Copy Example[/red]",
        style="bold green",
        align="center",
    )

    code_block = """temp = ["a", "b", "c"]
first__list = [temp, 2, 3]
second_list = copy.copy(first__list)  # shallow copy
third__list = copy.deepcopy(first__list)  # deep copy
    """

    it.print_comment("""
# Create both shallow and deep copies to compare
# A deep copy creates new objects for all nested elements
                     """)
    code = Syntax(
        code_block,
        "python",
        line_numbers=True,
    )
    c.print(code)

    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)
    third__list = copy.deepcopy(first__list)

    # Check if the lists are the same object
    ic(first__list is third__list)
    c.input(">>>")

    it.print_comment("""
# Let's compare the IDs of each element in all three lists
                     """)

    code = Syntax(
        """ic(id(first__list[0]), id(second_list[0]), id(third__list[0]))
ic(id(first__list[1]), id(second_list[1]), id(third__list[1]))
ic(id(first__list[2]), id(second_list[2]), id(third__list[2]))""",
        "python",
        line_numbers=True,
    )
    c.print(code)

    ic(id(first__list[0]), id(second_list[0]), id(third__list[0]))
    ic(id(first__list[1]), id(second_list[1]), id(third__list[1]))
    ic(id(first__list[2]), id(second_list[2]), id(third__list[2]))
    c.input(">>>")


# -----------------------------------------------------------------------
# Modifying Immutable Objects Example
# -----------------------------------------------------------------------
def ex4():
    """Reference Code:
    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)
    third__list = copy.deepcopy(first__list)
    first__list[1] = 999
    """

    c.rule(
        title="[red]Modifying Immutable Objects Example[/red]",
        style="bold green",
        align="center",
    )

    code_block = """temp = ["a", "b", "c"]
first__list = [temp, 2, 3]
second_list = copy.copy(first__list)
third__list = copy.deepcopy(first__list)

# Modify an immutable object in the original list
first__list[1] = 999
    """

    it.print_comment("""
# Let's modify an immutable object (integer) in the original list
# and see how it affects the copies
                     """)
    code = Syntax(
        code_block,
        "python",
        line_numbers=True,
    )
    c.print(code)

    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)
    third__list = copy.deepcopy(first__list)

    first__list[1] = 999

    it.print_comment("# Let's see what changed in each list")
    code = Syntax(
        """ic(first__list)
ic(second_list)
ic(third__list)""",
        "python",
        line_numbers=True,
    )
    c.print(code)

    ic(first__list)
    ic(second_list)
    ic(third__list)
    c.input(">>>")


# -----------------------------------------------------------------------
# Modifying Nested Mutable Objects Example
# -----------------------------------------------------------------------
def ex5():
    """Reference Code:
    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)
    third__list = copy.deepcopy(first__list)
    first__list[0][2] = "zzz"
    """

    c.rule(
        title="[red]Modifying Nested Mutable Objects Example[/red]",
        style="bold green",
        align="center",
    )

    code_block = """temp = ["a", "b", "c"]
first__list = [temp, 2, 3]
second_list = copy.copy(first__list)
third__list = copy.deepcopy(first__list)

# Modify a mutable object nested in the original list
first__list[0][2] = "zzz"
    """

    it.print_comment("""
# Now let's modify a mutable object (inner list) in the original list
# This is where shallow vs deep copy differences become clear
                     """)
    code = Syntax(
        code_block,
        "python",
        line_numbers=True,
    )
    c.print(code)

    temp = ["a", "b", "c"]
    first__list = [temp, 2, 3]
    second_list = copy.copy(first__list)
    third__list = copy.deepcopy(first__list)


    first__list[0][2] = "zzz"

    it.print_comment("# Let's see what changed in each list")
    code = Syntax(
        """ic(first__list)
ic(second_list)
ic(third__list)""",
        "python",
        line_numbers=True,
    )
    c.print(code)

    ic(first__list)
    ic(second_list)
    ic(third__list)
    c.input(">>>")

    it.print_comment("""
# Let's check the IDs again to understand what happened
# Note how the IDs for the modified elements have changed
                     """)

    code = Syntax(
        """ic(id(first__list[0]), id(second_list[0]), id(third__list[0]))
ic(id(first__list[0][2]), id(second_list[0][2]), id(third__list[0][2]))
ic(id(first__list[1]), id(second_list[1]), id(third__list[1]))
ic(id(first__list[2]), id(second_list[2]), id(third__list[2]))""",
        "python",
        line_numbers=True,
    )
    c.print(code)

    ic(id(first__list[0]), id(second_list[0]), id(third__list[0]))
    ic(id(first__list[0][2]), id(second_list[0][2]), id(third__list[0][2]))
    ic(id(first__list[1]), id(second_list[1]), id(third__list[1]))
    ic(id(first__list[2]), id(second_list[2]), id(third__list[2]))
    # c.input(">>>")


# -----------------------------------------------------------------------
# Summary
# -----------------------------------------------------------------------
def ex6():
    """Summary of shallow vs deep copy"""

    c.rule(
        title="[red]Summary: Shallow vs Deep Copy[/red]",
        style="bold green",
        align="center",
    )

    it.print_comment("""
# Key takeaways:
#
# 1. A direct reference (assignment) creates a new name that points to the same object
#    - Changes to either variable affect the same object
#
# 2. A shallow copy (copy.copy, list(), or [:]) creates a new object but:
#    - Nested mutable objects are still shared references
#    - Modifying the top-level object doesn't affect the original
#    - Modifying nested mutable objects affects both copies
#
# 3. A deep copy (copy.deepcopy) creates completely independent copies:
#    - All nested objects are recursively copied
#    - Modifications to any level don't affect the original
#
# When to use each:
# - Use direct reference when you want multiple names to work with the same data
# - Use shallow copy when you want to modify the top-level container without affecting the original
# - Use deep copy when you need complete independence, especially with nested data structures
                     """)
    c.input(">>> Press Enter to exit...")


if __name__ == "__main__":
    try:
        ex1()
        ex2()
        ex3()
        ex4()
        ex5()
        ex6()
    except KeyboardInterrupt:
        print(">>> Exiting...\n")