"""This is an example of re-raising exceptions.

The main function demonstrates different exception handling scenarios.

It includes:

- Normal processing (file exists, no errors in data)
- File not found (no attempt at recovery)
- Data format error
- Unexpected error

"""

# pylint: disable=W1203,C0301

import logging
import argparse
import os
from typing import Optional

from rich.logging import RichHandler
from rich import print as rprint


class DataFormatError(Exception):
    """Custom exception for data format issues."""

    pass


def parse_data(data: str, logger: logging.Logger) -> str:
    """Simulates parsing data that may or may not have issues.

    Args:
        data: The data string to parse.
        logger: The logger instance.

    Returns:
        The processed data (uppercased).

    Raises:
        ValueError: If "ERROR_1" is found in the data.
        TypeError: If "ERROR_2" is found in the data.
        IndexError: If "ERROR_3" is found in the data.
        Exception: For any other unexpected error during parsing.
    """
    try:
        # Simulate some possible parsing errors
        if "ERROR_1" in data:
            raise ValueError("Invalid character found.")
        if "ERROR_3" in data:
            raise IndexError("Index out of range.")

        # Simulate successful data parsing
        processed_data = data.upper()  # Convert to uppercase
        return processed_data
    except Exception:
        logger.error("Error parsing data within parse_data", exc_info=True)
        raise


def process_data(filename: str, logger: logging.Logger) -> Optional[str]:
    """Processes data from a file, demonstrating exception handling and re-raising.

    Args:
        filename: The name of the file to process.
        logger: The logger instance.

    Returns:
        The processed data as a string, or None if an error occurred, and the error has been logged.

    Raises:
        FileNotFoundError: If the specified file does not exist.
        DataFormatError:  If a ValueError (related to data format) occurs within parse_data.
        Exception: For any other unexpected error during file processing.

    """
    try:
        with open(filename, "r", encoding="utf-8") as file:
            data = file.read()
            result = parse_data(data, logger)
            return result
    except FileNotFoundError:
        logger.error(f"Missing input file: {filename}")
        # raise  # Re-raises the FileNotFoundError
    except ValueError as e:
        # Transform this exception into a more specific one
        logger.warning(f"Data format issue in {filename}: {str(e)}")
        raise DataFormatError(f"Invalid format in {filename}: {str(e)}") from e
    except Exception:
        # Log unexpected errors with full details before re-raising
        logger.critical(f"Unexpected error processing {filename}", exc_info=True)
        # Re-raise whatever exception occurred
        raise


def setup_logging(log_level: str) -> logging.Logger:
    """Configures logging based on the specified log level.

    Args:
        log_level: The desired log level (e.g., "DEBUG", "INFO").

    Returns:
        The configured logger instance.

    Raises:
        ValueError: If the provided log level is invalid.
    """
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {log_level}")

    logging.basicConfig(
        level=numeric_level,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler()],
    )
    logger = logging.getLogger("rich")
    logger.debug(f"Logging level set to: {log_level.upper()}")  # Confirm setting
    return logger


def main() -> None:
    """Main function to parse arguments and demonstrate logging."""
    parser = argparse.ArgumentParser(
        description="Demonstrates logging with RichHandler and command-line log level control."
    )
    parser.add_argument(
        "--log-level",
        dest="log_level",
        default="INFO",  # Default log level
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level (DEBUG, INFO, WARNING, ERROR, CRITICAL).",
    )

    args = parser.parse_args()

    logger = setup_logging(args.log_level)

    # Create example files
    if not os.path.exists("data.txt"):
        with open("data.txt", "w", encoding="utf-8") as f:
            f.write("This is some good data.")
    if not os.path.exists("bad_data.txt"):
        with open("bad_data.txt", "w", encoding="utf-8") as f:
            f.write("ERROR_1:  Data with an invalid character.")
    if not os.path.exists("error_data.txt"):
        with open("error_data.txt", "w", encoding="utf-8") as f:
            f.write("ERROR_3: This will cause another type of problem")

    try:
        rprint(
            "[bold green]=== Case 1: Normal processing \
(file exists, no errors in data) ===[/bold green]"
        )
        result = process_data("data.txt", logger=logger)
        logger.info(f"Successfully processed data.txt: {result[:20]}...")
    except Exception as e:
        logger.error(f"Error processing data.txt: {e}")

    # Case 2: File not found (no attempt at recovery)
    try:
        rprint(
            "\n[bold red]=== Case 2: File not found (no attempt at recovery) ===[/bold red]"
        )
        result = process_data("missing_file.txt", logger=logger)
        logger.info(f"Successfully processed missing_file.txt: {result}")
    except FileNotFoundError as e:
        logger.error(f"Error processing missing_file.txt: {e}")
    except Exception as e:
        logger.error(f"Unexpected error processing missing_file.txt: {e}")

    # Case 3: Data format error
    try:
        rprint(
            "\n[bold magenta]=== Case 3: Data format error ===[/bold magenta]"
        )
        result = process_data("bad_data.txt", logger=logger)
        logger.info(f"Successfully processed bad_data.txt: {result}")
    except DataFormatError as e:
        logger.error(f"Error processing bad_data.txt: {e}")
    except Exception as e:
        logger.error(f"Unexpected error processing bad_data.txt: {e}")

    # Case 4: Unexpected error
    try:
        rprint("\n[bold cyan]=== Case 4: Unexpected error ===[/bold cyan]")
        result = process_data("error_data.txt", logger=logger)
        logger.info(f"Successfully processed error_data.txt: {result}")
    except Exception as e:
        logger.error(f"Error processing error_data.txt: {e}")


if __name__ == "__main__":
    main()