from rich import print as rprint
from rich.console import Console
from rich.syntax import Syntax
from rich.panel import Panel
from rich.markdown import Markdown
from rich.pretty import Pretty
from dataclasses import dataclass
from typing import List, Optional, Dict

@dataclass
class Example:
    """Represents a single code example with description and expected output"""
    code: str
    description: str
    expected_output: Optional[str] = None
    show_output: bool = True

class ExampleSection:
    """Represents a group of related examples"""
    def __init__(self, title: str, description: str, examples: List[Example]):
        self.title = title
        self.description = description
        self.examples = examples

class PythonDemo:
    """Base class for Python concept demonstrations"""
    def __init__(self):
        self.console = Console()
        self.locals = {}  # Store local variables for execution

    def display_code(self, code: str) -> None:
        """Display syntax-highlighted code"""
        syntax = Syntax(
            code,
            "python",
            theme="monokai",
            line_numbers=True,
            word_wrap=True,
        )
        self.console.print(syntax)

    def display_section_header(self, title: str, description: str) -> None:
        """Display a section header with description"""
        self.console.rule(f"[red]{title}[/red]", style="bold green", align="center")
        self.console.print(f"\n[bold cyan]{description}[/bold cyan]\n")

    def display_var(self, name: str, value: any) -> None:
        """Display a variable and its value"""
        self.console.print(f"[yellow]{name}[/yellow] = ", end="")
        self.console.print(Pretty(value))

    def execute_code(self, code: str) -> None:
        """Execute code after cleaning REPL markers and show results"""
        exec_code = (code
                    .replace(">>> ", "")
                    .replace("... ", ""))

        # First try to execute the whole block (for function definitions)
        try:
            exec(exec_code, globals(), self.locals)
            return
        except Exception:
            # If that fails, try line by line
            lines = exec_code.strip().split('\n')
            current_block = []

            for line in lines:
                if not line.strip() or line.strip().startswith('#'):
                    continue

                # If it's a function definition or starts with whitespace, collect the block
                if line.startswith('def ') or line[0].isspace():
                    current_block.append(line)
                    continue

                # If we have a block to execute
                if current_block:
                    try:
                        block_code = '\n'.join(current_block)
                        exec(block_code, globals(), self.locals)
                    except Exception as e:
                        self.console.print(f"[red]Error executing block: {e}[/red]")
                    current_block = []

                # Handle special cases for variable inspection
                if line.startswith('ic('):
                    var_name = line[3:-1]  # Extract variable name from ic(var)
                    if var_name in self.locals:
                        self.display_var(var_name, self.locals[var_name])
                else:
                    # Execute the single line
                    try:
                        exec(line, globals(), self.locals)
                    except Exception as e:
                        self.console.print(f"[red]Error executing line: {e}[/red]")

            # Execute any remaining block
            if current_block:
                try:
                    block_code = '\n'.join(current_block)
                    exec(block_code, globals(), self.locals)
                except Exception as e:
                    self.console.print(f"[red]Error executing final block: {e}[/red]")

    def wait_for_user(self) -> None:
        """Wait for user input to continue"""
        self.console.input("\n[bold green]Press Enter to continue...[/bold green]")

    def run_example(self, example: Example) -> None:
        """Run a single example"""
        # Show description
        self.console.print(f"\n[bold cyan]{example.description}[/bold cyan]")

        # Display and execute code
        self.display_code(example.code)
        if example.show_output:
            self.execute_code(example.code)

        # Check expected output if provided
        if example.expected_output:
            self.console.print(f"\n[yellow]Expected output:[/yellow]")
            self.console.print(example.expected_output)

        self.wait_for_user()

    def run_section(self, section: ExampleSection) -> None:
        """Run all examples in a section"""
        self.display_section_header(section.title, section.description)
        for example in section.examples:
            self.run_example(example)

class MutabilityDemo(PythonDemo):
    """Demonstration of Python's mutability concepts"""

    def __init__(self):
        super().__init__()
        self.sections = self._create_sections()

    def _create_sections(self) -> List[ExampleSection]:
        """Create all demonstration sections"""
        return [
            ExampleSection(
                "Mutable List Example",
                "Demonstrating how mutable objects share references",
                [
                    Example(
                        ">>> my_list = [1, 2, 3]\n>>> ic(my_list)",
                        "Creating a list and assigning it to a variable"
                    ),
                    Example(
                        ">>> another_list = my_list\n>>> ic(another_list)",
                        "Creating a reference to the same list"
                    ),
                    Example(
                        ">>> another_list.append(4)\n>>> ic(my_list)\n>>> ic(another_list)",
                        "Modifying the list through either reference shows changes in both"
                    ),
                ]
            ),
            ExampleSection(
                "Immutable String Example",
                "Demonstrating how immutable objects create new references",
                [
                    Example(
                        ">>> my_string = 'hello'\n>>> ic(my_string)",
                        "Creating a string (immutable object)"
                    ),
                    Example(
                        ">>> another_string = my_string\n>>> ic(another_string)",
                        "Creating a reference to the same string"
                    ),
                    Example(
                        ">>> another_string = another_string + ' world'\n>>> ic(my_string)\n>>> ic(another_string)",
                        "Modifying creates a new string, leaving original unchanged"
                    ),
                ]
            ),
            ExampleSection(
                "Function Arguments Example",
                "Demonstrating how mutability affects function arguments",
                [
                    Example(
                        """>>> def modify_list(a_list):
...     a_list.append(5)  # Modifies the original list
...
>>> def modify_string(a_string):
...     a_string = a_string + "!"  # Creates a new string
...     return a_string""",
                        "Defining functions that demonstrate mutability differences",
                        show_output=True
                    ),
                    Example(
                        ">>> test_list = [1, 2]\n>>> modify_list(test_list)\n>>> ic(test_list)",
                        "Mutable objects are modified directly within functions"
                    ),
                    Example(
                        ">>> test_string = 'Hi'\n>>> new_string = modify_string(test_string)\n>>> ic(test_string)\n>>> ic(new_string)",
                        "Immutable objects require returning new values from functions"
                    ),
                ]
            ),
        ]

    def run(self) -> None:
        """Run the complete demonstration"""
        try:
            # Display welcome message
            self.console.print(
                "\n[bold blue]Welcome to Python Mutability Demonstration![/bold blue]\n"
                "[cyan]This demo will show you how Python handles mutable and immutable objects.[/cyan]\n"
            )
            self.wait_for_user()

            # Run all sections
            for section in self.sections:
                self.run_section(section)

            # Display completion message
            self.console.rule("[red]Demo Complete[/red]", style="bold green", align="center")
            self.console.print("\n[bold green]Thanks for exploring Python mutability![/bold green]")

        except KeyboardInterrupt:
            self.console.print("\n[bold red]Demo interrupted by user[/bold red]")
        except Exception as e:
            self.console.print(f"\n[bold red]Error: {str(e)}[/bold red]")

if __name__ == "__main__":
    demo = MutabilityDemo()
    demo.run()