# instructor_tools.py
from typing import Any, Callable, Dict, Optional, Tuple, TypeVar
import os
import inspect
import io
import sys
from rich.console import Console
from rich.syntax import Syntax

def test_import():
    print("Hello from instructor_tools.py")


# Type definitions
T = TypeVar('T')
ExampleDict = Dict[str, Tuple[str, Any]]

class InstructorToolsError(Exception):
    """Base exception class for instructor tools."""
    pass

class ConsoleOutputError(InstructorToolsError):
    """Exception raised for errors in console output operations."""
    pass

# Initialize Rich console
console = Console()

def print_comment(comment: str):
    console.print(f"[bold cyan]{comment}[/bold cyan]")


def clear_screen() -> None:
    """Clears the terminal screen."""
    try:
        os.system('cls' if os.name == 'nt' else 'clear')
    except Exception as e:
        raise ConsoleOutputError(f"Failed to clear screen: {e}")

def display_menu(title: str, examples: ExampleDict) -> None:
    """
    Displays a formatted menu of examples.

    Args:
        title: The menu title
        examples: Dictionary mapping keys to (name, function) tuples

    Raises:
        ConsoleOutputError: If there's an error printing to the console
    """
    try:
        console.rule(f"[bold magenta]{title}")
        for key, (name, _) in examples.items():
            console.print(f"[magenta]{key}.[/magenta] [cyan]{name}[/cyan]")
        console.print("[red]0. Exit[/red]")
    except Exception as e:
        raise ConsoleOutputError(f"Failed to display menu: {e}")

def cprint(*args: Any, **kwargs: Any) -> None:
    """
    Prints to the console with rich formatting.

    Raises:
        ConsoleOutputError: If there's an error printing to the console
    """
    try:
        console.print(*args, **kwargs)
    except Exception as e:
        raise ConsoleOutputError(f"Failed to print to console: {e}")

def print_code(code_string: str, language: str = "python") -> None:
    """
    Prints a code block with syntax highlighting.

    Args:
        code_string: The code to display
        language: The programming language for syntax highlighting

    Raises:
        ConsoleOutputError: If there's an error printing the code
        ValueError: If code_string is empty or None
    """
    if not code_string:
        raise ValueError("Code string cannot be empty or None")

    try:
        syntax = Syntax(
            code_string,
            language,
            theme="monokai",
            line_numbers=False,
            word_wrap=True
        )
        console.print(syntax)
    except Exception as e:
        raise ConsoleOutputError(f"Failed to print code block: {e}")

def print_example_header(description: str) -> None:
    """
    Prints a colored header for an example.

    Args:
        description: The header text to display

    Raises:
        ConsoleOutputError: If there's an error printing the header
        ValueError: If description is empty or None
    """
    if not description:
        raise ValueError("Description cannot be empty or None")

    try:
        console.rule(f"[bold blue]{description}[/bold blue]")
    except Exception as e:
        raise ConsoleOutputError(f"Failed to print example header: {e}")

def run_example(func: Callable[..., T], *args: Any, **kwargs: Any) -> Optional[T]:
    """
    Runs an example function and captures/prints its output.

    Args:
        func: The function to run
        *args: Positional arguments for the function
        **kwargs: Keyword arguments for the function

    Returns:
        The function's return value if any

    Raises:
        ConsoleOutputError: If there's an error in console output
        ValueError: If func is None or not callable
    """
    if not func or not callable(func):
        raise ValueError("Function must be callable")

    try:
        print_example_header(f"Example: {func.__name__}")
        print_code(inspect.getsource(func))

        cprint("\n[bold magenta]Output:[/bold magenta]")
        result = func(*args, **kwargs)

        if result is not None:
            cprint(result, style="green")

        console.rule()
        return result

    except Exception as e:
        raise ConsoleOutputError(f"Failed to run example: {e}")
